'
' <copyright file="Authentication.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Collections
Imports System.Collections.Specialized
Imports System.Net
Imports System.Web
Imports System.Security.Cryptography
Imports System.Globalization
Imports System.Diagnostics


Namespace Microsoft.Samples.ServiceHosting.StorageClient
	''' <summary>
	''' This type represents the different constituent parts that make up a resource Uri in the context of cloud services.
	''' </summary>
	Public Class ResourceUriComponents
		''' <summary>
		''' The account name in the URI.
		''' </summary>
		Private privateAccountName As String
		Public Property AccountName() As String
			Get
				Return privateAccountName
			End Get
			Set(ByVal value As String)
				privateAccountName = value
			End Set
		End Property

		''' <summary>
		''' This is really the first component (delimited by '/') after the account name. Since it happens to
		''' be a container name in the context of all our storage services (containers in blob storage,
		''' queues in the queue service and table names in table storage), it's named as ContainerName to make it more 
		''' readable at the cost of slightly being incorrectly named.
		''' </summary>
		Private privateContainerName As String
		Public Property ContainerName() As String
			Get
				Return privateContainerName
			End Get
			Set(ByVal value As String)
				privateContainerName = value
			End Set
		End Property

		''' <summary>
		''' The remaining string in the URI.
		''' </summary>
		Private privateRemainingPart As String
		Public Property RemainingPart() As String
			Get
				Return privateRemainingPart
			End Get
			Set(ByVal value As String)
				privateRemainingPart = value
			End Set
		End Property

		''' <summary>
		''' Construct a ResourceUriComponents object.
		''' </summary>
		''' <param name="accountName">The account name that should become part of the URI.</param>
		''' <param name="containerName">The container name (container, queue or table name) that should become part of the URI.</param>
		''' <param name="remainingPart">Remaining part of the URI.</param>
		Public Sub New(ByVal accountName As String, ByVal containerName As String, ByVal remainingPart As String)
			Me.AccountName = accountName
			Me.ContainerName = containerName
			Me.RemainingPart = remainingPart
		End Sub

		''' <summary>
		''' Construct a ResourceUriComponents object.        
		''' </summary>
		''' <param name="accountName">The account name that should become part of the URI.</param>
		''' <param name="containerName">The container name (container, queue or table name) that should become part of the URI.</param>
		Public Sub New(ByVal accountName As String, ByVal containerName As String)
			Me.New(accountName, containerName, Nothing)
		End Sub

		''' <summary>
		''' Construct a ResourceUriComponents object.        
		''' </summary>
		''' <param name="accountName">The account name that should become part of the URI.</param>
		Public Sub New(ByVal accountName As String)
			Me.New(accountName, Nothing, Nothing)
		End Sub

		''' <summary>
		''' Construct a ResourceUriComponents object.        
		''' </summary>
		Public Sub New()
		End Sub
	End Class

	Friend NotInheritable Class MessageCanonicalizer
		''' <summary>
		''' An internal class that stores the canonicalized string version of an HTTP request.
		''' </summary>
		Private Class CanonicalizedString
			Private canonicalizedString_Renamed As New StringBuilder()

			''' <summary>
			''' Property for the canonicalized string.
			''' </summary>
			Friend ReadOnly Property Value() As String
				Get
					Return canonicalizedString_Renamed.ToString()
				End Get
			End Property

			''' <summary>
			''' Constructor for the class.
			''' </summary>
			''' <param name="initialElement">The first canonicalized element to start the string with.</param>
			Friend Sub New(ByVal initialElement As String)
				canonicalizedString_Renamed.Append(initialElement)
			End Sub

			''' <summary>
			''' Append additional canonicalized element to the string.
			''' </summary>
			''' <param name="element">An additional canonicalized element to append to the string.</param>
			Friend Sub AppendCanonicalizedElement(ByVal element As String)
				canonicalizedString_Renamed.Append(StorageHttpConstants.ConstChars.Linefeed)
				canonicalizedString_Renamed.Append(element)
			End Sub
		End Class

		''' <summary>
		''' Create a canonicalized string out of HTTP request header contents for signing 
		''' blob/queue requests with the Shared Authentication scheme. 
		''' </summary>
		''' <param name="address">The uri address of the HTTP request.</param>
		''' <param name="uriComponents">Components of the Uri extracted out of the request.</param>
		''' <param name="method">The method of the HTTP request (GET/PUT, etc.).</param>
		''' <param name="contentType">The content type of the HTTP request.</param>
		''' <param name="date">The date of the HTTP request.</param>
		''' <param name="headers">Should contain other headers of the HTTP request.</param>
		''' <returns>A canonicalized string of the HTTP request.</returns>
		Private Sub New()
		End Sub
		<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1308:NormalizeStringsToUppercase", Justification := "Authentication algorithm requires canonicalization by converting to lower case")> _
		Friend Shared Function CanonicalizeHttpRequest(ByVal address As Uri, ByVal uriComponents As ResourceUriComponents, ByVal method As String, ByVal contentType As String, ByVal [date] As String, ByVal headers As NameValueCollection) As String
			' The first element should be the Method of the request.
			' I.e. GET, POST, PUT, or HEAD.
			Dim canonicalizedString As New CanonicalizedString(method)

			' The second element should be the MD5 value.
			' This is optional and may be empty.
			Dim httpContentMD5Value As String = String.Empty

			' First extract all the content MD5 values from the header.
			Dim httpContentMD5Values As ArrayList = HttpRequestAccessor.GetHeaderValues(headers, StorageHttpConstants.HeaderNames.ContentMD5)

			' If we only have one, then set it to the value we want to append to the canonicalized string.
			If httpContentMD5Values.Count = 1 Then
				httpContentMD5Value = CStr(httpContentMD5Values(0))
			End If

			canonicalizedString.AppendCanonicalizedElement(httpContentMD5Value)

			' The third element should be the content type.
			canonicalizedString.AppendCanonicalizedElement(contentType)

			' The fourth element should be the request date.
			' See if there's an storage date header.
			' If there's one, then don't use the date header.
			Dim httpStorageDateValues As ArrayList = HttpRequestAccessor.GetHeaderValues(headers, StorageHttpConstants.HeaderNames.StorageDateTime)
			If httpStorageDateValues.Count > 0 Then
				[date] = Nothing
			End If

			canonicalizedString.AppendCanonicalizedElement([date])

			' Look for header names that start with StorageHttpConstants.HeaderNames.PrefixForStorageHeader
			' Then sort them in case-insensitive manner.
			Dim httpStorageHeaderNameArray As New ArrayList()
			For Each key As String In headers.Keys
				If key.ToLowerInvariant().StartsWith(StorageHttpConstants.HeaderNames.PrefixForStorageHeader, StringComparison.Ordinal) Then
					httpStorageHeaderNameArray.Add(key.ToLowerInvariant())
				End If
			Next key

			httpStorageHeaderNameArray.Sort()

			' Now go through each header's values in the sorted order and append them to the canonicalized string.
			For Each key As String In httpStorageHeaderNameArray
				Dim canonicalizedElement As New StringBuilder(key)
				Dim delimiter As String = ":"
				Dim values As ArrayList = HttpRequestAccessor.GetHeaderValues(headers, key)

				' Go through values, unfold them, and then append them to the canonicalized element string.
				For Each value As String In values
					' Unfolding is simply removal of CRLF.
					Dim unfoldedValue As String = value.Replace(StorageHttpConstants.ConstChars.CarriageReturnLinefeed, String.Empty)

					' Append it to the canonicalized element string.
					canonicalizedElement.Append(delimiter)
					canonicalizedElement.Append(unfoldedValue)
					delimiter = ","
				Next value

				' Now, add this canonicalized element to the canonicalized header string.
				canonicalizedString.AppendCanonicalizedElement(canonicalizedElement.ToString())
			Next key

			' Now we append the canonicalized resource element.
			Dim canonicalizedResource As String = GetCanonicalizedResource(address, uriComponents)
			canonicalizedString.AppendCanonicalizedElement(canonicalizedResource)

			Return canonicalizedString.Value
		End Function

		Friend Shared Function GetCanonicalizedResource(ByVal address As Uri, ByVal uriComponents As ResourceUriComponents) As String
			' Algorithem is as follows
			' 1. Start with the empty string ("")
			' 2. Append the account name owning the resource preceded by a �/�. This is not 
			'    the name of the account making the request but the account that owns the 
			'    resource being accessed.
			' 3. Append the path part of the un-decoded HTTP Request-URI, up-to but not 
			'    including the query string.
			' 4. If the request addresses a particular component of a resource, like?comp=
			'    metadata then append the sub-resource including question mark (like ?comp=
			'    metadata)
			Dim canonicalizedResource As New StringBuilder(StorageHttpConstants.ConstChars.Slash)
			canonicalizedResource.Append(uriComponents.AccountName)

			' Note that AbsolutePath starts with a '/'.
			canonicalizedResource.Append(address.AbsolutePath)

			Dim queryVariables As NameValueCollection = HttpUtility.ParseQueryString(address.Query)
			Dim compQueryParameterValue As String = queryVariables(StorageHttpConstants.QueryParams.QueryParamComp)
			If compQueryParameterValue IsNot Nothing Then
				canonicalizedResource.Append(StorageHttpConstants.ConstChars.QuestionMark)
				canonicalizedResource.Append(StorageHttpConstants.QueryParams.QueryParamComp)
				canonicalizedResource.Append(StorageHttpConstants.QueryParams.SeparatorForParameterAndValue)
				canonicalizedResource.Append(compQueryParameterValue)
			End If

			Return canonicalizedResource.ToString()
		End Function


		''' <summary>
		''' Canonicalize HTTP header contents.
		''' </summary>
		''' <param name="request">An HttpWebRequest object.</param>
		''' <param name="uriComponents">Components of the Uri extracted out of the request.</param>
		''' <returns>The canonicalized string of the given HTTP request's header.</returns>
		Friend Shared Function CanonicalizeHttpRequest(ByVal request As HttpWebRequest, ByVal uriComponents As ResourceUriComponents) As String
			Return CanonicalizeHttpRequest(request.Address, uriComponents, request.Method, request.ContentType, String.Empty, request.Headers)
		End Function

		''' <summary>
		''' Creates a standard datetime string for the shared key lite authentication scheme.
		''' </summary>
		''' <param name="dateTime">DateTime value to convert to a string in the expected format.</param>
		''' <returns></returns>
		Friend Shared Function ConvertDateTimeToHttpString(ByVal dateTime As DateTime) As String
			' On the wire everything should be represented in UTC. This assert will catch invalid callers who
			' are violating this rule.
			Debug.Assert(dateTime = DateTime.MaxValue OrElse dateTime = DateTime.MinValue OrElse dateTime.Kind = DateTimeKind.Utc)

			' 'R' means rfc1123 date which is what the storage services use for all dates...
			' It will be in the following format:
			' Sun, 28 Jan 2008 12:11:37 GMT
			Return dateTime.ToString("R", CultureInfo.InvariantCulture)
		End Function

		Private Shared Function AppendStringToCanonicalizedString(ByVal canonicalizedString As StringBuilder, ByVal stringToAppend As String) As String
			canonicalizedString.Append(StorageHttpConstants.ConstChars.Linefeed)
			canonicalizedString.Append(stringToAppend)
			Return canonicalizedString.ToString()
		End Function

		Friend Shared Function CanonicalizeHttpRequestForSharedKeyLite(ByVal request As HttpWebRequest, ByVal uriComponents As ResourceUriComponents, ByVal [date] As String) As String
			Dim canonicalizedString As New StringBuilder([date])
			AppendStringToCanonicalizedString(canonicalizedString, MessageCanonicalizer.GetCanonicalizedResource(request.Address, uriComponents))

			Return canonicalizedString.ToString()
		End Function
	End Class

	''' <summary>
	''' Use this class to extract various header values from Http requests.
	''' </summary>
	Friend NotInheritable Class HttpRequestAccessor
		''' <summary>
		''' A helper function for extracting HTTP header values from a NameValueCollection object.
		''' </summary>
		''' <param name="headers">A NameValueCollection object that should contain HTTP header name-values pairs.</param>
		''' <param name="headerName">Name of the header that we want to get values of.</param>
		''' <returns>A array list of values for the header. The values are in the same order as they are stored in the NameValueCollection object.</returns>
		Private Sub New()
		End Sub
		Friend Shared Function GetHeaderValues(ByVal headers As NameValueCollection, ByVal headerName As String) As ArrayList
			Dim arrayOfValues As New ArrayList()
			Dim values() As String = headers.GetValues(headerName)

			If values IsNot Nothing Then
				For Each value As String In values
					' canonization formula requires the string to be left trimmed.
					arrayOfValues.Add(value.TrimStart())
				Next value
			End If

			Return arrayOfValues
		End Function


		''' <summary>
		''' Constructs an URI given all its constituents
		''' </summary>
		''' <param name="endpoint">
		''' This is the service endpoint in case of path-style URIs and a host suffix in case of host-style URIs
		''' IMPORTANT: This does NOT include the service name or account name
		''' </param>
		''' <param name="uriComponents">Uri constituents</param>
		''' <param name="pathStyleUri">Indicates whether to construct a path-style Uri (true) or host-style URI (false)</param>
		''' <returns>Full uri</returns>
		Public Shared Function ConstructResourceUri(ByVal endpoint As Uri, ByVal uriComponents As ResourceUriComponents, ByVal pathStyleUri As Boolean) As Uri
			Return If(pathStyleUri, ConstructPathStyleResourceUri(endpoint, uriComponents), ConstructHostStyleResourceUri(endpoint, uriComponents))
		End Function

		''' <summary>
		''' Constructs a path-style resource URI given all its constituents
		''' </summary>
		Private Shared Function ConstructPathStyleResourceUri(ByVal endpoint As Uri, ByVal uriComponents As ResourceUriComponents) As Uri
			Dim path As New StringBuilder(String.Empty)
			If uriComponents.AccountName IsNot Nothing Then
				path.Append(uriComponents.AccountName)

				If uriComponents.ContainerName IsNot Nothing Then
					path.Append(StorageHttpConstants.ConstChars.Slash)
					path.Append(uriComponents.ContainerName)

					If uriComponents.RemainingPart IsNot Nothing Then
						path.Append(StorageHttpConstants.ConstChars.Slash)
						path.Append(uriComponents.RemainingPart)
					End If
				End If
			End If

			Return ConstructUriFromUriAndString(endpoint, path.ToString())
		End Function

		''' <summary>
		''' Constructs a host-style resource URI given all its constituents
		''' </summary>
		Private Shared Function ConstructHostStyleResourceUri(ByVal hostSuffix As Uri, ByVal uriComponents As ResourceUriComponents) As Uri
			If uriComponents.AccountName Is Nothing Then
				' When there is no account name, full URI is same as hostSuffix
				Return hostSuffix
			Else
				' accountUri will be something like "http://accountname.hostSuffix/" and then we append
				' container name and remaining part if they are present.
				Dim accountUri As Uri = ConstructHostStyleAccountUri(hostSuffix, uriComponents.AccountName)
				Dim path As New StringBuilder(String.Empty)
				If uriComponents.ContainerName IsNot Nothing Then
					path.Append(uriComponents.ContainerName)

					If uriComponents.RemainingPart IsNot Nothing Then
						path.Append(StorageHttpConstants.ConstChars.Slash)
						path.Append(uriComponents.RemainingPart)
					End If
				End If

				Return ConstructUriFromUriAndString(accountUri, path.ToString())
			End If
		End Function


		''' <summary>
		''' Given the host suffix part, service name and account name, this method constructs the account Uri
		''' </summary>
		Private Shared Function ConstructHostStyleAccountUri(ByVal hostSuffix As Uri, ByVal accountName As String) As Uri
			' Example: 
			' Input: serviceEndpoint="http://blob.windows.net/", accountName="youraccount"
			' Output: accountUri="http://youraccount.blob.windows.net/"
			Dim serviceUri As Uri = hostSuffix

			' serviceUri in our example would be "http://blob.windows.net/"
			Dim accountUriString As String = String.Format(CultureInfo.InvariantCulture, "{0}{1}{2}.{3}:{4}/", serviceUri.Scheme, Uri.SchemeDelimiter, accountName, serviceUri.Host, serviceUri.Port)

			Return New Uri(accountUriString)
		End Function

		Private Shared Function ConstructUriFromUriAndString(ByVal endpoint As Uri, ByVal path As String) As Uri
			' This is where we encode the url path to be valid
			Dim encodedPath As String = HttpUtility.UrlEncode(path)
			Return New Uri(endpoint, encodedPath)
		End Function
	End Class

	''' <summary>
	''' Objects of this class contain the credentials (name and key) of a storage account.
	''' </summary>
	Public Class SharedKeyCredentials

		''' <summary>
		''' Create a SharedKeyCredentials object given an account name and a shared key.
		''' </summary>
		Public Sub New(ByVal accountName As String, ByVal key() As Byte)
			Me.accountName = accountName
			Me.key = key
		End Sub

		''' <summary>
		''' Signs the request appropriately to make it an authenticated request.
		''' Note that this method takes the URI components as decoding the URI components requires the knowledge
		''' of whether the URI is in path-style or host-style and a host-suffix if it's host-style.
		''' </summary>
		Public Sub SignRequest(ByVal request As HttpWebRequest, ByVal uriComponents As ResourceUriComponents)
			If request Is Nothing Then
				Throw New ArgumentNullException("request")
			End If
			Dim message As String = MessageCanonicalizer.CanonicalizeHttpRequest(request, uriComponents)
			Dim computedBase64Signature As String = ComputeMacSha(message)
			request.Headers.Add(StorageHttpConstants.HeaderNames.Authorization, String.Format(CultureInfo.InvariantCulture, "{0} {1}:{2}", StorageHttpConstants.AuthenticationSchemeNames.SharedKeyAuthSchemeName, accountName, computedBase64Signature))
		End Sub

		''' <summary>
		''' Signs requests using the SharedKeyLite authentication scheme with is used for the table storage service.
		''' </summary>
		<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId := "Lite", Justification := "Name of the authentication scheme in the REST protocol")> _
		Public Sub SignRequestForSharedKeyLite(ByVal request As HttpWebRequest, ByVal uriComponents As ResourceUriComponents)
			If request Is Nothing Then
				Throw New ArgumentNullException("request")
			End If

			' add the date header to the request
			Dim dateString As String = MessageCanonicalizer.ConvertDateTimeToHttpString(DateTime.UtcNow)
			request.Headers.Add(StorageHttpConstants.HeaderNames.StorageDateTime, dateString)

			' compute the signature and add the authentication scheme
			Dim message As String = MessageCanonicalizer.CanonicalizeHttpRequestForSharedKeyLite(request, uriComponents, dateString)
			Dim computedBase64Signature As String = ComputeMacSha(message)
			request.Headers.Add(StorageHttpConstants.HeaderNames.Authorization, String.Format(CultureInfo.InvariantCulture, "{0} {1}:{2}", StorageHttpConstants.AuthenticationSchemeNames.SharedKeyLiteAuthSchemeName, accountName, computedBase64Signature))
		End Sub


		Private Function ComputeMacSha(ByVal canonicalizedString As String) As String
			Dim dataToMAC() As Byte = System.Text.Encoding.UTF8.GetBytes(canonicalizedString)

			Using hmacsha1 As New HMACSHA256(key)
				Return System.Convert.ToBase64String(hmacsha1.ComputeHash(dataToMAC))
			End Using
		End Function

		Private accountName As String
		Private key() As Byte
	End Class
End Namespace
