'
' <copyright file="BlobStorage.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.IO
Imports System.Collections.Specialized
Imports System.Threading
Imports System.Diagnostics.CodeAnalysis
Imports System.Globalization
Imports System.Diagnostics

<Assembly: CLSCompliant(True)> 

' disable the generation of warnings for missing documentation elements for 
' public classes/members in this file

Namespace Microsoft.Samples.ServiceHosting.StorageClient

    ''' <summary>
    ''' This delegate define the shape of a retry policy. A retry policy will invoke the given
    ''' <paramref name="action"/> as many times as it wants to in the face of 
    ''' retriable StorageServerExceptions.
    ''' </summary>
    ''' <param name="action">The action to retry</param>
    ''' <returns></returns>
    Public Delegate Sub RetryPolicy(ByVal action As Action)

#Region "Blob Storage API"
    ''' <summary>
    ''' The entry point of the blob storage API
    ''' </summary>
    Public MustInherit Class BlobStorage

        ''' <summary>
        ''' Factory method for BlobStorage
        ''' </summary>
        ''' <param name="baseUri">The base URI of the blob storage service</param>
        ''' <param name="usePathStyleUris">If true, path-style URIs (http://baseuri/accountname/containername/objectname) are used.
        ''' If false host-style URIs (http://accountname.baseuri/containername/objectname) are used,
        ''' where baseuri is the URI of the service.
        ''' If null, the choice is made automatically: path-style URIs if host name part of base URI is an 
        ''' IP addres, host-style otherwise.</param>
        ''' <param name="accountName">The name of the storage account</param>
        ''' <param name="base64Key">Authentication key used for signing requests</param>
        ''' <returns>A newly created BlobStorage instance</returns>
        Public Shared Function Create(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String) As BlobStorage
            Return New BlobStorageRest(baseUri, usePathStyleUris, accountName, base64Key)
        End Function

        ''' <summary>
        ''' Factory method for BlobStorage
        ''' </summary>
        ''' <param name="accountInfo">Account information</param>
        ''' <returns>A newly created BlobStorage instance</returns>
        Public Shared Function Create(ByVal accountInfo As StorageAccountInfo) As BlobStorage
            Return New BlobStorageRest(accountInfo.BaseUri, accountInfo.UsePathStyleUris, accountInfo.AccountName, accountInfo.Base64Key)
        End Function


        ''' <summary>
        ''' Get a reference to a newly created BlobContainer object.
        ''' This method does not make any calls to the storage service.
        ''' </summary>
        ''' <param name="containerName">The name of the container</param>
        ''' <returns>A reference to a newly created BlobContainer object</returns>
        Public MustOverride Function GetBlobContainer(ByVal containerName As String) As BlobContainer


        ''' <summary>
        ''' Lists the containers within the account.
        ''' </summary>
        ''' <returns>A list of containers</returns>
        Public MustOverride Function ListBlobContainers() As IEnumerable(Of BlobContainer)

        ''' <summary>
        ''' The time out for each request to the storage service.
        ''' </summary>
        Private privateTimeout As TimeSpan
        Public Property Timeout() As TimeSpan
            Get
                Return privateTimeout
            End Get
            Set(ByVal value As TimeSpan)
                privateTimeout = value
            End Set
        End Property

        ''' <summary>
        ''' The retry policy used for retrying requests
        ''' </summary>
        Private privateRetryPolicy As RetryPolicy
        Public Property RetryPolicy() As RetryPolicy
            Get
                Return privateRetryPolicy
            End Get
            Set(ByVal value As RetryPolicy)
                privateRetryPolicy = value
            End Set
        End Property

        ''' <summary>
        ''' The base URI of the blob storage service
        ''' </summary>
        Public ReadOnly Property BaseUri() As Uri
            Get
                Return Me.baseUri_Renamed
            End Get
        End Property

        ''' <summary>
        ''' The name of the storage account
        ''' </summary>
        Public ReadOnly Property AccountName() As String
            Get
                Return Me.accountName_Renamed
            End Get
        End Property

        ''' <summary>
        ''' Indicates whether to use/generate path-style or host-style URIs
        ''' </summary>
        Public ReadOnly Property UsePathStyleUris() As Boolean
            Get
                Return Me.usePathStyleUris_Renamed
            End Get
        End Property

        ''' <summary>
        ''' The default timeout
        ''' </summary>
        <SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes", Justification:="TimeSpan is a non-mutable type")> _
        Public Shared ReadOnly DefaultTimeout As TimeSpan = TimeSpan.FromSeconds(30)

        ''' <summary>
        ''' The default retry policy
        ''' </summary>
        <SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes", Justification:="RetryPolicy is a non-mutable type")> _
        Public Shared ReadOnly DefaultRetryPolicy As RetryPolicy = AddressOf RetryPolicies.NoRetry


        Protected Friend Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String)
            Me.baseUri_Renamed = baseUri
            Me.accountName_Renamed = accountName
            Me.Base64Key = base64Key
            If Not usePathStyleUris.HasValue Then
                Me.usePathStyleUris_Renamed = Utilities.StringIsIPAddress(baseUri.Host)
            Else
                Me.usePathStyleUris_Renamed = usePathStyleUris.Value
            End If

            Timeout = DefaultTimeout
            RetryPolicy = DefaultRetryPolicy
        End Sub

        Private usePathStyleUris_Renamed As Boolean
        Private baseUri_Renamed As Uri
        Private accountName_Renamed As String
        Private privateBase64Key As String
        Protected Friend Property Base64Key() As String
            Get
                Return privateBase64Key
            End Get
            Set(ByVal value As String)
                privateBase64Key = value
            End Set
        End Property
    End Class


    ''' <summary>
    ''' Provides definitions for some standard retry policies.
    ''' </summary>
    Public NotInheritable Class RetryPolicies

        Public Shared ReadOnly StandardMinBackoff As TimeSpan = TimeSpan.FromMilliseconds(100)
        Public Shared ReadOnly StandardMaxBackoff As TimeSpan = TimeSpan.FromSeconds(30)
        Private Shared ReadOnly Random As New Random()

        ''' <summary>
        ''' Policy that does no retries i.e., it just invokes <paramref name="action"/> exactly once
        ''' </summary>
        ''' <param name="action">The action to retry</param>
        ''' <returns>The return value of <paramref name="action"/></returns>
        Private Sub New()
        End Sub
        Public Shared Sub NoRetry(ByVal action As Action)
            Try
                action()
            Catch e As TableRetryWrapperException
                Throw e.InnerException
            End Try
        End Sub

        ''' <summary>
        ''' Policy that retries a specified number of times with a specified fixed time interval between retries
        ''' </summary>
        ''' <param name="numberOfRetries">The number of times to retry. Should be a non-negative number</param>
        ''' <param name="intervalBetweenRetries">The time interval between retries. Use TimeSpan.Zero to specify immediate
        ''' retries</param>
        ''' <returns></returns>
        ''' <remarks>When <paramref name="numberOfRetries"/> is 0 and <paramref name="intervalBetweenRetries"/> is
        ''' TimeSpan.Zero this policy is equivalent to the NoRetry policy</remarks>
        Public Shared Function RetryN(ByVal numberOfRetries As Integer, ByVal intervalBetweenRetries As TimeSpan) As RetryPolicy
            Return New RetryPolicy(Function(action) AnonymousMethod1(action, numberOfRetries, intervalBetweenRetries))
        End Function

        Private Shared Function AnonymousMethod1(ByVal action As Action, ByVal numberOfRetries As Integer, ByVal intervalBetweenRetries As TimeSpan) As Boolean
            RetryNImpl(action, numberOfRetries, intervalBetweenRetries)
            Return True
        End Function

        ''' <summary>
        ''' Policy that retries a specified number of times with a randomized exponential backoff scheme
        ''' </summary>
        ''' <param name="numberOfRetries">The number of times to retry. Should be a non-negative number.</param>
        ''' <param name="deltaBackoff">The multiplier in the exponential backoff scheme</param>
        ''' <returns></returns>
        ''' <remarks>For this retry policy, the minimum amount of milliseconds between retries is given by the 
        ''' StandardMinBackoff constant, and the maximum backoff is predefined by the StandardMaxBackoff constant. 
        ''' Otherwise, the backoff is calculated as random(2^currentRetry) * deltaBackoff.</remarks>
        Public Shared Function RetryExponentialN(ByVal numberOfRetries As Integer, ByVal deltaBackoff As TimeSpan) As RetryPolicy
            Return New RetryPolicy(Function(action) AnonymousMethod2(action, numberOfRetries, deltaBackoff))
        End Function

        Private Shared Function AnonymousMethod2(ByVal action As Action, ByVal numberOfRetries As Integer, ByVal deltaBackoff As TimeSpan) As Boolean
            RetryExponentialNImpl(action, numberOfRetries, StandardMinBackoff, StandardMaxBackoff, deltaBackoff)
            Return True
        End Function

        ''' <summary>
        ''' Policy that retries a specified number of times with a randomized exponential backoff scheme
        ''' </summary>
        ''' <param name="numberOfRetries">The number of times to retry. Should be a non-negative number</param>
        ''' <param name="deltaBackoff">The multiplier in the exponential backoff scheme</param>
        ''' <param name="minBackoff">The minimum backoff interval</param>
        ''' <param name="maxBackoff">The maximum backoff interval</param>
        ''' <returns></returns>
        ''' <remarks>For this retry policy, the minimum amount of milliseconds between retries is given by the 
        ''' minBackoff parameter, and the maximum backoff is predefined by the maxBackoff parameter. 
        ''' Otherwise, the backoff is calculated as random(2^currentRetry) * deltaBackoff.</remarks>
        Public Shared Function RetryExponentialN(ByVal numberOfRetries As Integer, ByVal minBackoff As TimeSpan, ByVal maxBackoff As TimeSpan, ByVal deltaBackoff As TimeSpan) As RetryPolicy
            If minBackoff > maxBackoff Then
                Throw New ArgumentException("The minimum backoff must not be larger than the maximum backoff period.")
            End If
            If minBackoff < TimeSpan.Zero Then
                Throw New ArgumentException("The minimum backoff period must not be negative.")
            End If

            Return New RetryPolicy(Function(action) AnonymousMethod3(action, numberOfRetries, minBackoff, maxBackoff, deltaBackoff))
        End Function

        Private Shared Function AnonymousMethod3(ByVal action As Action, ByVal numberOfRetries As Integer, ByVal minBackoff As TimeSpan, ByVal maxBackoff As TimeSpan, ByVal deltaBackoff As TimeSpan) As Boolean
            RetryExponentialNImpl(action, numberOfRetries, minBackoff, maxBackoff, deltaBackoff)
            Return True
        End Function

#Region "private helper methods"

        Private Shared Sub RetryNImpl(ByVal action As Action, ByVal numberOfRetries As Integer, ByVal intervalBetweenRetries As TimeSpan)
            Try
                action()
                Return
            Catch e1 As StorageServerException
                If numberOfRetries = 0 Then
                    Throw
                End If
                If intervalBetweenRetries > TimeSpan.Zero Then
                    Thread.Sleep(intervalBetweenRetries)
                End If
            Catch e As TableRetryWrapperException
                If numberOfRetries = 0 Then
                    Throw e.InnerException
                End If
                If intervalBetweenRetries > TimeSpan.Zero Then
                    Thread.Sleep(intervalBetweenRetries)
                End If
            End Try
            If (numberOfRetries > 0) Then Return
            numberOfRetries -= 1
        End Sub

        Private Shared Sub RetryExponentialNImpl(ByVal action As Action, ByVal numberOfRetries As Integer, ByVal minBackoff As TimeSpan, ByVal maxBackoff As TimeSpan, ByVal deltaBackoff As TimeSpan)
            Dim totalNumberOfRetries As Integer = numberOfRetries
            Dim backoff As TimeSpan

            ' sanity check
            ' this is already checked when creating the retry policy in case other than the standard settings are used
            ' because this library is available in source code, the standard settings can be changed and thus we 
            ' check again at this point
            If minBackoff > maxBackoff Then
                Throw New ArgumentException("The minimum backoff must not be larger than the maximum backoff period.")
            End If
            If minBackoff < TimeSpan.Zero Then
                Throw New ArgumentException("The minimum backoff period must not be negative.")
            End If

            Try
                action()
                Return
            Catch e1 As StorageServerException
                If numberOfRetries = 0 Then
                    Throw
                End If
                backoff = CalculateCurrentBackoff(minBackoff, maxBackoff, deltaBackoff, totalNumberOfRetries - numberOfRetries)
                Debug.Assert(backoff >= minBackoff)
                Debug.Assert(backoff <= maxBackoff)
                If backoff > TimeSpan.Zero Then
                    Thread.Sleep(backoff)
                End If
            Catch e As TableRetryWrapperException
                If numberOfRetries = 0 Then
                    Throw e.InnerException
                End If
                backoff = CalculateCurrentBackoff(minBackoff, maxBackoff, deltaBackoff, totalNumberOfRetries - numberOfRetries)
                Debug.Assert(backoff >= minBackoff)
                Debug.Assert(backoff <= maxBackoff)
                If backoff > TimeSpan.Zero Then
                    Thread.Sleep(backoff)
                End If
            End Try
            If (numberOfRetries > 0) Then Return
            numberOfRetries -= 1
        End Sub

        Private Shared Function CalculateCurrentBackoff(ByVal minBackoff As TimeSpan, ByVal maxBackoff As TimeSpan, ByVal deltaBackoff As TimeSpan, ByVal curRetry As Integer) As TimeSpan
            Dim backoff As Long

            If curRetry > 30 Then
                backoff = maxBackoff.Ticks
            Else
                Try
                    ' only randomize the multiplier here 
                    ' it would be as correct to randomize the whole backoff result
                    backoff = Random.Next((1 << curRetry) + 1)
                    ' Console.WriteLine("backoff:" + backoff);
                    ' Console.WriteLine("random range: [0, " + ((1 << curRetry) + 1) + "]");
                    backoff *= deltaBackoff.Ticks
                    backoff += minBackoff.Ticks
                Catch e1 As OverflowException
                    backoff = maxBackoff.Ticks
                End Try
                If backoff > maxBackoff.Ticks Then
                    backoff = maxBackoff.Ticks
                End If
            End If
            Debug.Assert(backoff >= minBackoff.Ticks)
            Debug.Assert(backoff <= maxBackoff.Ticks)
            Return TimeSpan.FromTicks(backoff)
        End Function

#End Region
    End Class

    ''' <summary>
    ''' Access control for containers
    ''' </summary>
    Public Enum ContainerAccessControl
        [Private]
        [Public]
    End Enum

    ''' <summary>
    ''' The blob container class.
    ''' Used to access and enumerate blobs in the container.
    ''' Storage key credentials are needed to access private blobs but not for public blobs.
    ''' </summary>
    Public MustInherit Class BlobContainer
        ''' <summary>
        ''' Use this constructor to access private blobs.
        ''' </summary>
        ''' <param name="baseUri">The base Uri for the storage endpoint</param>
        ''' <param name="accountName">Name of the storage account</param>
        ''' <param name="containerName">Name of the container</param>
        Protected Friend Sub New(ByVal baseUri As Uri, ByVal accountName As String, ByVal containerName As String)
            Me.New(baseUri, True, accountName, containerName, DateTime.MinValue)
        End Sub

        ''' <summary>
        ''' Use this constructor to access private blobs.
        ''' </summary>
        ''' <param name="baseUri">The base Uri for the storage endpoint</param>
        ''' <param name="usePathStyleUris">
        ''' If true, path-style URIs (http://baseuri/accountname/containername/objectname) are used and if false 
        ''' host-style URIs (http://accountname.baseuri/containername/objectname) are used, where baseuri is the 
        ''' URI of the service
        ''' </param>
        ''' <param name="accountName">Name of the storage account</param>
        ''' <param name="containerName">Name of the container</param>
        ''' <param name="lastModified">Date of last modification</param>
        Protected Friend Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris As Boolean, ByVal accountName As String, ByVal containerName As String, ByVal lastModified As DateTime)
            If (Not Utilities.IsValidContainerOrQueueName(containerName)) Then
                Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "The specified container name ""{0}"" is not valid!" & "Please choose a name that conforms to the naming conventions for containers!", containerName))
            End If
            Me.baseUri_Renamed = baseUri
            Me.usePathStyleUris_Renamed = usePathStyleUris
            Me.accountName_Renamed = accountName
            Me.containerName_Renamed = containerName
            Me.Timeout = BlobStorage.DefaultTimeout
            Me.RetryPolicy = BlobStorage.DefaultRetryPolicy
            Me.LastModifiedTime = lastModified
        End Sub


        ''' <summary>
        ''' The time out for each request to the storage service.
        ''' </summary>
        Private privateTimeout As TimeSpan
        Public Property Timeout() As TimeSpan
            Get
                Return privateTimeout
            End Get
            Set(ByVal value As TimeSpan)
                privateTimeout = value
            End Set
        End Property

        ''' <summary>
        ''' The retry policy used for retrying requests
        ''' </summary>
        Private privateRetryPolicy As RetryPolicy
        Public Property RetryPolicy() As RetryPolicy
            Get
                Return privateRetryPolicy
            End Get
            Set(ByVal value As RetryPolicy)
                privateRetryPolicy = value
            End Set
        End Property

        ''' <summary>
        ''' The base URI of the blob storage service
        ''' </summary>
        Public ReadOnly Property BaseUri() As Uri
            Get
                Return Me.baseUri_Renamed
            End Get
        End Property

        ''' <summary>
        ''' The name of the storage account
        ''' </summary>
        Public ReadOnly Property AccountName() As String
            Get
                Return Me.accountName_Renamed
            End Get
        End Property

        ''' <summary>
        ''' The name of the blob container.
        ''' </summary>
        Public ReadOnly Property ContainerName() As String
            Get
                Return Me.containerName_Renamed
            End Get
        End Property

        ''' <summary>
        ''' Indicates whether to use/generate path-style or host-style URIs
        ''' </summary>
        Public ReadOnly Property UsePathStyleUris() As Boolean
            Get
                Return Me.usePathStyleUris_Renamed
            End Get
        End Property

        ''' <summary>
        ''' The URI of the container
        ''' </summary>
        Public MustOverride ReadOnly Property ContainerUri() As Uri

        ''' <summary>
        ''' The timestamp for last modification of container.
        ''' </summary>
        Private privateLastModifiedTime As DateTime
        Public Property LastModifiedTime() As DateTime
            Get
                Return privateLastModifiedTime
            End Get
            Protected Set(ByVal value As DateTime)
                privateLastModifiedTime = value
            End Set
        End Property

        ''' <summary>
        ''' Create the container if it does not exist.
        ''' The container is created with private access control and no metadata.
        ''' </summary>
        ''' <returns>true if the container was created. false if the container already exists</returns>
        Public MustOverride Function CreateContainer() As Boolean

        ''' <summary>
        ''' Create the container with the specified metadata and access control if it does not exist
        ''' </summary>
        ''' <param name="metadata">The metadata for the container. Can be null to indicate no metadata</param>
        ''' <param name="accessControl">The access control (public or private) with which to create the container</param>
        ''' <returns>true if the container was created. false if the container already exists</returns>
        Public MustOverride Function CreateContainer(ByVal metadata As NameValueCollection, ByVal accessControl As ContainerAccessControl) As Boolean

        ''' <summary>
        ''' Check if the blob container exists
        ''' </summary>
        ''' <returns>true if the container exists, false otherwise.</returns>
        Public MustOverride Function DoesContainerExist() As Boolean

        ''' <summary>
        ''' Get the properties for the container if it exists.
        ''' </summary>
        ''' <returns>The properties for the container if it exists, null otherwise</returns>
        <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate", Justification:="The method makes a call to the blob service")> _
        Public MustOverride Function GetContainerProperties() As ContainerProperties

        ''' <summary>
        ''' Get the access control permissions associated with the container.
        ''' </summary>
        ''' <returns></returns>
        <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate", Justification:="The method makes a call to the blob service")> _
        Public MustOverride Function GetContainerAccessControl() As ContainerAccessControl

        ''' <summary>
        ''' Set the access control permissions associated with the container.
        ''' </summary>
        ''' <param name="acl">The permission to set</param>
        <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate", Justification:="The method makes a call to the blob service")> _
        Public MustOverride Sub SetContainerAccessControl(ByVal acl As ContainerAccessControl)

        ''' <summary>
        ''' Deletes the current container.
        ''' </summary>
        Public MustOverride Function DeleteContainer() As Boolean

        ''' <summary>
        ''' Check if the blob container exists
        ''' </summary>
        ''' <param name="blobName">Name of the BLOB.</param>
        ''' <returns>true if the blob exists, false otherwise.</returns>
        Public MustOverride Function DoesBlobExist(ByVal blobName As String) As Boolean

        ''' <summary>
        ''' Create a new blob or overwrite an existing blob.
        ''' </summary>
        ''' <param name="blobProperties">The properties of the blob</param>
        ''' <param name="blobContents">The contents of the blob</param>
        ''' <param name="overwrite">Should this request overwrite an existing blob ?</param>
        ''' <returns>true if the blob was created. false if the blob already exists and <paramref name="overwrite"/>was set to false</returns>
        ''' <remarks>The LastModifiedTime property of <paramref name="blobProperties"/> is set as a result of this call.
        ''' This method also has an effect on the ETag values that are managed by the service.</remarks>
        Public MustOverride Function CreateBlob(ByVal blobProperties As BlobProperties, ByVal blobContents As BlobContents, ByVal overwrite As Boolean) As Boolean

        ''' <summary>
        ''' Updates an existing blob if it has not been modified since the specified time which is typically
        ''' the last modified time of the blob when you retrieved it.
        ''' Use this method to implement optimistic concurrency by avoiding clobbering changes to the blob
        ''' made by another writer.
        ''' </summary>
        ''' <param name="blob">The properties of the blob. This object should be one previously
        ''' obtained from a call to GetBlob or GetBlobProperties and have its LastModifiedTime property set.</param>
        ''' <param name="contents">The contents of the blob. The contents of the blob should be readable</param>
        ''' <returns>true if the blob was updated. false if the blob has changed since the last time</returns>
        ''' <remarks>The LastModifiedTime property of <paramref name="blob"/> is set as a result of this call.
        ''' This method also has an effect on the ETag values that are managed by the service if the update was 
        ''' successful.</remarks>
        Public MustOverride Function UpdateBlobIfNotModified(ByVal blob As BlobProperties, ByVal contents As BlobContents) As Boolean

        ''' <summary>
        ''' Get the blob contents and properties if the blob exists
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <param name="blobContents">Object in which the contents are returned.
        ''' This object should contain a writable stream or should be a default constructed object.</param>
        ''' <param name="transferAsChunks">Should the blob be gotten in pieces. This requires more round-trips, but will retry smaller pieces in case of failure.</param>
        ''' <returns>The properties of the blob if the blob exists.</returns>
        Public MustOverride Function GetBlob(ByVal name As String, ByVal blobContents As BlobContents, ByVal transferAsChunks As Boolean) As BlobProperties

        ''' <summary>
        ''' Gets the blob contents and properties if the blob has not been modified since the time specified.
        ''' Use this method if you have cached the contents of a blob and want to avoid retrieving the blob
        ''' if it has not changed since the last time you retrieved it.
        ''' </summary>
        ''' <param name="blobProperties">The properties of the blob obtained from an earlier call to GetBlob. This
        ''' parameter is updated by the call if the blob has been modified</param>
        ''' <param name="blobContents">Contains the stream to which the contents of the blob are written if it has been
        ''' modified</param>
        ''' <param name="transferAsChunks">Should the blob be gotten in pieces. This requires more round-trips, but will retry smaller pieces in case of failure.</param>
        ''' <returns>true if the blob has been modified, false otherwise</returns>
        Public MustOverride Function GetBlobIfModified(ByVal blobProperties As BlobProperties, ByVal blobContents As BlobContents, ByVal transferAsChunks As Boolean) As Boolean

        ''' <summary>
        ''' Get the properties of the blob if it exists.
        ''' This method is also the simplest way to check if a blob exists.
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <returns>The properties of the blob if it exists. null otherwise.
        ''' The properties for the contents of the blob are not set</returns>
        Public MustOverride Function GetBlobProperties(ByVal name As String) As BlobProperties

        ''' <summary>
        ''' Set the metadata of an existing blob.
        ''' </summary>
        ''' <param name="blobProperties">The blob properties object whose metadata is to be updated</param>
        Public MustOverride Sub UpdateBlobMetadata(ByVal blobProperties As BlobProperties)

        ''' <summary>
        ''' Set the metadata of an existing blob if it has not been modified since it was last retrieved.
        ''' </summary>
        ''' <param name="blobProperties">The blob properties object whose metadata is to be updated.
        ''' Typically obtained by a previous call to GetBlob or GetBlobProperties</param>
        ''' <returns>true if the blob metadata was updated. false if it was not updated because the blob
        ''' has been modified</returns>
        Public MustOverride Function UpdateBlobMetadataIfNotModified(ByVal blobProperties As BlobProperties) As Boolean

        ''' <summary>
        ''' Delete a blob with the given name
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <returns>true if the blob exists and was successfully deleted, false if the blob does not exist</returns>
        Public MustOverride Function DeleteBlob(ByVal name As String) As Boolean

        ''' <summary>
        ''' Delete a blob with the given name if the blob has not been modified since it was last obtained.
        ''' Use this method for optimistic concurrency to avoid deleting a blob that has been modified since
        ''' the last time you retrieved it
        ''' </summary>
        ''' <param name="blob">A blob object (typically previously obtained from a GetBlob call)</param>
        ''' <param name="modified">This out parameter is set to true if the blob was not deleted because
        ''' it was modified</param>
        ''' <returns>true if the blob exists and was successfully deleted, false if the blob does not exist or was
        ''' not deleted because the blob was modified.</returns>
        Public MustOverride Function DeleteBlobIfNotModified(ByVal blob As BlobProperties, <System.Runtime.InteropServices.Out()> ByRef modified As Boolean) As Boolean

        ''' <summary>
        ''' Enumerates all blobs with a given prefix.
        ''' </summary>
        ''' <param name="prefix"></param>
        ''' <param name="combineCommonPrefixes">If true common prefixes with "/" as seperator</param>
        ''' <returns>The list of blob properties and common prefixes</returns>
        Public MustOverride Function ListBlobs(ByVal prefix As String, ByVal combineCommonPrefixes As Boolean) As IEnumerable(Of Object)

        Private baseUri_Renamed As Uri
        Private accountName_Renamed As String
        Private containerName_Renamed As String
        Private usePathStyleUris_Renamed As Boolean
    End Class

    ''' <summary>
    ''' The properties of a blob.
    ''' No member of this class makes a storage service request.
    ''' </summary>
    Public Class BlobProperties
        ''' <summary>
        ''' Construct a new BlobProperties object
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        Public Sub New(ByVal name As String)
            name = name
        End Sub


        ''' <summary>
        ''' Name of the blob
        ''' </summary>
        Private privateName As String
        Public Property Name() As String
            Get
                Return privateName
            End Get
            Friend Set(ByVal value As String)
                privateName = value
            End Set
        End Property

        ''' <summary>
        ''' URI of the blob
        ''' </summary>
        Private privateUri As Uri
        Public Property Uri() As Uri
            Get
                Return privateUri
            End Get
            Friend Set(ByVal value As Uri)
                privateUri = value
            End Set
        End Property

        ''' <summary>
        ''' Content encoding of the blob if it set, null otherwise.
        ''' </summary>
        Private privateContentEncoding As String
        Public Property ContentEncoding() As String
            Get
                Return privateContentEncoding
            End Get
            Set(ByVal value As String)
                privateContentEncoding = value
            End Set
        End Property

        ''' <summary>
        ''' Content Type of the blob if it is set, null otherwise.
        ''' </summary>
        Private privateContentType As String
        Public Property ContentType() As String
            Get
                Return privateContentType
            End Get
            Set(ByVal value As String)
                privateContentType = value
            End Set
        End Property

        ''' <summary>
        ''' Content Language of the blob if it is set, null otherwise.
        ''' </summary>
        Private privateContentLanguage As String
        Public Property ContentLanguage() As String
            Get
                Return privateContentLanguage
            End Get
            Set(ByVal value As String)
                privateContentLanguage = value
            End Set
        End Property

        ''' <summary>
        ''' The length of the blob content, null otherwise.
        ''' </summary>
        Private privateContentLength As Long
        Public Property ContentLength() As Long
            Get
                Return privateContentLength
            End Get
            Friend Set(ByVal value As Long)
                privateContentLength = value
            End Set
        End Property

        ''' <summary>
        ''' Metadata for the blob in the form of name-value pairs.
        ''' </summary>
        Private privateMetadata As NameValueCollection
        Public Property Metadata() As NameValueCollection
            Get
                Return privateMetadata
            End Get
            Set(ByVal value As NameValueCollection)
                privateMetadata = value
            End Set
        End Property

        ''' <summary>
        ''' The last modified time for the blob. 
        ''' </summary>
        Private privateLastModifiedTime As DateTime
        Public Property LastModifiedTime() As DateTime
            Get
                Return privateLastModifiedTime
            End Get
            Friend Set(ByVal value As DateTime)
                privateLastModifiedTime = value
            End Set
        End Property

        ''' <summary>
        ''' The ETag of the blob. This is an identifier assigned to the blob by the storage service
        ''' and is used to distinguish contents of two blobs (or versions of the same blob).
        ''' </summary>
        Private privateETag As String
        Public Property ETag() As String
            Get
                Return privateETag
            End Get
            Friend Set(ByVal value As String)
                privateETag = value
            End Set
        End Property

        Friend Sub Assign(ByVal other As BlobProperties)
            Name = other.Name
            Uri = other.Uri
            ContentEncoding = other.ContentEncoding
            ContentLanguage = other.ContentLanguage
            ContentLength = other.ContentLength
            ContentType = other.ContentType
            ETag = other.ETag
            LastModifiedTime = other.LastModifiedTime
            Metadata = (If(other.Metadata IsNot Nothing, New NameValueCollection(other.Metadata), Nothing))
        End Sub
    End Class

    ''' <summary>
    ''' The properties of a container.
    ''' No member of this class makes a storage service request.
    ''' </summary>
    Public Class ContainerProperties
        Public Sub New(ByVal name As String)
            name = name
        End Sub

        Private privateName As String
        Public Property Name() As String
            Get
                Return privateName
            End Get
            Friend Set(ByVal value As String)
                privateName = value
            End Set
        End Property
        Private privateLastModifiedTime As DateTime
        Public Property LastModifiedTime() As DateTime
            Get
                Return privateLastModifiedTime
            End Get
            Friend Set(ByVal value As DateTime)
                privateLastModifiedTime = value
            End Set
        End Property

        Private privateETag As String
        Public Property ETag() As String
            Get
                Return privateETag
            End Get
            Friend Set(ByVal value As String)
                privateETag = value
            End Set
        End Property

        Private privateUri As Uri
        Public Property Uri() As Uri
            Get
                Return privateUri
            End Get
            Friend Set(ByVal value As Uri)
                privateUri = value
            End Set
        End Property

        Private privateMetadata As NameValueCollection
        Public Property Metadata() As NameValueCollection
            Get
                Return privateMetadata
            End Get
            Friend Set(ByVal value As NameValueCollection)
                privateMetadata = value
            End Set
        End Property
    End Class


    ''' <summary>
    ''' The contents of the Blob in various forms.
    ''' </summary>
    Public Class BlobContents
        ''' <summary>
        ''' Construct a new BlobContents object from a stream.
        ''' </summary>
        ''' <param name="stream">The stream to/from which blob contents are written/read. The
        ''' stream should be seekable in order for requests to be retried.</param>
        Public Sub New(ByVal stream As Stream)
            Me.stream = stream
        End Sub

        ''' <summary>
        ''' Construct a new BlobContents object from a byte array.
        ''' </summary>
        ''' <param name="value">The byte array to/from which contents are written/read.</param>
        Public Sub New(ByVal value() As Byte)
            Me.bytes = value
            Me.stream = New MemoryStream(value, False)
        End Sub

        ''' <summary>
        ''' Get the contents of a blob as a stream.
        ''' </summary>
        Public ReadOnly Property AsStream() As Stream
            Get
                Return stream
            End Get

        End Property

        ''' <summary>
        ''' Get the contents of a blob as a byte array.
        ''' </summary>
        Public Function AsBytes() As Byte()
            If bytes IsNot Nothing Then
                Return bytes
            End If
            If stream IsNot Nothing Then
                stream.Seek(0, SeekOrigin.Begin)
                bytes = New Byte(stream.Length - 1) {}
                Dim n As Integer = 0
                Dim offset As Integer = 0
                Do
                    n = stream.Read(bytes, offset, bytes.Length - offset)
                    offset += n

                Loop While n > 0
            End If
            Return bytes
        End Function

        Private stream As Stream
        Private bytes() As Byte
    End Class
#End Region
End Namespace
