'
' <copyright file="Errors.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.IO
Imports System.Collections.Generic
Imports System.Net
Imports System.Collections.Specialized
Imports System.Runtime.Serialization
Imports System.Security.Permissions
Imports System.Diagnostics.CodeAnalysis
Imports System.Web
Imports System.Linq
Imports System.Xml
Imports System.Xml.Linq


' disable the generation of warnings for missing documentation elements for 
' public classes/members in this file
' justification is that this file contains many public constants whose names 
' sufficiently reflect their intended usage 

Namespace Microsoft.Samples.ServiceHosting.StorageClient
	''' <summary>
	''' Error codes that can be returned by the storage service or the client library.
	''' These are divided into server errors and client errors depending on which side
	''' the error can be attributed to.
	''' </summary>
	Public Enum StorageErrorCode
		None = 0

		'Server errors
		ServiceInternalError = 1
		ServiceTimeout
		ServiceIntegrityCheckFailed
		TransportError
		ServiceBadResponse

		'Client errors
		ResourceNotFound
		AccountNotFound
		ContainerNotFound
		BlobNotFound
		AuthenticationFailure
		AccessDenied
		ResourceAlreadyExists
		ContainerAlreadyExists
		BlobAlreadyExists
		BadRequest
		ConditionFailed
		BadGateway
	End Enum

	<Serializable> _
	Public Class StorageExtendedErrorInformation
		Private privateErrorCode As String
		Public Property ErrorCode() As String
			Get
				Return privateErrorCode
			End Get
			Friend Set(ByVal value As String)
				privateErrorCode = value
			End Set
		End Property
		Private privateErrorMessage As String
		Public Property ErrorMessage() As String
			Get
				Return privateErrorMessage
			End Get
			Friend Set(ByVal value As String)
				privateErrorMessage = value
			End Set
		End Property
		Private privateAdditionalDetails As NameValueCollection
		Public Property AdditionalDetails() As NameValueCollection
			Get
				Return privateAdditionalDetails
			End Get
			Friend Set(ByVal value As NameValueCollection)
				privateAdditionalDetails = value
			End Set
		End Property
	End Class

	''' <summary>
	''' The base class for storage service exceptions
	''' </summary>
	<Serializable> _
	Public MustInherit Class StorageException
		Inherits Exception
		''' <summary>
		''' The Http status code returned by the storage service
		''' </summary>
		Private privateStatusCode As HttpStatusCode
		Public Property StatusCode() As HttpStatusCode
			Get
				Return privateStatusCode
			End Get
			Private Set(ByVal value As HttpStatusCode)
				privateStatusCode = value
			End Set
		End Property

		''' <summary>
		''' The specific error code returned by the storage service
		''' </summary>
		Private privateErrorCode As StorageErrorCode
		Public Property ErrorCode() As StorageErrorCode
			Get
				Return privateErrorCode
			End Get
			Private Set(ByVal value As StorageErrorCode)
				privateErrorCode = value
			End Set
		End Property

		''' <summary>
		''' 
		''' </summary>
		Private privateExtendedErrorInformation As StorageExtendedErrorInformation
		Public Property ExtendedErrorInformation() As StorageExtendedErrorInformation
			Get
				Return privateExtendedErrorInformation
			End Get
			Private Set(ByVal value As StorageExtendedErrorInformation)
				privateExtendedErrorInformation = value
			End Set
		End Property

		Protected Sub New()
		End Sub

		Protected Sub New(ByVal errorCode As StorageErrorCode, ByVal message As String, ByVal statusCode As HttpStatusCode, ByVal extendedErrorInfo As StorageExtendedErrorInformation, ByVal innerException As Exception)
			MyBase.New(message, innerException)
			Me.ErrorCode = errorCode
			Me.StatusCode = statusCode
			Me.ExtendedErrorInformation = extendedErrorInfo
		End Sub

		''' <summary>
		''' Initializes a new instance of the <see cref="StorageException"/> class with
		''' serialized data.
		''' </summary>
		''' <param name="info">The <see cref="SerializationInfo"/> object that contains serialized object
		''' data about the exception being thrown</param>
		''' <param name="context">The <see cref="StreamingContext"/> object that contains contextual information
		''' about the source or destionation. </param>
		Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
			MyBase.New(info, context)
			If Nothing Is info Then
				Throw New ArgumentNullException("info")
			End If

			Me.StatusCode = CType(info.GetValue("StatusCode", GetType(HttpStatusCode)), HttpStatusCode)
			Me.ErrorCode = CType(info.GetValue("ErrorCode", GetType(StorageErrorCode)), StorageErrorCode)
			Me.ExtendedErrorInformation = CType(info.GetValue("ExtendedErrorInformation", GetType(StorageExtendedErrorInformation)), StorageExtendedErrorInformation)
		End Sub

		''' <summary>
		''' Sets the <see cref="SerializationInfo"/> object with additional exception information
		''' </summary>
		''' <param name="info">The <see cref="SerializationInfo"/> object that holds the 
		''' serialized object data.</param>
		''' <param name="context">The <see cref="StreamingContext"/> object that contains contextual information
		''' about the source or destionation. </param>
		<SecurityPermissionAttribute(SecurityAction.Demand, SerializationFormatter := True)> _
		Public Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
			If Nothing Is info Then
				Throw New ArgumentNullException("info")
			End If

			info.AddValue("StatusCode", Me.StatusCode)
			info.AddValue("ErrorCode", Me.ErrorCode)
			info.AddValue("ExtendedErrorInformation", Me.ExtendedErrorInformation)
			MyBase.GetObjectData(info, context)
		End Sub

	End Class

	''' <summary>
	''' Server exceptions are those due to server side problems.
	''' These may be transient and requests resulting in such exceptions
	''' can be retried with the same parameters.
	''' </summary>
	<SuppressMessage("Microsoft.Design", "CA1032:ImplementStandardExceptionConstructors", Justification := "Since this exception comes from the server, there must be an HTTP response code associated with it, hence we exclude the default constructor taking only a string but no status code."), Serializable> _
	Public Class StorageServerException
		Inherits StorageException
		Friend Sub New(ByVal errorCode As StorageErrorCode, ByVal message As String, ByVal statusCode As HttpStatusCode, ByVal innerException As Exception)
			MyBase.New(errorCode, message, statusCode, Nothing, innerException)
		End Sub

		Friend Sub New(ByVal errorCode As StorageErrorCode, ByVal message As String, ByVal statusCode As HttpStatusCode, ByVal extendedErrorInfo As StorageExtendedErrorInformation, ByVal innerException As Exception)
			MyBase.New(errorCode, message, statusCode, extendedErrorInfo, innerException)
		End Sub


		''' <summary>
		''' Initializes a new instance of the <see cref="StorageServerException"/> class with
		''' serialized data.
		''' </summary>
		''' <param name="info">The <see cref="SerializationInfo"/> object that contains serialized object
		''' data about the exception being thrown</param>
		''' <param name="context">The <see cref="StreamingContext"/> object that contains contextual information
		''' about the source or destionation. </param>
		Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
			MyBase.New(info, context)
		End Sub

		Public Sub New()
		End Sub
	End Class

	''' <summary>
	''' Client side exceptions are due to incorrect parameters to the request.
	''' These requests should not be retried with the same parameters
	''' </summary>
	<SuppressMessage("Microsoft.Design", "CA1032:ImplementStandardExceptionConstructors", Justification := "Since this exception comes from the server, there must be an HTTP response code associated with it, hence we exclude the default constructor taking only a string but no status code."), Serializable> _
	Public Class StorageClientException
		Inherits StorageException
		Friend Sub New(ByVal errorCode As StorageErrorCode, ByVal message As String, ByVal statusCode As HttpStatusCode, ByVal extendedErrorInfo As StorageExtendedErrorInformation, ByVal innerException As Exception)
			MyBase.New(errorCode, message, statusCode, extendedErrorInfo, innerException)
		End Sub


		''' <summary>
		''' Initializes a new instance of the <see cref="StorageClientException"/> class with
		''' serialized data.
		''' </summary>
		''' <param name="info">The <see cref="SerializationInfo"/> object that contains serialized object
		''' data about the exception being thrown</param>
		''' <param name="context">The <see cref="StreamingContext"/> object that contains contextual information
		''' about the source or destionation. </param>
		Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
			MyBase.New(info, context)
		End Sub

		Public Sub New()
		End Sub

	End Class

	#Region "Error code strings that can be returned in the StorageExtendedErrorInformation.ErrorCode"
	''' <summary>
	''' Error code strings that are common to all storage services
	''' </summary>
	Public NotInheritable Class StorageErrorCodeStrings
		Public Const UnsupportedHttpVerb As String = "UnsupportedHttpVerb"
		Public Const MissingContentLengthHeader As String = "MissingContentLengthHeader"
		Public Const MissingRequiredHeader As String = "MissingRequiredHeader"
		Public Const MissingRequiredXmlNode As String = "MissingRequiredXmlNode"
		Public Const UnsupportedHeader As String = "UnsupportedHeader"
		Public Const UnsupportedXmlNode As String = "UnsupportedXmlNode"
		Public Const InvalidHeaderValue As String = "InvalidHeaderValue"
		Public Const InvalidXmlNodeValue As String = "InvalidXmlNodeValue"
		Public Const MissingRequiredQueryParameter As String = "MissingRequiredQueryParameter"
		Public Const UnsupportedQueryParameter As String = "UnsupportedQueryParameter"
		Public Const InvalidQueryParameterValue As String = "InvalidQueryParameterValue"
		Public Const OutOfRangeQueryParameterValue As String = "OutOfRangeQueryParameterValue"
		Public Const InvalidUri As String = "InvalidUri"
		Public Const InvalidHttpVerb As String = "InvalidHttpVerb"
		Public Const EmptyMetadataKey As String = "EmptyMetadataKey"
		Public Const RequestBodyTooLarge As String = "RequestBodyTooLarge"
		Public Const InvalidXmlDocument As String = "InvalidXmlDocument"
		Public Const InternalError As String = "InternalError"
		Public Const AuthenticationFailed As String = "AuthenticationFailed"
		Public Const Md5Mismatch As String = "Md5Mismatch"
		Public Const InvalidMd5 As String = "InvalidMd5"
		Public Const OutOfRangeInput As String = "OutOfRangeInput"
		Public Const InvalidInput As String = "InvalidInput"
		Public Const OperationTimedOut As String = "OperationTimedOut"
		Public Const ResourceNotFound As String = "ResourceNotFound"
		Public Const InvalidMetadata As String = "InvalidMetadata"
		Public Const MetadataTooLarge As String = "MetadataTooLarge"
		Public Const ConditionNotMet As String = "ConditionNotMet"
		Public Const InvalidRange As String = "InvalidRange"
		Public Const ContainerNotFound As String = "ContainerNotFound"
		Public Const ContainerAlreadyExists As String = "ContainerAlreadyExists"
		Public Const ContainerDisabled As String = "ContainerDisabled"
		Public Const ContainerBeingDeleted As String = "ContainerBeingDeleted"
		Public Const ServerBusy As String = "ServerBusy"
	End Class

	''' <summary>
	''' Error code strings that are specific to blob service
	''' </summary>
	Public NotInheritable Class BlobErrorCodeStrings
		Public Const InvalidBlockId As String = "InvalidBlockId"
		Public Const BlobNotFound As String = "BlobNotFound"
		Public Const BlobAlreadyExists As String = "BlobAlreadyExists"
		Public Const InvalidBlobOrBlock As String = "InvalidBlobOrBlock"
		Public Const InvalidBlockList As String = "InvalidBlockList"
	End Class

	''' <summary>
	''' Error code strings that are specific to queue service
	''' </summary>
	Public NotInheritable Class QueueErrorCodeStrings
		Public Const QueueNotFound As String = "QueueNotFound"
		Public Const QueueDisabled As String = "QueueDisabled"
		Public Const QueueAlreadyExists As String = "QueueAlreadyExists"
		Public Const QueueNotEmpty As String = "QueueNotEmpty"
		Public Const QueueBeingDeleted As String = "QueueBeingDeleted"
		Public Const PopReceiptMismatch As String = "PopReceiptMismatch"
		Public Const InvalidParameter As String = "InvalidParameter"
		Public Const MessageNotFound As String = "MessageNotFound"
		Public Const MessageTooLarge As String = "MessageTooLarge"
		Public Const InvalidMarker As String = "InvalidMarker"
	End Class

	''' <summary>
	''' Error code strings that are specific to queue service
	''' </summary>
	'''     public static class TableErrorCodeStrings
	Public NotInheritable Class TableErrorCodeStrings
		Public Const XMethodNotUsingPost As String = "XMethodNotUsingPost"
		Public Const XMethodIncorrectValue As String = "XMethodIncorrectValue"
		Public Const XMethodIncorrectCount As String = "XMethodIncorrectCount"

		Public Const TableHasNoProperties As String = "TableHasNoProperties"
		Public Const DuplicatePropertiesSpecified As String = "DuplicatePropertiesSpecified"
		Public Const TableHasNoSuchProperty As String = "TableHasNoSuchProperty"
		Public Const DuplicateKeyPropertySpecified As String = "DuplicateKeyPropertySpecified"
		Public Const TableAlreadyExists As String = "TableAlreadyExists"
		Public Const TableNotFound As String = "TableNotFound"
		Public Const EntityNotFound As String = "EntityNotFound"
		Public Const EntityAlreadyExists As String = "EntityAlreadyExists"
		Public Const PartitionKeyNotSpecified As String = "PartitionKeyNotSpecified"
		Public Const OperatorInvalid As String = "OperatorInvalid"
		Public Const UpdateConditionNotSatisfied As String = "UpdateConditionNotSatisfied"
		Public Const PropertiesNeedValue As String = "PropertiesNeedValue"

		Public Const PartitionKeyPropertyCannotBeUpdated As String = "PartitionKeyPropertyCannotBeUpdated"
		Public Const TooManyProperties As String = "TooManyProperties"
		Public Const EntityTooLarge As String = "EntityTooLarge"
		Public Const PropertyValueTooLarge As String = "PropertyValueTooLarge"
		Public Const InvalidValueType As String = "InvalidValueType"
		Public Const TableBeingDeleted As String = "TableBeingDeleted"
		Public Const TableServerOutOfMemory As String = "TableServerOutOfMemory"
		Public Const PrimaryKeyPropertyIsInvalidType As String = "PrimaryKeyPropertyIsInvalidType"
		Public Const PropertyNameTooLong As String = "PropertyNameTooLong"
		Public Const PropertyNameInvalid As String = "PropertyNameInvalid"

		Public Const BatchOperationNotSupported As String = "BatchOperationNotSupported"
		Public Const JsonFormatNotSupported As String = "JsonFormatNotSupported"
		Public Const MethodNotAllowed As String = "MethodNotAllowed"
		Public Const NotImplemented As String = "NotImplemented"
	End Class
	#End Region

	#Region "Helper functions dealing with errors"
	Friend Partial NotInheritable Class Utilities
		Private Sub New()
		End Sub
		Friend Shared Sub ProcessUnexpectedStatusCode(ByVal response As HttpWebResponse)
			Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, response.StatusDescription, response.StatusCode, Nothing)
		End Sub

		Friend Shared Function TranslateWebException(ByVal e As Exception) As Exception
			Dim we As WebException = TryCast(e, WebException)
			If Nothing Is we Then
				Return e
			End If

			' If the response is not null, let's first see what the status code is.
			If we.Response IsNot Nothing Then
				Dim response As HttpWebResponse = (CType(we.Response, HttpWebResponse))

				Dim extendedError As StorageExtendedErrorInformation = GetExtendedErrorDetailsFromResponse(response.GetResponseStream(), response.ContentLength)
				Dim translatedException As Exception = Nothing
				If extendedError IsNot Nothing Then
					translatedException = TranslateExtendedError(extendedError, response.StatusCode, response.StatusDescription, e)
					If translatedException IsNot Nothing Then
						Return translatedException
					End If
				End If
				translatedException = TranslateFromHttpStatus(response.StatusCode, response.StatusDescription, extendedError, we)
				If translatedException IsNot Nothing Then
					Return translatedException
				End If

			End If

			Select Case we.Status
				Case WebExceptionStatus.RequestCanceled
					Return New StorageServerException(StorageErrorCode.ServiceTimeout, "The server request did not complete within the specified timeout", HttpStatusCode.GatewayTimeout, we)

				Case WebExceptionStatus.ConnectFailure
					Return we

				Case Else
					Return New StorageServerException(StorageErrorCode.ServiceInternalError, "The server encountered an unknown failure: " & e.Message, HttpStatusCode.InternalServerError, we)
			End Select
		End Function

		Friend Shared Function TranslateFromHttpStatus(ByVal statusCode As HttpStatusCode, ByVal statusDescription As String, ByVal details As StorageExtendedErrorInformation, ByVal inner As Exception) As Exception
			Select Case statusCode
				Case HttpStatusCode.Forbidden
					Return New StorageClientException(StorageErrorCode.AccessDenied, statusDescription, HttpStatusCode.Forbidden, details, inner)

				Case HttpStatusCode.Gone, HttpStatusCode.NotFound
					Return New StorageClientException(StorageErrorCode.ResourceNotFound, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.BadRequest
					Return New StorageClientException(StorageErrorCode.BadRequest, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.PreconditionFailed, HttpStatusCode.NotModified
					Return New StorageClientException(StorageErrorCode.BadRequest, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.Conflict
					Return New StorageClientException(StorageErrorCode.ResourceAlreadyExists, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.GatewayTimeout
					Return New StorageServerException(StorageErrorCode.ServiceTimeout, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.RequestedRangeNotSatisfiable
					Return New StorageClientException(StorageErrorCode.BadRequest, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.InternalServerError
					Return New StorageServerException(StorageErrorCode.ServiceInternalError, statusDescription, statusCode, details, inner)

				Case HttpStatusCode.BadGateway
					Return New StorageServerException(StorageErrorCode.BadGateway, statusDescription, statusCode, details, inner)
			End Select
			Return Nothing
		End Function

		Private Shared Function TranslateExtendedError(ByVal details As StorageExtendedErrorInformation, ByVal statusCode As HttpStatusCode, ByVal statusDescription As String, ByVal inner As Exception) As Exception
			Dim errorCode As StorageErrorCode = Nothing
			Select Case details.ErrorCode
				Case StorageErrorCodeStrings.UnsupportedHttpVerb, StorageErrorCodeStrings.MissingContentLengthHeader, StorageErrorCodeStrings.MissingRequiredHeader, StorageErrorCodeStrings.UnsupportedHeader, StorageErrorCodeStrings.InvalidHeaderValue, StorageErrorCodeStrings.MissingRequiredQueryParameter, StorageErrorCodeStrings.UnsupportedQueryParameter, StorageErrorCodeStrings.InvalidQueryParameterValue, StorageErrorCodeStrings.OutOfRangeQueryParameterValue, StorageErrorCodeStrings.InvalidUri, StorageErrorCodeStrings.InvalidHttpVerb, StorageErrorCodeStrings.EmptyMetadataKey, StorageErrorCodeStrings.RequestBodyTooLarge, StorageErrorCodeStrings.InvalidXmlDocument, StorageErrorCodeStrings.InvalidXmlNodeValue, StorageErrorCodeStrings.MissingRequiredXmlNode, StorageErrorCodeStrings.InvalidMd5, StorageErrorCodeStrings.OutOfRangeInput, StorageErrorCodeStrings.InvalidInput, StorageErrorCodeStrings.InvalidMetadata, StorageErrorCodeStrings.MetadataTooLarge, StorageErrorCodeStrings.InvalidRange
					errorCode = StorageErrorCode.BadRequest
				Case StorageErrorCodeStrings.AuthenticationFailed
					errorCode = StorageErrorCode.AuthenticationFailure
				Case StorageErrorCodeStrings.ResourceNotFound
					errorCode = StorageErrorCode.ResourceNotFound
				Case StorageErrorCodeStrings.ConditionNotMet
					errorCode = StorageErrorCode.ConditionFailed
				Case StorageErrorCodeStrings.ContainerAlreadyExists
					errorCode = StorageErrorCode.ContainerAlreadyExists
				Case StorageErrorCodeStrings.ContainerNotFound
					errorCode = StorageErrorCode.ContainerNotFound
				Case BlobErrorCodeStrings.BlobNotFound
					errorCode = StorageErrorCode.BlobNotFound
				Case BlobErrorCodeStrings.BlobAlreadyExists
					errorCode = StorageErrorCode.BlobAlreadyExists
			End Select

            If Not errorCode <> StorageErrorCode.None Then
                Return New StorageClientException(errorCode, statusDescription, statusCode, details, inner)
            End If

            Select Case details.ErrorCode
                Case StorageErrorCodeStrings.InternalError, StorageErrorCodeStrings.ServerBusy
                    errorCode = StorageErrorCode.ServiceInternalError
                Case StorageErrorCodeStrings.Md5Mismatch
                    errorCode = StorageErrorCode.ServiceIntegrityCheckFailed
                Case StorageErrorCodeStrings.OperationTimedOut
                    errorCode = StorageErrorCode.ServiceTimeout
            End Select
            If errorCode <> StorageErrorCode.None Then
                Return New StorageServerException(errorCode, statusDescription, statusCode, details, inner)
            End If

            Return Nothing
		End Function


		' This is the limit where we allow for the error message returned by the server.
		' Message longer than that will be truncated. 
		Private Const ErrorTextSizeLimit As Integer = 8 * 1024

		Private Shared Function GetExtendedErrorDetailsFromResponse(ByVal httpResponseStream As Stream, ByVal contentLength As Long) As StorageExtendedErrorInformation
			Try
				Dim bytesToRead As Integer = CInt(Fix(Math.Max(CLng(Fix(contentLength)), CLng(Fix(ErrorTextSizeLimit)))))
				Dim responseBuffer(bytesToRead - 1) As Byte
				Dim bytesRead As Integer = CopyStreamToBuffer(httpResponseStream, responseBuffer, CInt(Fix(bytesToRead)))
				Return GetErrorDetailsFromStream(New MemoryStream(responseBuffer, 0, bytesRead, False))
			Catch e1 As WebException
				'Ignore network errors when reading error details.
				Return Nothing
			Catch e2 As IOException
				Return Nothing
			Catch e3 As TimeoutException
				Return Nothing
			End Try
		End Function

		Private Shared Function GetErrorDetailsFromStream(ByVal inputStream As Stream) As StorageExtendedErrorInformation
			Dim extendedError As New StorageExtendedErrorInformation()
			Try
				Using reader As XmlReader = XmlReader.Create(inputStream)
					reader.Read()
					reader.ReadStartElement(StorageHttpConstants.XmlElementNames.ErrorRootElement)
					extendedError.ErrorCode = reader.ReadElementString(StorageHttpConstants.XmlElementNames.ErrorCode)
					extendedError.ErrorMessage = reader.ReadElementString(StorageHttpConstants.XmlElementNames.ErrorMessage)
					extendedError.AdditionalDetails = New NameValueCollection()

					' After error code and message we can have a number of additional details optionally followed
					' by ExceptionDetails element - we'll read all of these into the additionalDetails collection
					Do
						If reader.IsStartElement() Then
							If String.Compare(reader.LocalName, StorageHttpConstants.XmlElementNames.ErrorException, StringComparison.Ordinal) = 0 Then
								' Need to read exception details - we have message and stack trace
								reader.ReadStartElement(StorageHttpConstants.XmlElementNames.ErrorException)
								extendedError.AdditionalDetails.Add(StorageHttpConstants.XmlElementNames.ErrorExceptionMessage, reader.ReadElementString(StorageHttpConstants.XmlElementNames.ErrorExceptionMessage))
								extendedError.AdditionalDetails.Add(StorageHttpConstants.XmlElementNames.ErrorExceptionStackTrace, reader.ReadElementString(StorageHttpConstants.XmlElementNames.ErrorExceptionStackTrace))
								reader.ReadEndElement()
							Else
								Dim elementName As String = reader.LocalName
								extendedError.AdditionalDetails.Add(elementName, reader.ReadString())
							End If
						End If
					Loop While reader.Read()
				End Using
			Catch e1 As XmlException
				'If there is a parsing error we cannot return extended error information
				Return Nothing
			End Try
			Return extendedError
		End Function

		Friend Shared Function GetExtendedErrorFromXmlMessage(ByVal xmlErrorMessage As String) As StorageExtendedErrorInformation
			Dim message As String = Nothing
			Dim errorCode As String = Nothing

			Dim xnErrorCode As XName = XName.Get(StorageHttpConstants.XmlElementNames.TableErrorCodeElement, StorageHttpConstants.XmlElementNames.DataWebMetadataNamespace)
			Dim xnMessage As XName = XName.Get(StorageHttpConstants.XmlElementNames.TableErrorMessageElement, StorageHttpConstants.XmlElementNames.DataWebMetadataNamespace)

			Using reader As New StringReader(xmlErrorMessage)
				Dim xDocument As XDocument = Nothing
				Try
					xDocument = XDocument.Load(reader)
				Catch e1 As XmlException
					' The XML could not be parsed. This could happen either because the connection 
					' could not be made to the server, or if the response did not contain the
					' error details (for example, if the response status code was neither a failure
					' nor a success, but a 3XX code such as NotModified.
					Return Nothing
				End Try

				Dim errorCodeElement As XElement = xDocument.Descendants(xnErrorCode).FirstOrDefault()

				If errorCodeElement Is Nothing Then
					Return Nothing
				End If

				errorCode = errorCodeElement.Value

				Dim messageElement As XElement = xDocument.Descendants(xnMessage).FirstOrDefault()

				If messageElement IsNot Nothing Then
					message = messageElement.Value
				End If

			End Using

			Dim errorDetails As New StorageExtendedErrorInformation()
			errorDetails.ErrorMessage = message
			errorDetails.ErrorCode = errorCode
			Return errorDetails
		End Function
	End Class


	#End Region
End Namespace
