'
' <copyright file="Queue.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Configuration
Imports System.Collections.Specialized
Imports System.Xml
Imports System.IO
Imports System.Globalization
Imports System.Diagnostics.CodeAnalysis


' disable the generation of warnings for missing documentation elements for 
' public classes/members in this file


Namespace Microsoft.Samples.ServiceHosting.StorageClient

	''' <summary>
	''' The entry point of the queue storage API
	''' </summary>
	Public MustInherit Class QueueStorage
		''' <summary>
		''' Factory method for QueueStorage
		''' </summary>
		''' <param name="baseUri">The base URI of the queue service</param>
		''' <param name="usePathStyleUris">If true, path-style URIs (http://baseuri/accountname/containername/objectname) are used.
		''' If false host-style URIs (http://accountname.baseuri/containername/objectname) are used,
		''' where baseuri is the URI of the service.
		''' If null, the choice is made automatically: path-style URIs if host name part of base URI is an 
		''' IP addres, host-style otherwise.</param>
		''' <param name="accountName">The name of the storage account</param>
		''' <param name="base64Key">Authentication key used for signing requests</param>
		''' <returns>A newly created QueueStorage instance</returns>
		Public Shared Function Create(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String) As QueueStorage
			Return New QueueStorageRest(New StorageAccountInfo(baseUri, usePathStyleUris, accountName, base64Key))
		End Function

		Public Shared Function Create(ByVal accountInfo As StorageAccountInfo) As QueueStorage
			Return New QueueStorageRest(accountInfo)
		End Function

		''' <summary>
		''' Get a reference to a Queue object with a specified name. This method does not make a call to
		''' the queue service.
		''' </summary>
		''' <param name="queueName">The name of the queue</param>
		''' <returns>A newly created queue object</returns>
		Public MustOverride Function GetQueue(ByVal queueName As String) As MessageQueue


		''' <summary>
		''' Lists the queues within the account.
		''' </summary>
		''' <returns>A list of queues</returns>
		Public Overridable Function ListQueues() As IEnumerable(Of MessageQueue)
			Return ListQueues(Nothing)
		End Function

		''' <summary>
		''' Lists the queues within the account that start with the given prefix.
		''' </summary>
		''' <param name="prefix">If prefix is null returns all queues.</param>
		''' <returns>A list of queues.</returns>
		Public MustOverride Function ListQueues(ByVal prefix As String) As IEnumerable(Of MessageQueue)

		''' <summary>
		''' The time out for each request to the storage service.
		''' </summary>
		Private privateTimeout As TimeSpan
		Public Property Timeout() As TimeSpan
			Get
				Return privateTimeout
			End Get
			Set(ByVal value As TimeSpan)
				privateTimeout = value
			End Set
		End Property

		''' <summary>
		''' The retry policy used for retrying requests
		''' </summary>
		Private privateRetryPolicy As RetryPolicy
		Public Property RetryPolicy() As RetryPolicy
			Get
				Return privateRetryPolicy
			End Get
			Set(ByVal value As RetryPolicy)
				privateRetryPolicy = value
			End Set
		End Property

		''' <summary>
		''' The base URI of the blob storage service
		''' </summary>
		Public ReadOnly Property BaseUri() As Uri
			Get
				Return AccountInfo.BaseUri
			End Get
		End Property

		''' <summary>
		''' The name of the storage account
		''' </summary>
		Public ReadOnly Property AccountName() As String
			Get
				Return AccountInfo.AccountName
			End Get
		End Property

		''' <summary>
		''' Indicates whether to use/generate path-style or host-style URIs
		''' </summary>
		Public ReadOnly Property UsePathStyleUris() As Boolean
			Get
				Return AccountInfo.UsePathStyleUris
			End Get
		End Property

		''' <summary>
		''' The default timeout
		''' </summary>
		<SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes", Justification := "TimeSpan is a non-mutable type")> _
		Public Shared ReadOnly DefaultTimeout As TimeSpan = TimeSpan.FromSeconds(30)

		''' <summary>
		''' The default retry policy
		''' </summary>
        <SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes", Justification:="RetryPolicy is a non-mutable type")> _
        Public Shared ReadOnly DefaultRetryPolicy As RetryPolicy = AddressOf RetryPolicies.NoRetry

		Protected Friend Sub New(ByVal accountInfo As StorageAccountInfo)
			Me.AccountInfo = accountInfo
			Timeout = DefaultTimeout
			RetryPolicy = DefaultRetryPolicy
		End Sub

		Protected Sub New(ByVal other As QueueStorage)
			Me.AccountInfo = other.AccountInfo
			Me.Timeout = other.Timeout
			Me.RetryPolicy = other.RetryPolicy
		End Sub

		Private privateAccountInfo As StorageAccountInfo
		Protected Friend Property AccountInfo() As StorageAccountInfo
			Get
				Return privateAccountInfo
			End Get
			Set(ByVal value As StorageAccountInfo)
				privateAccountInfo = value
			End Set
		End Property
		Private privateCredentials As SharedKeyCredentials
		Protected Friend Property Credentials() As SharedKeyCredentials
			Get
				Return privateCredentials
			End Get
			Set(ByVal value As SharedKeyCredentials)
				privateCredentials = value
			End Set
		End Property
	End Class

	''' <summary>
	''' Objects of this class represent a single message in the queue.
	''' </summary>
	Public Class Message

		''' <summary>
		''' The maximum message size in bytes.
		''' </summary>
		Public Shared ReadOnly MaxMessageSize As Integer = 8 * 1024

		''' <summary>
		''' The maximum amount of time a message is kept in the queue. Max value is 7 days. 
		''' Value is given in seconds.
		''' </summary>
		Public Shared ReadOnly MaxTimeToLive As Integer = 7 * 24 * 60 * 60

		''' <summary>
		''' This constructor is not publicly exposed.
		''' </summary>
		Friend Sub New()
		End Sub

		''' <summary>
		''' Creates a message and initializes the content of the message to be the specified string.
		''' </summary>
		''' <param name="content">A string representing the contents of the message.</param>
		Public Sub New(ByVal content As String)
			If content Is Nothing Then
				Throw New ArgumentNullException("content")
			End If
			Me.content = Encoding.UTF8.GetBytes(content)
		End Sub

		''' <summary>
		''' Creates a message and given the specified byte contents.
		''' In this implementation, regardless of whether an XML or binary data is passed into this 
		''' function, message contents are converted to base64 before passing the data to the queue service.
		''' When calculating the size of the message, the size of the base64 encoding is thus the important
		''' parameter.
		''' </summary>
		''' <param name="content"></param>
		Public Sub New(ByVal content() As Byte)
			If content Is Nothing Then
				Throw New ArgumentNullException("content")
			End If
			If Convert.ToBase64String(content).Length > MaxMessageSize Then
				Throw New ArgumentException("Message body is too big!")
			End If
			Me.content = content
		End Sub


		''' <summary>
		''' A unique ID of the message as returned from queue operations.
		''' </summary>
		Private privateId As String
		Public Property Id() As String
			Get
				Return privateId
			End Get
			Friend Set(ByVal value As String)
				privateId = value
			End Set
		End Property

		''' <summary>
		''' When a message is retrieved from a queue, a PopReceipt is returned. The PopReceipt is used when 
		''' deleting a message from the queue.
		''' </summary>
		Private privatePopReceipt As String
		Public Property PopReceipt() As String
			Get
				Return privatePopReceipt
			End Get
			Friend Set(ByVal value As String)
				privatePopReceipt = value
			End Set
		End Property

		''' <summary>
		''' The point in time when the message was put into the queue.
		''' </summary>
		Private privateInsertionTime As DateTime
		Public Property InsertionTime() As DateTime
			Get
				Return privateInsertionTime
			End Get
			Friend Set(ByVal value As DateTime)
				privateInsertionTime = value
			End Set
		End Property

		''' <summary>
		''' A message's expiration time.
		''' </summary>
		Private privateExpirationTime As DateTime
		Public Property ExpirationTime() As DateTime
			Get
				Return privateExpirationTime
			End Get
			Friend Set(ByVal value As DateTime)
				privateExpirationTime = value
			End Set
		End Property

		''' <summary>
		''' The point in time when a message becomes visible again after a Get() operation was called 
		''' that returned the message.
		''' </summary>
		Private privateTimeNextVisible As DateTime
		Public Property TimeNextVisible() As DateTime
			Get
				Return privateTimeNextVisible
			End Get
			Friend Set(ByVal value As DateTime)
				privateTimeNextVisible = value
			End Set
		End Property

		''' <summary>
		''' Returns the the contents of the message as a string.
		''' </summary>
		Public Function ContentAsString() As String
			Return Encoding.UTF8.GetString(Me.content)
		End Function

		''' <summary>
		''' Returns the content of the message as a byte array
		''' </summary>
		Public Function ContentAsBytes() As Byte()
			Return content
		End Function

		''' <summary>
		''' When calling the Get() operation on a queue, the content of messages 
		''' returned in the REST protocol are represented as Base64-encoded strings.
		''' This internal function transforms the Base64 representation into a byte array.
		''' </summary>
		''' <param name="str">The Base64-encoded string.</param>
		Friend Sub SetContentFromBase64String(ByVal str As String)
			If str Is Nothing OrElse str = String.Empty Then
				' we got a message with an empty <MessageText> element
				Me.content = Encoding.UTF8.GetBytes(String.Empty)
			Else
				Me.content = Convert.FromBase64String(str)
			End If
		End Sub

		''' <summary>
		''' Internal method used for creating the XML that becomes part of a REST request
		''' </summary>
		Friend Function GetContentXMLRepresentation(<System.Runtime.InteropServices.Out()> ByRef length As Integer) As Byte()
			length = 0
			Dim ret() As Byte = Nothing
			Dim builder As New StringBuilder()
			Dim settings As New XmlWriterSettings()
			settings.OmitXmlDeclaration = True
			Dim enc As System.Text.UTF8Encoding = New UTF8Encoding(False)
			Using writer As XmlWriter = XmlWriter.Create(builder, settings)
				writer.WriteStartDocument()
				writer.WriteStartElement(StorageHttpConstants.XmlElementNames.QueueMessage)
				writer.WriteStartElement(StorageHttpConstants.XmlElementNames.MessageText)
				writer.WriteRaw(Convert.ToBase64String(content))
				writer.WriteEndElement()
				writer.WriteEndElement()
				writer.WriteEndDocument()
				writer.Flush()
			End Using
			ret = enc.GetBytes(builder.ToString())
			If ret IsNot Nothing Then
				length = ret.Length
			End If
			Return ret
		End Function

		Private content() As Byte
	End Class

	''' <summary>
	''' Queues in the storage client library expose a functionality for listening for incoming messages. 
	''' If a message is put into a queue, a corresponding event is issued and this delegate is called. This functionality
	''' is implemented internally in this library by periodically polling for incoming messages.
	''' </summary>
	''' <param name="sender">The queue that has received a new event.</param>
	''' <param name="e">The event argument containing the message.</param>
	Public Delegate Sub MessageReceivedEventHandler(ByVal sender As Object, ByVal e As MessageReceivedEventArgs)

	''' <summary>
	''' The argument class for the MessageReceived event.
	''' </summary>
	Public Class MessageReceivedEventArgs
		Inherits EventArgs
		''' <summary>
		''' The message itself.
		''' </summary>
		Private _msg As Message

		''' <summary>
		''' Constructor for creating a message received argument.
		''' </summary>
		''' <param name="msg"></param>
		Public Sub New(ByVal msg As Message)
			If msg Is Nothing Then
				Throw New ArgumentNullException("msg")
			End If
			_msg = msg
		End Sub

		''' <summary>
		''' The message received by the queue.
		''' </summary>
		Public Property Message() As Message
			Get
				Return _msg
			End Get
			Set(ByVal value As Message)
				If value Is Nothing Then
					Throw New ArgumentNullException("value")
				End If
				_msg = value
			End Set
		End Property
	End Class

	Public Class QueueProperties
		''' <summary>
		''' The approximated amount of messages in the queue.
		''' </summary>
		Private privateApproximateMessageCount As Integer
		Public Property ApproximateMessageCount() As Integer
			Get
				Return privateApproximateMessageCount
			End Get
			Friend Set(ByVal value As Integer)
				privateApproximateMessageCount = value
			End Set
		End Property

		''' <summary>
		''' Metadata for the queue in the form of name-value pairs.
		''' </summary>
		Private privateMetadata As NameValueCollection
		Public Property Metadata() As NameValueCollection
			Get
				Return privateMetadata
			End Get
			Set(ByVal value As NameValueCollection)
				privateMetadata = value
			End Set
		End Property
	End Class

	''' <summary>
	''' Objects of this class represent a queue in a user's storage account.
	''' </summary>
	Public MustInherit Class MessageQueue

		''' <summary>
		''' The name of the queue.
		''' </summary>
		Private _name As String

		''' <summary>
		''' The user account this queue lives in.
		''' </summary>
		Private _account As StorageAccountInfo


		''' <summary>
		''' This constructor is only called by subclasses.
		''' </summary>
		Protected Friend Sub New()
			' queues are generated using factory methods
		End Sub

		Protected Friend Sub New(ByVal name As String, ByVal account As StorageAccountInfo)
			If String.IsNullOrEmpty(name) Then
				Throw New ArgumentException("Queue name cannot be null or empty!")
			End If
			If account Is Nothing Then
				Throw New ArgumentNullException("account")
			End If
			If (Not Utilities.IsValidContainerOrQueueName(name)) Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "The specified queue name ""{0}"" is not valid!" & "Please choose a name that conforms to the naming conventions for queues!", name))
			End If
			_name = name
			_account = account
		End Sub

		''' <summary>
		''' The name of the queue exposed as a public property.
		''' </summary>
		Public ReadOnly Property Name() As String
			Get
				Return _name
			End Get
		End Property

		''' <summary>
		''' The account info object this queue lives in -- exposed as an internal property.
		''' </summary>
		Friend ReadOnly Property AccountInfo() As StorageAccountInfo
			Get
				Return _account
			End Get
		End Property

		''' <summary>
		''' Indicates whether to use/generate path-style or host-style URIs
		''' </summary>
		Public ReadOnly Property UsePathStyleUris() As Boolean
			Get
				Return _account.UsePathStyleUris
			End Get
		End Property

		''' <summary>
		''' The URI of the queue
		''' </summary>
		Public MustOverride ReadOnly Property QueueUri() As Uri

		''' <summary>
		''' The retry policy used for retrying requests; this is the retry policy of the 
		''' storage account where this queue was created
		''' </summary>
		Private privateRetryPolicy As RetryPolicy
		Public Property RetryPolicy() As RetryPolicy
			Get
				Return privateRetryPolicy
			End Get
			Set(ByVal value As RetryPolicy)
				privateRetryPolicy = value
			End Set
		End Property

		''' <summary>
		''' The timeout of requests.
		''' </summary>
		Private privateTimeout As TimeSpan
		Public Property Timeout() As TimeSpan
			Get
				Return privateTimeout
			End Get
			Set(ByVal value As TimeSpan)
				privateTimeout = value
			End Set
		End Property

		''' <summary>
		''' Creates a queue in the specified storage account.
		''' </summary>
		''' <param name="queueAlreadyExists">true if a queue with the same name already exists.</param>
		''' <returns>true if the queue was successfully created.</returns>
		Public MustOverride Function CreateQueue(<System.Runtime.InteropServices.Out()> ByRef queueAlreadyExists As Boolean) As Boolean

		''' <summary>
		''' Creates a queue in the specified storage account.
		''' </summary>
		''' <returns>true if the queue was successfully created.</returns>
		Public Overridable Function CreateQueue() As Boolean
			Dim ignore As Boolean
			Return CreateQueue(ignore)
		End Function


		''' <summary>
		''' Determines whether a queue with the same name already exists in an account.
		''' </summary>
		''' <returns>true if a queue with the same name already exists.</returns>
		Public Overridable Function DoesQueueExist() As Boolean
			Try
				Me.GetProperties()
				Return True
			Catch e As StorageClientException
				If e.ErrorCode = StorageErrorCode.ResourceNotFound Then
					Return False
				End If
				Throw
			End Try
		End Function

		''' <summary>
		''' Deletes the queue. The queue will be deleted regardless of whether there are messages in the 
		''' queue or not.
		''' </summary>
		''' <returns>true if the queue was successfully deleted.</returns>
		Public MustOverride Function DeleteQueue() As Boolean

		''' <summary>
		''' Sets the properties of a queue.
		''' </summary>
		''' <param name="properties">The queue's properties to set.</param>
		''' <returns>true if the properties were successfully written to the queue.</returns>
		Public MustOverride Function SetProperties(ByVal properties As QueueProperties) As Boolean

		''' <summary>
		''' Retrieves the queue's properties.
		''' </summary>
		''' <returns>The queue's properties.</returns>
		Public MustOverride Function GetProperties() As QueueProperties

		''' <summary>
		''' Retrieves the approximate number of messages in a queue.
		''' </summary>
		''' <returns>The approximate number of messages in this queue.</returns>
		Public MustOverride Function ApproximateCount() As Integer

		''' <summary>
		''' Puts a message in the queue.
		''' </summary>
		''' <param name="msg">The message to store in the queue.</param>
		''' <returns>true if the message has been successfully enqueued.</returns>
		Public MustOverride Function PutMessage(ByVal msg As Message) As Boolean

		''' <summary>
		''' Puts a message in the queue.
		''' </summary>
		''' <param name="msg">The message to store in the queue.</param>
		''' <param name="timeToLiveInSeconds">The time to live for the message in seconds.</param>
		''' <returns>true if the message has been successfully enqueued.</returns>
		Public MustOverride Function PutMessage(ByVal msg As Message, ByVal timeToLiveInSeconds As Integer) As Boolean

		''' <summary>
		''' Retrieves a message from the queue. 
		''' </summary>
		''' <returns>The message retrieved or null if the queue is empty.</returns>
		Public MustOverride Function GetMessage() As Message

		''' <summary>
		''' Retrieves a message and sets its visibility timeout to the specified number of seconds.
		''' </summary>
		''' <param name="visibilityTimeoutInSeconds">Visibility timeout of the message retrieved in seconds.</param>
		''' <returns></returns>
		Public MustOverride Function GetMessage(ByVal visibilityTimeoutInSeconds As Integer) As Message

		''' <summary>
		''' Tries to retrieve the given number of messages.
		''' </summary>
		''' <param name="numberOfMessages">Maximum number of messages to retrieve.</param>
		''' <returns>The list of messages retrieved.</returns>
		Public MustOverride Function GetMessages(ByVal numberOfMessages As Integer) As IEnumerable(Of Message)

		''' <summary>
		''' Tries to retrieve the given number of messages.
		''' </summary>
		''' <param name="numberOfMessages">Maximum number of messages to retrieve.</param>
		''' <param name="visibilityTimeoutInSeconds">The visibility timeout of the retrieved messages in seconds.</param>
		''' <returns>The list of messages retrieved.</returns>
		Public MustOverride Function GetMessages(ByVal numberOfMessages As Integer, ByVal visibilityTimeoutInSeconds As Integer) As IEnumerable(Of Message)

		''' <summary>
		''' Get a message from the queue but do not actually dequeue it. The message will remain visible 
		''' for other parties requesting messages.
		''' </summary>
		''' <returns>The message retrieved or null if there are no messages in the queue.</returns>
		Public MustOverride Function PeekMessage() As Message

		''' <summary>
		''' Tries to get a copy of messages in the queue without actually dequeuing the messages.
		''' The messages will remain visible in the queue.
		''' </summary>
		''' <param name="numberOfMessages">Maximum number of message to retrieve.</param>
		''' <returns>The list of messages retrieved.</returns>
		Public MustOverride Function PeekMessages(ByVal numberOfMessages As Integer) As IEnumerable(Of Message)

		''' <summary>
		''' Deletes a message from the queue.
		''' </summary>
		''' <param name="msg">The message to retrieve with a valid popreceipt.</param>
		''' <returns>true if the operation was successful.</returns>
		Public MustOverride Function DeleteMessage(ByVal msg As Message) As Boolean

		''' <summary>
		''' Delete all messages in a queue.
		''' </summary>
		''' <returns>true if all messages were deleted successfully.</returns>
		Public MustOverride Function Clear() As Boolean

		''' <summary>
		''' The default time interval between polling the queue for messages. 
		''' Polling is only enabled if the user has called StartReceiving().
		''' </summary>
		Public Shared ReadOnly DefaultPollInterval As Integer = 5000

		''' <summary>
		''' The poll interval in milliseconds. If not explicitly set, this defaults to 
		''' the DefaultPollInterval.
		''' </summary>
		Public MustOverride Property PollInterval() As Integer

		''' <summary>
		''' Starts the automatic reception of messages.
		''' </summary>
		''' <returns>true if the operation was successful.</returns>
		Public MustOverride Function StartReceiving() As Boolean

		''' <summary>
		''' Stop the automatic reception of messages.
		''' </summary>
		Public MustOverride Sub StopReceiving()

		''' <summary>
		''' The event users subscribe to in order to automatically receive messages
		''' from a queue.
		''' </summary>
        Public Event MessageReceived As MessageReceivedEventHandler

	End Class

End Namespace
