'
' <copyright file="RestBlobStorage.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Collections.Specialized
Imports System.Globalization
Imports System.IO
Imports System.Net
Imports System.Text
Imports System.Xml
Imports Microsoft.Samples.ServiceHosting.StorageClient.StorageHttpConstants

	Imports System.Reflection
Namespace Microsoft.Samples.ServiceHosting.StorageClient

	Friend Class BlobStorageRest
		Inherits BlobStorage
		Private credentials As SharedKeyCredentials
		Friend Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String)
			MyBase.New(baseUri, usePathStyleUris, accountName, base64Key)
			Dim key() As Byte = Nothing
			Me.Base64Key = base64Key
			If base64Key IsNot Nothing Then
				key = Convert.FromBase64String(base64Key)
			End If
			Me.credentials = New SharedKeyCredentials(accountName, key)
		End Sub

		''' <summary>
		''' Get a reference to a BlobContainer object with the given name.
		''' This method does not make any calls to the storage service.
		''' </summary>
		''' <param name="containerName">The name of the container</param>
		''' <returns>A reference to a newly created BlobContainer object</returns>
		Public Overrides Function GetBlobContainer(ByVal containerName As String) As BlobContainer
			Return New BlobContainerRest(BaseUri, UsePathStyleUris, AccountName, containerName, Base64Key, DateTime.MinValue, Timeout, RetryPolicy)
		End Function

		''' <summary>
		''' Lists the containers within the account.
		''' </summary>
		''' <returns>A list of containers</returns>
		Public Overrides Function ListBlobContainers() As IEnumerable(Of BlobContainer)
			Dim marker As String = "", prefix As String = Nothing
			Const maxResults As Integer = StorageHttpConstants.ListingConstants.MaxContainerListResults
			Dim key() As Byte = Nothing
			If Base64Key IsNot Nothing Then
				key = Convert.FromBase64String(Base64Key)
			End If
			Dim credentials As New SharedKeyCredentials(AccountName, key)

            Dim containers = New List(Of BlobContainer)
			Do
				Dim result As ListContainersResult = ListContainersImpl(prefix, marker, maxResults)
                marker = result.NextMarker
                For Each container As ContainerProperties In result.Containers
                    containers.Add(New BlobContainerRest(BaseUri, UsePathStyleUris, AccountName, container.Name, Base64Key, container.LastModifiedTime, Timeout, RetryPolicy))
                Next container
			Loop While marker IsNot Nothing
            Return containers
		End Function

		Friend Class ListContainersResult
			Friend Sub New(ByVal containers As IEnumerable(Of ContainerProperties), ByVal nextMarker As String)
				Me.Containers = containers
				Me.NextMarker = nextMarker
			End Sub

			Private privateContainers As IEnumerable(Of ContainerProperties)
			Public Property Containers() As IEnumerable(Of ContainerProperties)
				Get
					Return privateContainers
				End Get
				Private Set(ByVal value As IEnumerable(Of ContainerProperties))
					privateContainers = value
				End Set
			End Property

			Private privateNextMarker As String
			Public Property NextMarker() As String
				Get
					Return privateNextMarker
				End Get
				Private Set(ByVal value As String)
					privateNextMarker = value
				End Set
			End Property

		End Class

		Private Function ListContainersImpl(ByVal prefix As String, ByVal marker As String, ByVal maxResults As Integer) As ListContainersResult
			Dim result As ListContainersResult = Nothing
			Dim uriComponents As ResourceUriComponents
			Dim listContainerssUri As Uri = CreateRequestUriForListContainers(prefix, marker, Nothing, maxResults, uriComponents)
			Dim request As HttpWebRequest = Utilities.CreateHttpRequest(listContainerssUri, StorageHttpConstants.HttpMethod.Get, Timeout)
			credentials.SignRequest(request, uriComponents)

			Try
				Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
					If response.StatusCode = HttpStatusCode.OK Then
						Using stream As Stream = response.GetResponseStream()
							result = ListContainersResultFromResponse(stream)
						End Using
					Else
						Utilities.ProcessUnexpectedStatusCode(response)
					End If
				End Using
			Catch ioe As IOException
				Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, Nothing, ioe)
			Catch te As System.TimeoutException
				Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during listing containers", HttpStatusCode.RequestTimeout, Nothing, te)
			Catch we As WebException
				Throw Utilities.TranslateWebException(we)
			End Try
			Return result
		End Function

		Private Function CreateRequestUriForListContainers(ByVal prefix As String, ByVal fromMarker As String, ByVal delimiter As String, ByVal maxResults As Integer, <System.Runtime.InteropServices.Out()> ByRef uriComponents As ResourceUriComponents) As Uri
			Dim queryParams As NameValueCollection = BlobStorageRest.CreateRequestUriForListing(prefix, fromMarker, delimiter, maxResults)

			Return Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, Nothing, Nothing, Timeout, queryParams, uriComponents)
		End Function

		Friend Shared Function CreateRequestUriForListing(ByVal prefix As String, ByVal fromMarker As String, ByVal delimiter As String, ByVal maxResults As Integer) As NameValueCollection
			Dim queryParams As New NameValueCollection()
			queryParams.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.List)

			If (Not String.IsNullOrEmpty(prefix)) Then
				queryParams.Add(StorageHttpConstants.QueryParams.QueryParamPrefix, prefix)
			End If

			If (Not String.IsNullOrEmpty(fromMarker)) Then
				queryParams.Add(StorageHttpConstants.QueryParams.QueryParamMarker, fromMarker)
			End If

			If (Not String.IsNullOrEmpty(delimiter)) Then
				queryParams.Add(StorageHttpConstants.QueryParams.QueryParamDelimiter, delimiter)
			End If

			queryParams.Add(StorageHttpConstants.QueryParams.QueryParamMaxResults, maxResults.ToString(CultureInfo.InvariantCulture))

			Return queryParams
		End Function

		Private Shared Function ListContainersResultFromResponse(ByVal responseBody As Stream) As ListContainersResult
			Dim containers As New List(Of ContainerProperties)()
			Dim nextMarker As String = Nothing

			Dim doc As New XmlDocument()
			Try
				doc.Load(responseBody)
			Catch xe As XmlException
				Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "The result of a ListBlobContainers operation could not be parsed", Nothing, xe)
			End Try


			' Get all the containers returned as the listing results
			Dim containerNodes As XmlNodeList = doc.SelectNodes(XPathQueryHelper.ContainerQuery)

			For Each containerNode As XmlNode In containerNodes
				Dim lastModified? As DateTime = XPathQueryHelper.LoadSingleChildDateTimeValue(containerNode, StorageHttpConstants.XmlElementNames.LastModified, False)

				Dim eTag As String = XPathQueryHelper.LoadSingleChildStringValue(containerNode, StorageHttpConstants.XmlElementNames.Etag, False)

				Dim containerUri As String = XPathQueryHelper.LoadSingleChildStringValue(containerNode, StorageHttpConstants.XmlElementNames.Url, True)

				Dim containerName As String = XPathQueryHelper.LoadSingleChildStringValue(containerNode, StorageHttpConstants.XmlElementNames.Name, True)

				Dim properties As New ContainerProperties(containerName)
				If lastModified.HasValue Then
					properties.LastModifiedTime = lastModified.Value
				End If
				properties.ETag = eTag

				Dim uri As Uri = Nothing
				Uri.TryCreate(containerUri, UriKind.Absolute, uri)
				properties.Uri = uri

				containers.Add(properties)
			Next containerNode

			' Get the nextMarker
			Dim nextMarkerNode As XmlNode = doc.SelectSingleNode(XPathQueryHelper.NextMarkerQuery)
			If nextMarkerNode IsNot Nothing AndAlso nextMarkerNode.FirstChild IsNot Nothing Then
				nextMarker = nextMarkerNode.FirstChild.Value
			End If

			Return New ListContainersResult(containers, nextMarker)
		End Function
	End Class

	Friend Class BlobContainerRest
		Inherits BlobContainer

		Friend Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris As Boolean, ByVal accountName As String, ByVal containerName As String, ByVal base64Key As String, ByVal lastModified As DateTime, ByVal timeOut As TimeSpan, ByVal retryPolicy As RetryPolicy)
			MyBase.New(baseUri, usePathStyleUris, accountName, containerName, lastModified)
			Dim uriComponents As New ResourceUriComponents(accountName, containerName, Nothing)
			containerUri_Renamed = HttpRequestAccessor.ConstructResourceUri(baseUri, uriComponents, usePathStyleUris)
			If base64Key IsNot Nothing Then
				key = Convert.FromBase64String(base64Key)
			End If
			credentials = New SharedKeyCredentials(accountName, key)
			Timeout = timeOut
			RetryPolicy = retryPolicy
		End Sub

		Public Overrides ReadOnly Property ContainerUri() As Uri
			Get
				Return Me.containerUri_Renamed
			End Get
		End Property

		Public Overrides Function CreateContainer() As Boolean
			Return CreateContainerImpl(Nothing, ContainerAccessControl.Private)
		End Function

		''' <summary>
		''' Create the container with the specified access control if it does not exist
		''' </summary>
		''' <param name="metadata">The metadata for the container. Can be null to indicate no metadata</param>
		''' <param name="accessControl">The access control (public or private) with which to create the container</param>
		''' <returns>true if the container was created. false if the container already exists</returns>
		Public Overrides Function CreateContainer(ByVal metadata As NameValueCollection, ByVal accessControl As ContainerAccessControl) As Boolean
			Return CreateContainerImpl(metadata, accessControl)
		End Function


		Public Overrides Function DoesContainerExist() As Boolean
            Dim result As Boolean = False
            RetryPolicy.Invoke(Function() DoesContainerExistAction(result))
            Return result
		End Function

        Private Function DoesContainerExistAction(ByRef result As Boolean) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, New NameValueCollection(), uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, Timeout)
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        result = True
                    ElseIf response.StatusCode = HttpStatusCode.Gone OrElse response.StatusCode = HttpStatusCode.NotFound Then
                        result = False
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                        result = False
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso ((CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Gone OrElse (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.NotFound) Then
                    result = False
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

        ''' <summary>
        ''' Get the properties for the container if it exists.
        ''' </summary>
        ''' <returns>The metadata for the container if it exists, null otherwise</returns>
        Public Overrides Function GetContainerProperties() As ContainerProperties
            Dim result As ContainerProperties = Nothing
            RetryPolicy.Invoke(Function() GetContainerPropertiesAction(result))
            Return result
        End Function

        Private Function GetContainerPropertiesAction(ByRef result As ContainerProperties) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, ContainerName, Nothing, TimeOut, New NameValueCollection(), uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, TimeOut)
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        result = ContainerPropertiesFromResponse(response)
                    ElseIf response.StatusCode = HttpStatusCode.Gone OrElse response.StatusCode = HttpStatusCode.NotFound Then
                        result = Nothing
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                        result = Nothing
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Gone OrElse ((CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.NotFound) Then
                    result = Nothing
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

        ''' <summary>
        ''' Get the access control permissions associated with the container.
        ''' </summary>
        Public Overrides Function GetContainerAccessControl() As ContainerAccessControl
            Dim accessControl As ContainerAccessControl = ContainerAccessControl.Private
            RetryPolicy.Invoke(Function() GetContainerAccessControlAction(accessControl))
            Return accessControl
        End Function

        Private Function GetContainerAccessControlAction(ByRef accessControl As ContainerAccessControl) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim queryParams As New NameValueCollection()
            queryParams.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.Acl)
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, queryParams, uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, Timeout)
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        Dim acl As String = response.Headers(StorageHttpConstants.HeaderNames.PublicAccess)
                        Dim publicAcl As Boolean = False
                        If acl IsNot Nothing AndAlso Boolean.TryParse(acl, publicAcl) Then
                            accessControl = (If(publicAcl, ContainerAccessControl.Public, ContainerAccessControl.Private))
                        Else
                            Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "The server did not respond with expected container access control header", Nothing, Nothing, Nothing)
                        End If
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

        ''' <summary>
        ''' Get the access control permissions associated with the container.
        ''' </summary>
        Public Overrides Sub SetContainerAccessControl(ByVal acl As ContainerAccessControl)
            RetryPolicy.Invoke(Function() SetContainerAccessControlAction(acl))
        End Sub

        Private Function SetContainerAccessControlAction(ByRef acl As ContainerAccessControl) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim queryParams As New NameValueCollection()
            queryParams.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.Acl)
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, queryParams, uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Put, Timeout)
            request.Headers.Add(StorageHttpConstants.HeaderNames.PublicAccess, (acl = ContainerAccessControl.Public).ToString())
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode <> HttpStatusCode.OK Then
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

        Public Overrides Function DeleteContainer() As Boolean
            Dim result As Boolean = False
            RetryPolicy.Invoke(Function() DeleteContainerAction(result))
            Return result
        End Function

        Private Function DeleteContainerAction(ByRef result As Boolean) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, New NameValueCollection(), uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Delete, Timeout)
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK OrElse response.StatusCode = HttpStatusCode.Accepted Then
                        result = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso ((CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.NotFound OrElse (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Gone OrElse (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Conflict) Then
                    result = False
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

        Private Function CreateContainerImpl(ByVal metadata As NameValueCollection, ByVal accessControl As ContainerAccessControl) As Boolean
            Dim result As Boolean = False
            RetryPolicy.Invoke(Function() CreateContainerImplAction(metadata, accessControl, result))
            Return result
        End Function

        Private Function CreateContainerImplAction(ByVal metadata As NameValueCollection, ByVal accessControl As ContainerAccessControl, ByRef result As Boolean) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, New NameValueCollection(), uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Put, Timeout)
            If metadata IsNot Nothing Then
                Utilities.AddMetadataHeaders(request, metadata)
            End If
            If accessControl = ContainerAccessControl.Public Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.PublicAccess, "true")
            End If
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.Created Then
                        result = True
                    ElseIf response.StatusCode = HttpStatusCode.Conflict Then
                        result = False
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Conflict Then
                    result = False
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

        Public Overrides Function DoesBlobExist(ByVal blobName As String) As Boolean
            'if the blob exists, the GetBlobProperties function should return for us a BlobProperties object, otherwise it returns null
            If GetBlobProperties(blobName) Is Nothing Then
                Return False
            Else
                Return True
            End If
        End Function

        ''' <summary>
        ''' Create a new blob or overwrite an existing blob.
        ''' </summary>
        ''' 
        ''' <param name="blobProperties">The properties of the blob</param>
        ''' <param name="blobContents">The contents of the blob</param>
        ''' <param name="overwrite">Should this request overwrite an existing blob ?</param>
        ''' <returns>true if the blob was created. false if the blob already exists and <paramref name="overwrite"/>was set to false</returns>
        ''' <remarks>The LastModifiedTime property of <paramref name="blobProperties"/> is set as a result of this call</remarks>
        Public Overrides Function CreateBlob(ByVal blobProperties As BlobProperties, ByVal blobContents As BlobContents, ByVal overwrite As Boolean) As Boolean
            Return PutBlobImpl(blobProperties, blobContents.AsStream, overwrite, Nothing)
        End Function

        ''' <summary>
        ''' Updates an existing blob if it has not been modified since the specified time which is typically
        ''' the last modified time of the blob when you retrieved it.
        ''' Use this method to implement optimistic concurrency by avoiding clobbering changes to the blob
        ''' made by another writer.
        ''' </summary>
        ''' <param name="blobProperties">The properties of the blob. This object should be one previously
        ''' obtained from a call to GetBlob or GetBlobProperties and have its LastModifiedTime property set.</param>
        ''' <param name="contents">The contents of the blob. The contents of the blob should be readable</param>
        ''' <returns>true if the blob was updated. false if the blob has changed since the last time</returns>
        ''' <remarks>The LastModifiedTime property of <paramref name="properties"/> is set as a result of this call</remarks>
        Public Overrides Function UpdateBlobIfNotModified(ByVal blobProperties As BlobProperties, ByVal contents As BlobContents) As Boolean
            Return PutBlobImpl(blobProperties, contents.AsStream, True, blobProperties.ETag)
        End Function

        ''' <summary>
        ''' Get the blob contents and properties if the blob exisits
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <param name="blobContents">Object in which the contents are returned.
        ''' This object should contain a writable stream or should be a default constructed object.</param>
        ''' <param name="transferAsChunks">Should the blob be gotten in pieces. This requires more round-trips, but will retry smaller piecs in case of failure.</param>
        ''' <returns>The properties of the blob if the blob exists.</returns>
        Public Overrides Function GetBlob(ByVal name As String, ByVal blobContents As BlobContents, ByVal transferAsChunks As Boolean) As BlobProperties
            Dim notModified As Boolean = False
            Return GetBlobImpl(name, blobContents.AsStream, Nothing, transferAsChunks, notModified)
        End Function

        ''' <summary>
        ''' Gets the blob contents and properties if the blob has not been modified since the time specified.
        ''' Use this method if you have cached the contents of a blob and want to avoid retrieving the blob
        ''' if it has not changed since the last time you retrieved it.
        ''' </summary>
        ''' <param name="blobProperties">The properties of the blob obtained from an earlier call to GetBlob. This
        ''' parameter is updated by the call if the blob has been modified</param>
        ''' <param name="blobContents">Contains the stream to which the contents of the blob are written if it has been
        ''' modified</param>
        ''' <param name="transferAsChunks">Should the blob be gotten in pieces. This requires more round-trips, but will retry smaller piecs in case of failure.</param>
        ''' <returns>true if the blob has been modified, false otherwise</returns>
        Public Overrides Function GetBlobIfModified(ByVal blobProperties As BlobProperties, ByVal blobContents As BlobContents, ByVal transferAsChunks As Boolean) As Boolean
            Dim modified As Boolean = True
            Dim newProperties As BlobProperties = GetBlobImpl(blobProperties.Name, blobContents.AsStream, blobProperties.ETag, transferAsChunks, modified)
            If modified Then
                blobProperties.Assign(newProperties)
            End If
            Return modified
        End Function

        ''' <summary>
        ''' Get the properties of the blob if it exists.
        ''' This method is also the simplest way to check if a blob exists.
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <returns>The properties of the blob if it exists. null otherwise.
        ''' The properties for the contents of the blob are not set</returns>
        Public Overrides Function GetBlobProperties(ByVal name As String) As BlobProperties
            Try
                Dim modified As Boolean = False
                Return GetBlobImpl(name, Nothing, Nothing, False, modified)
            Catch se As StorageClientException
                If se.ErrorCode = StorageErrorCode.ResourceNotFound OrElse se.ErrorCode = StorageErrorCode.BlobNotFound Then
                    Return Nothing
                End If
                Throw
            End Try
        End Function

        ''' <summary>
        ''' Set the metadata of an existing blob.
        ''' </summary>
        ''' <param name="blobProperties">The blob properties object whose metadata is to be updated</param>
        Public Overrides Sub UpdateBlobMetadata(ByVal blobProperties As BlobProperties)
            SetBlobMetadataImpl(blobProperties, Nothing)
        End Sub

        ''' <summary>
        ''' Set the metadata of an existing blob if it has not been modified since it was last retrieved.
        ''' </summary>
        ''' <param name="blobProperties">The blob properties object whose metadata is to be updated.
        ''' Typically obtained by a previous call to GetBlob or GetBlobProperties</param>
        Public Overrides Function UpdateBlobMetadataIfNotModified(ByVal blobProperties As BlobProperties) As Boolean
            Return SetBlobMetadataImpl(blobProperties, blobProperties.ETag)
        End Function

        ''' <summary>
        ''' Delete a blob with the given name
        ''' </summary>
        ''' <param name="name">The name of the blob</param>
        ''' <returns>true if the blob exists and was successfully deleted, false if the blob does not exist</returns>
        Public Overrides Function DeleteBlob(ByVal name As String) As Boolean
            Dim unused As Boolean = False
            Return DeleteBlobImpl(name, Nothing, unused)
        End Function

        ''' <summary>
        ''' Delete a blob with the given name if the blob has not been modified since it was last obtained.
        ''' Use this method for optimistic concurrency to avoid deleting a blob that has been modified since
        ''' the last time you retrieved it
        ''' </summary>
        ''' <param name="blob">A blob object (typically previously obtained from a GetBlob call)</param>
        ''' <param name="modified">This out parameter is set to true if the blob was not deleted because
        ''' it was modified</param>
        ''' <returns>true if the blob exists and was successfully deleted, false if the blob does not exist or was
        ''' not deleted because the blob was modified.</returns>
        Public Overrides Function DeleteBlobIfNotModified(ByVal blob As BlobProperties, <System.Runtime.InteropServices.Out()> ByRef modified As Boolean) As Boolean
            Return DeleteBlobImpl(blob.Name, blob.ETag, modified)
        End Function

        ''' <summary>
        ''' Enumerates all blobs with a given prefix.
        ''' </summary>
        ''' <param name="prefix"></param>
        ''' <param name="combineCommonPrefixes">If true common prefixes with "/" as seperator</param>
        ''' <returns>The list of blob properties and common prefixes</returns>
        Public Overrides Function ListBlobs(ByVal prefix As String, ByVal combineCommonPrefixes As Boolean) As IEnumerable(Of Object)
            Dim marker As String = ""
            Const maxResults As Integer = StorageHttpConstants.ListingConstants.MaxBlobListResults

            Dim delimiter As String = If(combineCommonPrefixes, "/", "")
            Dim blobs As New List(Of Object)
            Do
                Dim result As ListBlobsResult = ListBlobsImpl(prefix, marker, delimiter, maxResults)
                marker = result.NextMarker
                For Each commonPrefix As String In result.CommonPrefixes
                    blobs.Add(commonPrefix)
                Next commonPrefix

                For Each blob As BlobProperties In result.Blobs
                    blobs.Add(blob)
                Next blob
            Loop While marker IsNot Nothing
            Return blobs
        End Function

        Private Function CreateHttpRequestForPutBlob(ByVal uri As Uri, ByVal httpMethod As String, ByVal blobProperties As BlobProperties, ByVal contentLength As Long, ByVal overwrite As Boolean, ByVal eTag As String) As HttpWebRequest
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, httpMethod, Timeout)
            If blobProperties.ContentEncoding IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.ContentEncoding, blobProperties.ContentEncoding)
            End If
            If blobProperties.ContentLanguage IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.ContentLanguage, blobProperties.ContentLanguage)
            End If
            If blobProperties.ContentType IsNot Nothing Then
                request.ContentType = blobProperties.ContentType
            End If
            If eTag IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.IfMatch, eTag)
            End If

            If blobProperties.Metadata IsNot Nothing AndAlso blobProperties.Metadata.Count > 0 Then
                Utilities.AddMetadataHeaders(request, blobProperties.Metadata)
            End If
            request.ContentLength = contentLength
            If (Not overwrite) Then
                request.Headers.Set(StorageHttpConstants.HeaderNames.IfNoneMatch, "*")
            End If
            Return request
        End Function

        Private Function CreateHttpRequestForGetBlob(ByVal uri As Uri, ByVal httpMethod As String, ByVal ifNoneMatchETag As String, ByVal ifMatchETag As String) As HttpWebRequest
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, httpMethod, Timeout)
            If ifNoneMatchETag IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.IfNoneMatch, ifNoneMatchETag)
            End If
            If ifMatchETag IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.IfMatch, ifMatchETag)
            End If
            Return request
        End Function


        ''' <summary>
        ''' Uploads a blob in chunks.
        ''' </summary>
        ''' <param name="blobProperties"></param>
        ''' <param name="stream"></param>
        ''' <param name="overwrite"></param>
        ''' <param name="eTag"></param>
        ''' <returns></returns>
        Private Function PutLargeBlobImpl(ByVal blobProperties As BlobProperties, ByVal stream As Stream, ByVal overwrite As Boolean, ByVal eTag As String) As Boolean
            Dim retval As Boolean = False
            ' Since we got a large block, chunk it into smaller pieces called blocks
            Dim blockSize As Long = StorageHttpConstants.BlobBlockConstants.BlockSize
            Dim startPosition As Long = stream.Position
            Dim length As Long = stream.Length - startPosition
            Dim numBlocks As Integer = CInt(Fix(Math.Ceiling(CDbl(length) / blockSize)))
            Dim blockIds(numBlocks - 1) As String

            'We can retry only if the stream supports seeking. An alternative is to buffer the data in memory
            'but we do not do this currently.
            Dim R As RetryPolicy

            If (stream.CanSeek) Then
                R = Me.RetryPolicy
            Else
                R = AddressOf RetryPolicies.NoRetry
            End If

            'Upload each of the blocks, retrying any failed uploads
            For i As Integer = 0 To numBlocks - 1
                Dim blockId As String = Convert.ToBase64String(System.BitConverter.GetBytes(i))
                blockIds(i) = blockId
                ' Rewind the stream to appropriate location in case this is a retry
                R(Function() AnonymousMethod7(stream, blockSize, startPosition, i, blockId, length, blobProperties, overwrite, eTag, retval))
            Next i

            ' Now commit the list
            ' First create the output
            Using buffer As New MemoryStream()
                ' construct our own XML segment with correct blockId's
                Dim writer As New XmlTextWriter(buffer, Encoding.UTF8)
                writer.WriteStartDocument()
                writer.WriteStartElement(XmlElementNames.BlockList)
                For Each id As String In blockIds
                    writer.WriteElementString(XmlElementNames.Block, id)
                Next id
                writer.WriteEndElement()
                writer.WriteEndDocument()
                writer.Flush()
                buffer.Position = 0 'Rewind

                Dim nvc As New NameValueCollection()
                nvc.Add(QueryParams.QueryParamComp, CompConstants.BlockList)

                retval = UploadData(blobProperties, buffer, buffer.Length, overwrite, eTag, nvc)
            End Using

            Return retval
        End Function

        Private Function AnonymousMethod7(ByVal stream As Stream, ByVal blockSize As Long, ByVal startPosition As Long, ByVal i As Integer, ByVal blockId As String, ByVal length As Long, ByVal blobProperties As BlobProperties, ByVal overwrite As Boolean, ByVal eTag As String, ByVal retval As Boolean) As Boolean
            If stream.CanSeek Then
                stream.Position = startPosition + i * blockSize
            End If
            Dim nvc As New NameValueCollection()
            nvc.Add(QueryParams.QueryParamComp, CompConstants.Block)
            nvc.Add(QueryParams.QueryParamBlockId, blockId)
            Dim blockLength As Long = Math.Min(blockSize, length - stream.Position)
            retval = UploadData(blobProperties, stream, blockLength, overwrite, eTag, nvc)
            Return True
        End Function

        Private Function PutBlobImpl(ByVal blobProperties As BlobProperties, ByVal stream As Stream, ByVal overwrite As Boolean, ByVal eTag As String) As Boolean
            ' If the blob is large, we should use blocks to upload it in pieces.
            ' This will ensure that a broken connection will only impact a single piece
            Dim originalPosition As Long = stream.Position
            Dim length As Long = stream.Length - stream.Position
            If length > StorageHttpConstants.BlobBlockConstants.MaximumBlobSizeBeforeTransmittingAsBlocks Then
                Return PutLargeBlobImpl(blobProperties, stream, overwrite, eTag)
            End If

            Dim retval As Boolean = False
            Dim R As RetryPolicy

            If (stream.CanSeek) Then
                R = Me.RetryPolicy
            Else
                R = AddressOf RetryPolicies.NoRetry
            End If
            R(Function() AnonymousMethod8(stream, originalPosition, blobProperties, overwrite, eTag, length, retval))

            Return retval
        End Function

        Private Function AnonymousMethod8(ByVal stream As Stream, ByVal originalPosition As Long, ByVal blobProperties As BlobProperties, ByVal overwrite As Boolean, ByVal eTag As String, ByVal length As Long, ByVal retval As Boolean) As Boolean
            If stream.CanSeek Then
                stream.Position = originalPosition
            End If
            retval = UploadData(blobProperties, stream, length, overwrite, eTag, New NameValueCollection())
            Return True
        End Function

        Private Function UploadData(ByVal blobProperties As BlobProperties, ByVal stream As Stream, ByVal length As Long, ByVal overwrite As Boolean, ByVal eTag As String, ByVal nvc As NameValueCollection) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim blobUri As Uri = Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, blobProperties.Name, Timeout, nvc, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequestForPutBlob(blobUri, StorageHttpConstants.HttpMethod.Put, blobProperties, length, overwrite, eTag)
            credentials.SignRequest(request, uriComponents)
            Dim retval As Boolean = False

            Try
                Using requestStream As Stream = request.GetRequestStream()
                    Utilities.CopyStream(stream, requestStream, length)
                    Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                        If response.StatusCode = HttpStatusCode.Created Then
                            retval = True
                        ElseIf (Not overwrite) AndAlso (response.StatusCode = HttpStatusCode.PreconditionFailed OrElse response.StatusCode = HttpStatusCode.NotModified) Then

                            retval = False
                        Else
                            retval = False
                            Utilities.ProcessUnexpectedStatusCode(response)
                        End If

                        blobProperties.LastModifiedTime = response.LastModified.ToUniversalTime()
                        blobProperties.ETag = response.Headers(StorageHttpConstants.HeaderNames.ETag)
                        response.Close()
                    End Using
                    requestStream.Close()
                End Using
            Catch ioe As IOException
                Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, Nothing, ioe)
            Catch te As System.TimeoutException
                Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during blob upload", HttpStatusCode.RequestTimeout, Nothing, te)
            Catch we As WebException
                If we.Response IsNot Nothing Then
                    Dim response As HttpWebResponse = CType(we.Response, HttpWebResponse)


                    If (response.StatusCode = HttpStatusCode.PreconditionFailed OrElse response.StatusCode = HttpStatusCode.NotModified) AndAlso ((Not overwrite) OrElse eTag IsNot Nothing) Then

                        retval = False
                        Return retval
                    End If
                End If
                Throw Utilities.TranslateWebException(we)
            End Try
            Return retval
        End Function

        Private Function GetBlobImpl(ByVal blobName As String, ByVal stream As Stream, ByVal oldETag As String, ByVal chunked As Boolean, <System.Runtime.InteropServices.Out()> ByRef modified As Boolean) As BlobProperties
            'We introduce this local variable since lambda expressions cannot contain use an out parameter
            Dim blobProperties As BlobProperties = Nothing
            Dim localModified As Boolean = True

            'If we are interested only in the blob properties (stream == null) or we are performing
            ' a chunked download we first obtain just the blob properties
            If stream Is Nothing OrElse chunked Then
                RetryPolicy.Invoke(Function() AnonymousMethod9(blobName, oldETag, blobProperties, localModified))
                modified = localModified
                If stream Is Nothing Then
                    Return blobProperties
                End If
            End If


            Dim R As RetryPolicy

            If (stream.CanSeek) Then
                R = Me.RetryPolicy
            Else
                R = AddressOf RetryPolicies.NoRetry
            End If
            Dim originalPosition As Long = If(stream.CanSeek, stream.Position, 0)
            If chunked AndAlso blobProperties IsNot Nothing AndAlso blobProperties.ContentLength > 0 Then
                'Chunked download. Obtain ranges of the blobs in 'BlockSize' chunks
                'Ensure that the If-Match <Etag>header is used on each request so
                'that we are assured that all data belongs to the single blob we
                'started downloading.
                Dim location As Long = 0
                Do While location < blobProperties.ContentLength
                    Dim nBytes As Long = Math.Min(blobProperties.ContentLength - location, StorageHttpConstants.BlobBlockConstants.BlockSize)
                    ' Set the position to rewind in case of a retry
                    R(Function() AnonymousMethod10(stream, originalPosition, location, blobName, oldETag, blobProperties, localModified, nBytes))
                    location += nBytes
                Loop
            Else
                'Non-chunked download. Obtain the entire blob in a single request.
                ' Set the position to rewind in case of a retry
                R(Function() AnonymousMethod11(stream, originalPosition, blobName, oldETag, blobProperties, localModified))
            End If
            modified = localModified
            Return blobProperties
        End Function

        Private Function AnonymousMethod9(ByVal blobName As String, ByVal oldETag As String, ByVal blobProperties As BlobProperties, ByVal localModified As Boolean) As Boolean
            blobProperties = DownloadData(blobName, Nothing, oldETag, Nothing, 0, 0, New NameValueCollection(), localModified)
            Return True
        End Function

        Private Function AnonymousMethod10(ByVal stream As Stream, ByVal originalPosition As Long, ByVal location As Long, ByVal blobName As String, ByVal oldETag As String, ByVal blobProperties As BlobProperties, ByVal localModified As Boolean, ByVal nBytes As Long) As Boolean
            If stream.CanSeek Then
                stream.Position = originalPosition + location
            End If
            DownloadData(blobName, stream, oldETag, blobProperties.ETag, location, nBytes, New NameValueCollection(), localModified)
            Return True
        End Function

        Private Function AnonymousMethod11(ByVal stream As Stream, ByVal originalPosition As Long, ByVal blobName As String, ByVal oldETag As String, ByVal blobProperties As BlobProperties, ByVal localModified As Boolean) As Boolean
            If stream.CanSeek Then
                stream.Position = originalPosition
            End If
            blobProperties = DownloadData(blobName, stream, oldETag, Nothing, 0, 0, New NameValueCollection(), localModified)
            Return True
        End Function

        Private Function SetBlobMetadataImpl(ByVal blobProperties As BlobProperties, ByVal eTag As String) As Boolean
            Dim retval As Boolean = False
            RetryPolicy.Invoke(Function() AnonymousMethod12(blobProperties, eTag, retval))
            Return retval
        End Function

        Private Function AnonymousMethod12(ByVal blobProperties As BlobProperties, ByVal eTag As String, ByRef retval As Boolean) As Boolean
            Dim queryParams As New NameValueCollection()
            queryParams.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.Metadata)
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, blobProperties.Name, Timeout, queryParams, uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Put, Timeout)
            If blobProperties.Metadata IsNot Nothing Then
                Utilities.AddMetadataHeaders(request, blobProperties.Metadata)
            End If
            If eTag IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.IfMatch, eTag)
            End If
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        retval = True
                    ElseIf (response.StatusCode = HttpStatusCode.PreconditionFailed OrElse response.StatusCode = HttpStatusCode.NotModified) AndAlso eTag IsNot Nothing Then
                        retval = False
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                        retval = False
                    End If
                    blobProperties.LastModifiedTime = response.LastModified.ToUniversalTime()
                    blobProperties.ETag = response.Headers(StorageHttpConstants.HeaderNames.ETag)
                    response.Close()
                End Using
            Catch ioe As IOException
                Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, ioe)
            Catch te As System.TimeoutException
                Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during blob upload", HttpStatusCode.RequestTimeout, te)
            Catch we As WebException
                If we.Response IsNot Nothing Then
                    Dim response As HttpWebResponse = CType(we.Response, HttpWebResponse)
                    If eTag IsNot Nothing AndAlso (response.StatusCode = HttpStatusCode.PreconditionFailed OrElse response.StatusCode = HttpStatusCode.NotModified) Then
                        retval = False
                    End If
                End If
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

        Private Function GetBlockList(ByVal blobName As String, ByVal eTag As String) As List(Of Integer)
            Dim blocks As List(Of Integer)
            Using blockListStream As New MemoryStream()
                Dim temp As Boolean = True
                Dim nvc As New NameValueCollection()
                nvc.Add(QueryParams.QueryParamComp, CompConstants.BlockList)

                DownloadData(blobName, blockListStream, eTag, Nothing, 0, 0, nvc, temp)
                Dim doc As New XmlDocument()
                blockListStream.Position = 0
                doc.Load(blockListStream)

                blocks = New List(Of Integer)()
                For Each block As XmlNode In doc.SelectNodes(XPathQueryHelper.BlockQuery)
                    blocks.Add(CInt(Fix(XPathQueryHelper.LoadSingleChildLongValue(block, XmlElementNames.Size, False))))
                Next block
            End Using
            Return blocks
        End Function

        ''' <summary>
        ''' Helper method used for getting blobs, ranges of blobs and blob properties.
        ''' </summary>
        ''' <param name="blobName">Name of the blob</param>
        ''' <param name="stream">The output stream to write blob data to. Can be null if only retrieving blob properties</param>
        ''' <param name="eTagIfNoneMatch">The If-None-Match header. Used to avoid downloading blob data if the blob has not changed</param>
        ''' <param name="eTagIfMatch">The If-Match header. Used to ensure that all chunks of the blob are of the same blob</param>
        ''' <param name="offset">The offset of the blob data to begin downloading from. Set to 0 to download all data.</param>
        ''' <param name="length">The length of the blob data to download. Set to 0 to download all data</param>
        ''' <param name="nvc">Query paramters to add to the request.</param>
        ''' <param name="localModified">Whether the blob had been modfied with respect to the <paramref name="eTagIfNoneMatch"/></param>
        ''' <returns></returns>
        Private Function DownloadData(ByVal blobName As String, ByVal stream As Stream, ByVal eTagIfNoneMatch As String, ByVal eTagIfMatch As String, ByVal offset As Long, ByVal length As Long, ByVal nvc As NameValueCollection, ByRef localModified As Boolean) As BlobProperties
            Dim uriComponents As ResourceUriComponents
            Dim blobUri As Uri = Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, blobName, Timeout, nvc, uriComponents)
            Dim httpMethod As String = (If(stream Is Nothing, StorageHttpConstants.HttpMethod.Head, StorageHttpConstants.HttpMethod.Get))
            Dim request As HttpWebRequest = CreateHttpRequestForGetBlob(blobUri, httpMethod, eTagIfNoneMatch, eTagIfMatch)

            If offset <> 0 OrElse length <> 0 Then
                'Use the blob storage custom header for range since the standard HttpWebRequest.AddRange 
                'accepts only 32 bit integers and so does not work for large blobs
                Dim rangeHeaderValue As String = String.Format(CultureInfo.InvariantCulture, HeaderValues.RangeHeaderFormat, offset, offset + length - 1)
                request.Headers.Add(HeaderNames.StorageRange, rangeHeaderValue)
            End If
            credentials.SignRequest(request, uriComponents)

            Dim blobProperties As BlobProperties

            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK OrElse response.StatusCode = HttpStatusCode.PartialContent Then
                        blobProperties = BlobPropertiesFromResponse(blobName, blobUri, response)
                        If stream IsNot Nothing Then
                            Using responseStream As Stream = response.GetResponseStream()
                                Dim bytesCopied As Long = Utilities.CopyStream(responseStream, stream)
                                If blobProperties.ContentLength > 0 AndAlso bytesCopied < blobProperties.ContentLength Then
                                    Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Unable to read complete data from server", HttpStatusCode.RequestTimeout, Nothing)
                                End If
                            End Using
                        End If
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                        Return Nothing
                    End If
                End Using
                Return blobProperties
            Catch ioe As IOException
                Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, ioe)
            Catch te As System.TimeoutException
                Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during blob upload", HttpStatusCode.RequestTimeout, te)
            Catch we As WebException
                If we.Response IsNot Nothing Then
                    Dim response As HttpWebResponse = CType(we.Response, HttpWebResponse)
                    If eTagIfNoneMatch IsNot Nothing AndAlso (response.StatusCode = HttpStatusCode.PreconditionFailed OrElse response.StatusCode = HttpStatusCode.NotModified) Then
                        localModified = False
                        blobProperties = Nothing
                        Return blobProperties
                    End If
                End If
                Throw Utilities.TranslateWebException(we)
            End Try
        End Function

        Private Shared Function MetadataFromHeaders(ByVal headers As WebHeaderCollection) As NameValueCollection
            Dim prefixLength As Integer = StorageHttpConstants.HeaderNames.PrefixForMetadata.Length
            Dim headerNames() As String = headers.AllKeys
            Dim metadataEntries As New NameValueCollection()
            For Each headerName As String In headerNames
                If headerName.StartsWith(StorageHttpConstants.HeaderNames.PrefixForMetadata, StringComparison.OrdinalIgnoreCase) Then
                    ' strip out the metadata prefix
                    metadataEntries.Add(headerName.Substring(prefixLength), headers(headerName))
                End If
            Next headerName
            Return metadataEntries
        End Function

        Private Shared Function BlobPropertiesFromResponse(ByVal blobName As String, ByVal blobUri As Uri, ByVal response As HttpWebResponse) As BlobProperties
            Dim blobProperties As New BlobProperties(blobName)
            blobProperties.Uri = blobUri
            blobProperties.ContentEncoding = response.Headers(StorageHttpConstants.HeaderNames.ContentEncoding)
            blobProperties.LastModifiedTime = response.LastModified.ToUniversalTime()
            blobProperties.ETag = response.Headers(StorageHttpConstants.HeaderNames.ETag)
            blobProperties.ContentLanguage = response.Headers(StorageHttpConstants.HeaderNames.ContentLanguage)
            blobProperties.ContentLength = response.ContentLength
            blobProperties.ContentType = response.ContentType

            Dim metadataEntries As NameValueCollection = MetadataFromHeaders(response.Headers)
            If metadataEntries.Count > 0 Then
                blobProperties.Metadata = metadataEntries
            End If

            Return blobProperties
        End Function

        Private Function ContainerPropertiesFromResponse(ByVal response As HttpWebResponse) As ContainerProperties
            Dim prop As New ContainerProperties(ContainerName)
            prop.LastModifiedTime = response.LastModified.ToUniversalTime()
            prop.ETag = response.Headers(StorageHttpConstants.HeaderNames.ETag)
            prop.Uri = containerUri_Renamed
            prop.Metadata = MetadataFromHeaders(response.Headers)
            Return prop
        End Function

        Private Function DeleteBlobImpl(ByVal name As String, ByVal eTag As String, <System.Runtime.InteropServices.Out()> ByRef modified As Boolean) As Boolean
            Dim retval As Boolean = False
            Dim localModified As Boolean = False
            RetryPolicy.Invoke(Function() AnonymousMethod13(name, eTag, retval, localModified))
            modified = localModified
            Return retval
        End Function

        Private Function AnonymousMethod13(ByVal name As String, ByVal eTag As String, ByRef retval As Boolean, ByRef localModified As Boolean) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim blobUri As Uri = Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, name, Timeout, New NameValueCollection(), uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(blobUri, StorageHttpConstants.HttpMethod.Delete, Timeout)
            If eTag IsNot Nothing Then
                request.Headers.Add(StorageHttpConstants.HeaderNames.IfMatch, eTag)
            End If
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK OrElse response.StatusCode = HttpStatusCode.Accepted Then
                        retval = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch ioe As IOException
                Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, ioe)
            Catch te As System.TimeoutException
                Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during blob delete", HttpStatusCode.RequestTimeout, te)
            Catch we As WebException
                If we.Response IsNot Nothing Then
                    Dim status As HttpStatusCode = (CType(we.Response, HttpWebResponse)).StatusCode
                    If status = HttpStatusCode.NotFound OrElse status = HttpStatusCode.Gone Then
                        localModified = True
                        retval = False
                    ElseIf status = HttpStatusCode.PreconditionFailed OrElse status = HttpStatusCode.NotModified Then
                        retval = False
                        localModified = True
                    Else
                        Throw Utilities.TranslateWebException(we)
                    End If
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

        Friend Class ListBlobsResult
            Friend Sub New(ByVal blobs As IEnumerable(Of BlobProperties), ByVal commonPrefixes As IEnumerable(Of String), ByVal nextMarker As String)
                blobs = blobs
                commonPrefixes = commonPrefixes
                nextMarker = nextMarker
            End Sub

            Private privateBlobs As IEnumerable(Of BlobProperties)
            Friend Property Blobs() As IEnumerable(Of BlobProperties)
                Get
                    Return privateBlobs
                End Get
                Private Set(ByVal value As IEnumerable(Of BlobProperties))
                    privateBlobs = value
                End Set
            End Property

            Private privateCommonPrefixes As IEnumerable(Of String)
            Friend Property CommonPrefixes() As IEnumerable(Of String)
                Get
                    Return privateCommonPrefixes
                End Get
                Private Set(ByVal value As IEnumerable(Of String))
                    privateCommonPrefixes = value
                End Set
            End Property

            Private privateNextMarker As String
            Friend Property NextMarker() As String
                Get
                    Return privateNextMarker
                End Get
                Private Set(ByVal value As String)
                    privateNextMarker = value
                End Set
            End Property
        End Class

        Private Function ListBlobsImpl(ByVal prefix As String, ByVal fromMarker As String, ByVal delimiter As String, ByVal maxCount As Integer) As ListBlobsResult
            Dim result As ListBlobsResult = Nothing

            RetryPolicy.Invoke(Function() AnonymousMethod14(prefix, fromMarker, delimiter, maxCount, result))
            Return result
        End Function

        Private Function AnonymousMethod14(ByVal prefix As String, ByVal fromMarker As String, ByVal delimiter As String, ByVal maxCount As Integer, ByRef result As ListBlobsResult) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim listBlobsUri As Uri = CreateRequestUriForListBlobs(prefix, fromMarker, delimiter, maxCount, uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(listBlobsUri, StorageHttpConstants.HttpMethod.Get, Timeout)
            credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        Using stream As Stream = response.GetResponseStream()
                            result = ListBlobsResultFromResponse(stream)
                        End Using
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                End Using
            Catch ioe As IOException
                Throw New StorageServerException(StorageErrorCode.TransportError, "The connection may be lost", Nothing, ioe)
            Catch te As System.TimeoutException
                Throw New StorageServerException(StorageErrorCode.ServiceTimeout, "Timeout during blob upload", HttpStatusCode.RequestTimeout, te)
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

        Private Function CreateRequestUriForListBlobs(ByVal prefix As String, ByVal fromMarker As String, ByVal delimiter As String, ByVal maxResults As Integer, <System.Runtime.InteropServices.Out()> ByRef uriComponents As ResourceUriComponents) As Uri
            Dim queryParams As NameValueCollection = BlobStorageRest.CreateRequestUriForListing(prefix, fromMarker, delimiter, maxResults)
            Return Utilities.CreateRequestUri(BaseUri, Me.UsePathStyleUris, AccountName, ContainerName, Nothing, Timeout, queryParams, uriComponents)
        End Function

        Private Shared Function ListBlobsResultFromResponse(ByVal responseBody As Stream) As ListBlobsResult
            Dim blobs As New List(Of BlobProperties)()
            Dim commonPrefixes As New List(Of String)()
            Dim nextMarker As String = Nothing

            Dim doc As New XmlDocument()
            Try
                doc.Load(responseBody)
            Catch xe As XmlException
                Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "The result of a ListBlobs operation could not be parsed", Nothing, xe)
            End Try

            ' Get the commonPrefixes
            Dim prefixNodes As XmlNodeList = doc.SelectNodes(XPathQueryHelper.CommonPrefixQuery)

            For Each prefixNode As XmlNode In prefixNodes
                Dim blobPrefix As String = XPathQueryHelper.LoadSingleChildStringValue(prefixNode, StorageHttpConstants.XmlElementNames.BlobPrefixName, False)
                commonPrefixes.Add(blobPrefix)
            Next prefixNode

            ' Get all the blobs returned as the listing results
            Dim blobNodes As XmlNodeList = doc.SelectNodes(XPathQueryHelper.BlobQuery)

            For Each blobNode As XmlNode In blobNodes
                Dim lastModified? As DateTime = XPathQueryHelper.LoadSingleChildDateTimeValue(blobNode, StorageHttpConstants.XmlElementNames.LastModified, False)

                Dim eTag As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.Etag, False)

                Dim contentType As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.ContentType, False)

                Dim contentEncoding As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.ContentEncoding, False)

                Dim contentLanguage As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.ContentLanguage, False)

                Dim blobSize? As Long = XPathQueryHelper.LoadSingleChildLongValue(blobNode, StorageHttpConstants.XmlElementNames.Size, False)

                Dim blobUrl As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.Url, True)

                Dim blobName As String = XPathQueryHelper.LoadSingleChildStringValue(blobNode, StorageHttpConstants.XmlElementNames.BlobName, True)

                Dim properties As New BlobProperties(blobName)
                properties.Uri = New Uri(blobUrl)
                If lastModified.HasValue Then
                    properties.LastModifiedTime = lastModified.Value
                End If
                properties.ContentEncoding = contentEncoding
                properties.ContentLanguage = contentLanguage
                properties.ETag = eTag
                properties.ContentLength = (If(Not blobSize.HasValue, 0, blobSize.Value))
                properties.ContentType = contentType

                blobs.Add(properties)
            Next blobNode

            ' Get the nextMarker
            Dim nextMarkerNode As XmlNode = doc.SelectSingleNode(XPathQueryHelper.NextMarkerQuery)
            If nextMarkerNode IsNot Nothing AndAlso nextMarkerNode.FirstChild IsNot Nothing Then
                nextMarker = nextMarkerNode.FirstChild.Value
            End If

            Return New ListBlobsResult(blobs, commonPrefixes, nextMarker)
        End Function

        Private containerUri_Renamed As Uri
        Private key() As Byte
        Private credentials As SharedKeyCredentials
    End Class

    ''' <summary>
    ''' Helper class for loading values from an XML segment
    ''' </summary>
    Friend NotInheritable Class XPathQueryHelper
        ' In general queries are like "//{0}/{1}/{2}" - using Join as it's more efficient than Format

        Friend Shared ReadOnly NextMarkerQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.EnumerationResults, StorageHttpConstants.XmlElementNames.NextMarker})

        Friend Shared ReadOnly ContainerQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.EnumerationResults, StorageHttpConstants.XmlElementNames.Containers, StorageHttpConstants.XmlElementNames.Container})

        Friend Shared ReadOnly BlobQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.EnumerationResults, StorageHttpConstants.XmlElementNames.Blobs, StorageHttpConstants.XmlElementNames.Blob})

        Friend Shared ReadOnly BlockQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.BlockList, StorageHttpConstants.XmlElementNames.Block})

        Friend Shared ReadOnly QueueListQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.EnumerationResults, StorageHttpConstants.XmlElementNames.Queues, StorageHttpConstants.XmlElementNames.Queue})

        Friend Shared ReadOnly MessagesListQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.QueueMessagesList, StorageHttpConstants.XmlElementNames.QueueMessage})

        Friend Shared ReadOnly CommonPrefixQuery As String = String.Join(StorageHttpConstants.ConstChars.Slash, New String() {"", "", StorageHttpConstants.XmlElementNames.EnumerationResults, StorageHttpConstants.XmlElementNames.Blobs, StorageHttpConstants.XmlElementNames.BlobPrefix})

        Private Sub New()
        End Sub
        Friend Shared Function LoadSingleChildDateTimeValue(ByVal node As XmlNode, ByVal childName As String, ByVal throwIfNotFound As Boolean) As DateTime?
            Dim childNode As XmlNode = node.SelectSingleNode(childName)

            If childNode IsNot Nothing AndAlso childNode.FirstChild IsNot Nothing Then
                Dim dateTime? As DateTime
                If (Not Utilities.TryGetDateTimeFromHttpString(childNode.FirstChild.Value, dateTime)) Then
                    Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "Date time value returned from server " & childNode.FirstChild.Value & " can't be parsed.", Nothing, Nothing)
                End If
                Return dateTime
            ElseIf (Not throwIfNotFound) Then
                Return Nothing
            Else
                Return Nothing
            End If
        End Function


        Friend Shared Function LoadSingleChildStringValue(ByVal node As XmlNode, ByVal childName As String, ByVal throwIfNotFound As Boolean) As String
            Dim childNode As XmlNode = node.SelectSingleNode(childName)

            If childNode IsNot Nothing AndAlso childNode.FirstChild IsNot Nothing Then
                Return childNode.FirstChild.Value
            ElseIf (Not throwIfNotFound) Then
                Return Nothing
            Else
                Return Nothing ' unnecessary since Fail will throw, but keeps the compiler happy
            End If
        End Function

        Friend Shared Function LoadSingleChildLongValue(ByVal node As XmlNode, ByVal childName As String, ByVal throwIfNotFound As Boolean) As Long?
            Dim childNode As XmlNode = node.SelectSingleNode(childName)

            If childNode IsNot Nothing AndAlso childNode.FirstChild IsNot Nothing Then
                Return Long.Parse(childNode.FirstChild.Value, CultureInfo.InvariantCulture)
            ElseIf (Not throwIfNotFound) Then
                Return Nothing
            Else
                Return Nothing ' unnecessary since Fail will throw, but keeps the compiler happy
            End If
        End Function
    End Class
End Namespace

