'
' <copyright file="RestQueue.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Web
Imports System.Net
Imports System.Collections.Specialized
Imports System.IO
Imports System.Xml
Imports System.Diagnostics
Imports System.Globalization
Imports System.Threading


Namespace Microsoft.Samples.ServiceHosting.StorageClient

	Friend Class QueueStorageRest
		Inherits QueueStorage
		Private _credentials As SharedKeyCredentials

		Friend Sub New(ByVal accountInfo As StorageAccountInfo)
			MyBase.New(accountInfo)
			Dim key() As Byte = Nothing
			If accountInfo.Base64Key IsNot Nothing Then
				key = Convert.FromBase64String(accountInfo.Base64Key)
			End If
			_credentials = New SharedKeyCredentials(accountInfo.AccountName, key)
		End Sub

		''' <summary>
		''' Get a reference to a Queue object with a specified name. This method does not make a call to
		''' the queue service.
		''' </summary>
		''' <param name="queueName">The name of the queue</param>
		''' <returns>A newly created queue object</returns>
		Public Overrides Function GetQueue(ByVal queueName As String) As MessageQueue
			Return New QueueRest(queueName, AccountInfo, Timeout, RetryPolicy)
		End Function

		Friend Class ListQueueResult
			Friend Sub New(ByVal names As IEnumerable(Of String), ByVal urls As IEnumerable(Of String), ByVal nextMarker As String)
				Names = names
				Urls = urls
				NextMarker = nextMarker
			End Sub

			Private privateNames As IEnumerable(Of String)
			Friend Property Names() As IEnumerable(Of String)
				Get
					Return privateNames
				End Get
				Private Set(ByVal value As IEnumerable(Of String))
					privateNames = value
				End Set
			End Property

			Private privateUrls As IEnumerable(Of String)
			Friend Property Urls() As IEnumerable(Of String)
				Get
					Return privateUrls
				End Get
				Private Set(ByVal value As IEnumerable(Of String))
					privateUrls = value
				End Set
			End Property

			Private privateNextMarker As String
			Friend Property NextMarker() As String
				Get
					Return privateNextMarker
				End Get
				Private Set(ByVal value As String)
					privateNextMarker = value
				End Set
			End Property
		End Class

		''' <summary>
		''' Lists all queues with a given prefix within an account.
		''' </summary>
		''' <param name="prefix"></param>
		''' <returns>The list of queue names.</returns>
		Public Overrides Function ListQueues(ByVal prefix As String) As IEnumerable(Of MessageQueue)
			Dim marker As String = ""
			Const maxResults As Integer = StorageHttpConstants.ListingConstants.MaxQueueListResults

            Dim queues As New List(Of MessageQueue)
			Do
				Dim result As ListQueueResult = ListQueuesImpl(prefix, marker, maxResults)
				If result Is Nothing Then
					marker = Nothing
				Else
					marker = result.NextMarker

					For Each name As String In result.Names
                        queues.Add(New QueueRest(name, AccountInfo, Me.Timeout, Me.RetryPolicy))
					Next name
				End If
            Loop While marker IsNot Nothing
            Return queues
		End Function

		''' <summary>
		''' Lists the queues within the account.
		''' </summary>
		''' <returns>A list of queues</returns>
		Private Function ListQueuesImpl(ByVal prefix As String, ByVal marker As String, ByVal maxResult As Integer) As ListQueueResult
			Dim result As ListQueueResult = Nothing

            RetryPolicy.Invoke(Function() AnonymousMethod1(prefix, marker, maxResult, result))

			Return result
		End Function
		
        Private Function AnonymousMethod1(ByVal prefix As String, ByVal marker As String, ByVal maxResult As Integer, ByRef result As ListQueueResult) As Boolean
            Dim col As New NameValueCollection()
            col.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.List)
            If (Not String.IsNullOrEmpty(prefix)) Then
                col.Add(StorageHttpConstants.QueryParams.QueryParamPrefix, prefix)
            End If
            If (Not String.IsNullOrEmpty(marker)) Then
                col.Add(StorageHttpConstants.QueryParams.QueryParamMarker, marker)
            End If
            col.Add(StorageHttpConstants.QueryParams.QueryParamMaxResults, maxResult.ToString(CultureInfo.InvariantCulture))
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = Utilities.CreateRequestUri(AccountInfo.BaseUri, AccountInfo.UsePathStyleUris, AccountInfo.AccountName, Nothing, Nothing, Timeout, col, uriComponents)
            Dim request As HttpWebRequest = Utilities.CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, Timeout)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        Using stream As Stream = response.GetResponseStream()
                            result = GetQueuesFromResponse(stream)
                            stream.Close()
                        End Using
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Private Shared Function GetQueuesFromResponse(ByVal stream As Stream) As ListQueueResult
			Dim result As ListQueueResult = Nothing
			Dim names As New List(Of String)()
			Dim urls As New List(Of String)()
			Dim nextMarker As String = Nothing

			Dim doc As New XmlDocument()
			Try
				doc.Load(stream)
			Catch xe As XmlException
				Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "The result of a ListQueue operation could not be parsed", Nothing, xe)
			End Try

			' get queue names and urls
			Dim queueNameNodes As XmlNodeList = doc.SelectNodes(XPathQueryHelper.QueueListQuery)
			For Each queueNameNode As XmlNode In queueNameNodes
				Dim queueName As String = XPathQueryHelper.LoadSingleChildStringValue(queueNameNode, StorageHttpConstants.XmlElementNames.QueueName, True)
				names.Add(queueName)
				Dim url As String = XPathQueryHelper.LoadSingleChildStringValue(queueNameNode, StorageHttpConstants.XmlElementNames.Url, True)
				urls.Add(url)
			Next queueNameNode

			' Get the nextMarker
			Dim nextMarkerNode As XmlNode = doc.SelectSingleNode(XPathQueryHelper.NextMarkerQuery)
			If nextMarkerNode IsNot Nothing AndAlso nextMarkerNode.FirstChild IsNot Nothing Then
				nextMarker = nextMarkerNode.FirstChild.Value
			End If
			If names.Count > 0 Then
				Debug.Assert(names.Count = urls.Count)
				result = New ListQueueResult(names, urls, nextMarker)
			End If
			Return result
		End Function
	End Class

	<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1001:TypesThatOwnDisposableFieldsShouldBeDisposable", Justification := "Disposable types are only used for automatic receiving of messages, which handle the disposal themselves.")> _
	Friend Class QueueRest
		Inherits MessageQueue
		#Region "Member variables and constructors"

		Private _queueUri As Uri
		Private _credentials As SharedKeyCredentials
		Private _pollInterval As Integer = DefaultPollInterval


		Friend Sub New(ByVal name As String, ByVal account As StorageAccountInfo, ByVal timeout As TimeSpan, ByVal retryPolicy As RetryPolicy)
			MyBase.New(name, account)
			Dim key() As Byte = Nothing
			If AccountInfo.Base64Key IsNot Nothing Then
				key = Convert.FromBase64String(AccountInfo.Base64Key)
			End If
			Dim uriComponents As New ResourceUriComponents(account.AccountName, name, Nothing)
			_credentials = New SharedKeyCredentials(AccountInfo.AccountName, key)
			_queueUri = HttpRequestAccessor.ConstructResourceUri(account.BaseUri, uriComponents, account.UsePathStyleUris)
			Timeout = timeout
			RetryPolicy = retryPolicy
		End Sub
		#End Region

		#Region "Public interface"

		Public Overrides ReadOnly Property QueueUri() As Uri
			Get
				Return _queueUri
			End Get
		End Property

		Public Overrides Function CreateQueue(<System.Runtime.InteropServices.Out()> ByRef queueAlreadyExists As Boolean) As Boolean
			Dim result As Boolean = False
			queueAlreadyExists = False
			' cannot use ref or out parameters in the retry expression below
			Dim exists As Boolean = False

							' as an addition we could parse the result and retrieve
							' queue properties at this point
            RetryPolicy.Invoke(Function() AnonymousMethod2(exists, result))

			queueAlreadyExists = exists
			Return result
		End Function
		
        Private Function AnonymousMethod2(ByVal exists As Boolean, ByRef result As Boolean) As Boolean
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(Nothing, New NameValueCollection(), False, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Put)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.Created Then
                        exists = False
                        result = True
                    ElseIf response.StatusCode = HttpStatusCode.NoContent Then
                        exists = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Conflict Then
                    exists = True
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

		Public Overrides Function DeleteQueue() As Boolean
			Dim result As Boolean = False

            RetryPolicy.Invoke(Function() AnonymousMethod3(result))

			Return result
		End Function
		
        Private Function AnonymousMethod3(ByRef result As Boolean) As Boolean
            Dim col As New NameValueCollection()
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(Nothing, col, False, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Delete)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.NoContent Then
                        result = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                If we.Response IsNot Nothing AndAlso ((CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.NotFound OrElse (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.PreconditionFailed OrElse (CType(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.Conflict) Then
                    result = False
                Else
                    Throw Utilities.TranslateWebException(we)
                End If
            End Try
            Return True
        End Function

		Public Overrides Function SetProperties(ByVal properties As QueueProperties) As Boolean
			If properties Is Nothing Then
				Throw New ArgumentNullException("properties")
			End If

			Dim result As Boolean = False

            RetryPolicy.Invoke(Function() AnonymousMethod4(properties, result))

			Return result
		End Function
		
        Private Function AnonymousMethod4(ByVal properties As QueueProperties, ByRef result As Boolean) As Boolean
            Dim col As New NameValueCollection()
            col.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.Metadata)
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(Nothing, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Put, properties.Metadata)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.NoContent Then
                        result = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Public Overrides Function GetProperties() As QueueProperties
			Dim result As QueueProperties = Nothing

            RetryPolicy.Invoke(Function() AnonymousMethod5(result))

			Return result
		End Function
		
        Private Function AnonymousMethod5(ByRef result As QueueProperties) As Boolean
            Dim col As New NameValueCollection()
            col.Add(StorageHttpConstants.QueryParams.QueryParamComp, StorageHttpConstants.CompConstants.Metadata)
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(Nothing, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, Nothing)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        result = GetPropertiesFromHeaders(response)
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Public Overrides Function ApproximateCount() As Integer
			Dim props As QueueProperties = GetProperties()
			Return props.ApproximateMessageCount
		End Function

		' getting and putting messages

		Public Overrides Function PutMessage(ByVal msg As Message) As Boolean
			Return PutMessage(msg, -1)
		End Function

		Public Overrides Function PutMessage(ByVal msg As Message, ByVal timeToLiveInSeconds As Integer) As Boolean
			If timeToLiveInSeconds < -1 Then
				Throw New ArgumentException("ttl parameter must be equal or larger than 0.")
			ElseIf timeToLiveInSeconds > Message.MaxTimeToLive Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "timeToLiveHours parameter must be smaller or equal than {0}, which is 7 days in hours.", Message.MaxTimeToLive))
			End If
			If msg Is Nothing OrElse msg.ContentAsBytes() Is Nothing Then
				Throw New ArgumentNullException("msg")
			End If
			If Convert.ToBase64String(msg.ContentAsBytes()).Length > Message.MaxMessageSize Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "Messages cannot be larger than {0} bytes.", Message.MaxMessageSize))
			End If

			Dim result As Boolean = False

            RetryPolicy.Invoke(Function() AnonymousMethod6(timeToLiveInSeconds, msg, result))

			Return result
		End Function
		
        Private Function AnonymousMethod6(ByVal timeToLiveInSeconds As Integer, ByVal msg As Message, ByRef result As Boolean) As Boolean
            Dim col As New NameValueCollection()
            If timeToLiveInSeconds <> -1 Then
                col.Add(StorageHttpConstants.RequestParams.MessageTtl, timeToLiveInSeconds.ToString(CultureInfo.InvariantCulture))
            End If
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(StorageHttpConstants.RequestParams.Messages, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Post, Nothing)
            Dim len As Integer
            Dim body() As Byte = msg.GetContentXMLRepresentation(len)
            request.ContentLength = len
            _credentials.SignRequest(request, uriComponents)
            Try
                Using requestStream As Stream = request.GetRequestStream()
                    requestStream.Write(body, 0, body.Length)
                    requestStream.Close()
                    Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                        If response.StatusCode = HttpStatusCode.Created Then
                            result = True
                        Else
                            Utilities.ProcessUnexpectedStatusCode(response)
                        End If
                        response.Close()
                    End Using
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Public Overrides Function GetMessage() As Message
			Dim result As IEnumerable(Of Message) = GetMessages(1)
			If result Is Nothing OrElse result.Count() = 0 Then
				Return Nothing
			End If
			Return result.First()
		End Function

		Public Overrides Function GetMessage(ByVal visibilityTimeoutInSeconds As Integer) As Message
			Dim result As IEnumerable(Of Message) = GetMessages(1, visibilityTimeoutInSeconds)
			If result Is Nothing OrElse result.Count() = 0 Then
				Return Nothing
			End If
			Return result.First()
		End Function

		Public Overrides Function GetMessages(ByVal numberOfMessages As Integer) As IEnumerable(Of Message)
			Return GetMessages(numberOfMessages, -1)
		End Function

		Public Overrides Function GetMessages(ByVal numberOfMessages As Integer, ByVal visibilityTimeout As Integer) As IEnumerable(Of Message)
			Return InternalGet(numberOfMessages, visibilityTimeout, False)
		End Function

		Public Overrides Function PeekMessage() As Message
			Dim result As IEnumerable(Of Message) = PeekMessages(1)
			If result Is Nothing OrElse result.Count() = 0 Then
				Return Nothing
			End If
			Return result.First()
		End Function

		Public Overrides Function PeekMessages(ByVal numberOfMessages As Integer) As IEnumerable(Of Message)
			Return InternalGet(numberOfMessages, -1, True)
		End Function

		' deleting messages
		Public Overrides Function DeleteMessage(ByVal msg As Message) As Boolean
			If msg.PopReceipt Is Nothing Then
				Throw New ArgumentException("No PopReceipt for the given message!")
			End If

			Dim result As Boolean = False

            RetryPolicy.Invoke(Function() AnonymousMethod7(msg, result))

			Return result
		End Function
		
        Private Function AnonymousMethod7(ByVal msg As Message, ByRef result As Boolean) As Boolean
            Dim col As New NameValueCollection()
            col.Add(StorageHttpConstants.RequestParams.PopReceipt, msg.PopReceipt.ToString())
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(StorageHttpConstants.RequestParams.Messages & "/" & msg.Id, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Delete, Nothing)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.NoContent Then
                        result = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Public Overrides Function Clear() As Boolean
			Dim result As Boolean = False

            RetryPolicy.Invoke(Function() AnonymousMethod8(result))

			Return result
		End Function
		
        Private Function AnonymousMethod8(ByRef result As Boolean) As Boolean
            Dim col As New NameValueCollection()
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(StorageHttpConstants.RequestParams.Messages, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Delete, Nothing)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.NoContent Then
                        result = True
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		' automatic receiving of messages
		Public Overrides Property PollInterval() As Integer
			Get
				Return _pollInterval
			End Get
			Set(ByVal value As Integer)
				If value < 0 Then
					Throw New ArgumentException("The send threshold must be a positive value.")
				End If
				If _run Then
					Throw New ArgumentException("You cannot set the poll interval while the receive thread is running")
				End If
				_pollInterval = value
			End Set
		End Property

		Private _evStarted As AutoResetEvent
		Private _evStopped As AutoResetEvent
		Private _evQuit As AutoResetEvent
		Private _run As Boolean
		Private _receiveThread As Thread
		Private _internalPollInterval As Integer


		Private Sub PeriodicReceive()
			Dim msg As Message

			_evStarted.Set()
			_internalPollInterval = PollInterval
			Do While Not _evQuit.WaitOne(_internalPollInterval, False)
				' time is up, so we get the message and continue
				msg = GetMessage()
				If msg IsNot Nothing Then
                    RaiseEvent MessageReceived(Me, New MessageReceivedEventArgs(msg))
					' continue receiving fast until we get no message
					_internalPollInterval = 10
				Else
					' we got no message, so we can fall back to the normal speed
					_internalPollInterval = PollInterval
				End If
			Loop
			_evStopped.Set()
		End Sub

		Public Overrides Function StartReceiving() As Boolean
			SyncLock Me
				If _run Then
					Return True
				End If
				_run = True
			End SyncLock
			If _evStarted Is Nothing Then
				_evStarted = New AutoResetEvent(False)
			End If
			If _evStopped Is Nothing Then
				_evStopped = New AutoResetEvent(False)
			End If
			If _evQuit Is Nothing Then
				_evQuit = New AutoResetEvent(False)
			End If
			_receiveThread = New Thread(New ThreadStart(AddressOf Me.PeriodicReceive))
			_receiveThread.Start()
			If (Not _evStarted.WaitOne(10000, False)) Then
				_receiveThread.Abort()
				CloseEvents()
				_run = False
				Return False
			End If
			Return True
		End Function

		Public Overrides Sub StopReceiving()
			_evQuit.Set()
			If (Not _evStopped.WaitOne(10000, False)) Then
				_receiveThread.Abort()
			End If
			CloseEvents()
			_run = False
		End Sub

		Private Sub CloseEvents()
			If _evStarted IsNot Nothing Then
				_evStarted.Close()
			End If
			If _evStopped Is Nothing Then
				_evStopped.Close()
			End If
			If _evQuit Is Nothing Then
				_evQuit.Close()
			End If
		End Sub

        Public Shadows Event MessageReceived As MessageReceivedEventHandler

		#End Region

		#Region "Helper methods"

		Private Shared Function GetPropertiesFromHeaders(ByVal response As HttpWebResponse) As QueueProperties
			Dim properties As New QueueProperties()
			Dim prefixLength As Integer = StorageHttpConstants.HeaderNames.PrefixForMetadata.Length
			For Each key As String In response.Headers.AllKeys
				If key.Equals(StorageHttpConstants.HeaderNames.ApproximateMessagesCount, StringComparison.OrdinalIgnoreCase) Then
					properties.ApproximateMessageCount = Convert.ToInt32(response.Headers(key), CultureInfo.InvariantCulture)
				ElseIf key.StartsWith(StorageHttpConstants.HeaderNames.PrefixForMetadata, StringComparison.OrdinalIgnoreCase) Then
					If properties.Metadata Is Nothing Then
						properties.Metadata = New NameValueCollection()
					End If
					properties.Metadata.Add(key.Substring(prefixLength), response.Headers(key))
				End If
			Next key
			Return properties
		End Function


		Private Function InternalGet(ByVal numberOfMessages As Integer, ByVal visibilityTimeout As Integer, ByVal peekOnly As Boolean) As IEnumerable(Of Message)
			If peekOnly AndAlso visibilityTimeout <> -1 Then
				Throw New ArgumentException("A peek operation does not change the visibility of messages", "visibilityTimeout")
			End If
			If numberOfMessages < 1 Then
				Throw New ArgumentException("numberOfMessages must be a positive integer", "numberOfMessages")
			End If
			If visibilityTimeout < -1 Then
				Throw New ArgumentException("Visibility Timeout must be 0 or a positive integer", "visibilityTimeout")
			End If

			Dim result As IEnumerable(Of Message) = Nothing

            RetryPolicy.Invoke(Function() AnonymousMethod9(numberOfMessages, visibilityTimeout, peekOnly, result))


			Return result
		End Function
		
        Private Function AnonymousMethod9(ByVal numberOfMessages As Integer, ByVal visibilityTimeout As Integer, ByVal peekOnly As Boolean, ByRef result As IEnumerable(Of Message)) As Boolean
            Dim col As New NameValueCollection()
            col.Add(StorageHttpConstants.RequestParams.NumOfMessages, numberOfMessages.ToString(CultureInfo.InvariantCulture))
            If visibilityTimeout <> -1 Then
                col.Add(StorageHttpConstants.RequestParams.VisibilityTimeout, visibilityTimeout.ToString(CultureInfo.InvariantCulture))
            End If
            If peekOnly Then
                col.Add(StorageHttpConstants.RequestParams.PeekOnly, peekOnly.ToString(CultureInfo.InvariantCulture))
            End If
            Dim uriComponents As ResourceUriComponents
            Dim uri As Uri = CreateRequestUri(StorageHttpConstants.RequestParams.Messages, col, uriComponents)
            Dim request As HttpWebRequest = CreateHttpRequest(uri, StorageHttpConstants.HttpMethod.Get, Nothing)
            _credentials.SignRequest(request, uriComponents)
            Try
                Using response As HttpWebResponse = CType(request.GetResponse(), HttpWebResponse)
                    If response.StatusCode = HttpStatusCode.OK Then
                        Using stream As Stream = response.GetResponseStream()
                            result = GetMessageFromResponse(stream)
                            stream.Close()
                        End Using
                    Else
                        Utilities.ProcessUnexpectedStatusCode(response)
                    End If
                    response.Close()
                End Using
            Catch we As WebException
                Throw Utilities.TranslateWebException(we)
            End Try
            Return True
        End Function

		Private Shared Function GetMessageFromResponse(ByVal stream As Stream) As IEnumerable(Of Message)
			Dim result As List(Of Message) = Nothing
			Dim msg As Message

			Dim doc As New XmlDocument()
			Try
				doc.Load(stream)
			Catch xe As XmlException
				Throw New StorageServerException(StorageErrorCode.ServiceBadResponse, "The result of a get message opertation could not be parsed", Nothing, xe)
			End Try

			Dim messagesNodes As XmlNodeList = doc.SelectNodes(XPathQueryHelper.MessagesListQuery)
			If messagesNodes.Count > 0 Then
				result = New List(Of Message)()
			End If
			For Each messageNode As XmlNode In messagesNodes
				msg = New Message()
				msg.Id = messageNode.SelectSingleNode(StorageHttpConstants.XmlElementNames.MessageId).FirstChild.Value.Trim()
				Debug.Assert(msg.Id IsNot Nothing)
				If messageNode.SelectSingleNode(StorageHttpConstants.XmlElementNames.PopReceipt) IsNot Nothing Then
					msg.PopReceipt = messageNode.SelectSingleNode(StorageHttpConstants.XmlElementNames.PopReceipt).FirstChild.Value.Trim()
					Debug.Assert(msg.PopReceipt IsNot Nothing)
				End If
				msg.InsertionTime = XPathQueryHelper.LoadSingleChildDateTimeValue(messageNode, StorageHttpConstants.XmlElementNames.InsertionTime, False).Value
				msg.ExpirationTime = XPathQueryHelper.LoadSingleChildDateTimeValue(messageNode, StorageHttpConstants.XmlElementNames.ExpirationTime, False).Value
				If XPathQueryHelper.LoadSingleChildDateTimeValue(messageNode, StorageHttpConstants.XmlElementNames.TimeNextVisible, False) IsNot Nothing Then
					msg.TimeNextVisible = XPathQueryHelper.LoadSingleChildDateTimeValue(messageNode, StorageHttpConstants.XmlElementNames.TimeNextVisible, False).Value
				End If
				msg.SetContentFromBase64String(XPathQueryHelper.LoadSingleChildStringValue(messageNode, StorageHttpConstants.XmlElementNames.MessageText, False))
				result.Add(msg)
			Next messageNode
			Return result.AsEnumerable()
		End Function

		Private Function CreateHttpRequest(ByVal uri As Uri, ByVal httpMethod As String) As HttpWebRequest
			Return CreateHttpRequest(uri, httpMethod, Nothing)
		End Function

		Private Function CreateHttpRequest(ByVal uri As Uri, ByVal httpMethod As String, ByVal metadata As NameValueCollection) As HttpWebRequest
			Dim request As HttpWebRequest = CType(HttpWebRequest.Create(uri), HttpWebRequest)
			request.Timeout = CInt(Fix(Timeout.TotalMilliseconds))
			request.ReadWriteTimeout = CInt(Fix(Timeout.TotalMilliseconds))
			request.Method = httpMethod
			request.ContentLength = 0
			request.Headers.Add(StorageHttpConstants.HeaderNames.StorageDateTime, Utilities.ConvertDateTimeToHttpString(DateTime.UtcNow))
			If metadata IsNot Nothing Then
				Utilities.AddMetadataHeaders(request, metadata)
			End If
			Return request
		End Function

		Private Function CreateRequestUri(ByVal uriSuffix As String, ByVal queryParameters As NameValueCollection, <System.Runtime.InteropServices.Out()> ByRef uriComponents As ResourceUriComponents) As Uri
			Return CreateRequestUri(uriSuffix, queryParameters, False, uriComponents)
		End Function

		Private Function CreateRequestUri(ByVal uriSuffix As String, ByVal queryParameters As NameValueCollection, ByVal accountOperation As Boolean, <System.Runtime.InteropServices.Out()> ByRef uriComponents As ResourceUriComponents) As Uri
			Return Utilities.CreateRequestUri(Me.AccountInfo.BaseUri, Me.AccountInfo.UsePathStyleUris, Me.AccountInfo.AccountName,If(accountOperation, Nothing, Me.Name), uriSuffix, Me.Timeout, queryParameters, uriComponents)
		End Function

		#End Region

	End Class
End Namespace
