'
' <copyright file="StorageAccountInfo.cs" company="Microsoft">
'     Copyright (c) Microsoft Corporation.  All rights reserved.
' </copyright>
'

Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Globalization
Imports System.Configuration
Imports System.Collections.Specialized
Imports System.Diagnostics
Imports Microsoft.ServiceHosting.ServiceRuntime

Namespace Microsoft.Samples.ServiceHosting.StorageClient
	''' <summary>
	''' Objects of this class encapsulate information about a storage account and endpoint configuration. 
	''' Associated with a storage account is the account name, the base URI of the account and a shared key. 
	''' </summary>
	Public Class StorageAccountInfo

		''' <summary>
		''' The default configuration string in configuration files for setting the queue storage endpoint.
		''' </summary>
		Public Shared ReadOnly DefaultQueueStorageEndpointConfigurationString As String = "QueueStorageEndpoint"

		''' <summary>
		''' The default configuration string in configuration files for setting the blob storage endpoint.
		''' </summary>
		Public Shared ReadOnly DefaultBlobStorageEndpointConfigurationString As String = "BlobStorageEndpoint"

		''' <summary>
		''' The default configuration string in configuration files for setting the table storage endpoint.
		''' </summary>
		Public Shared ReadOnly DefaultTableStorageEndpointConfigurationString As String = "TableStorageEndpoint"

		''' <summary>
		''' The default configuration string in configuration files for setting the storage account name.
		''' </summary>
		Public Shared ReadOnly DefaultAccountNameConfigurationString As String = "AccountName"

		''' <summary>
		''' The default configuration string in configuration files for setting the shared key associated with a storage account.
		''' </summary>
		Public Shared ReadOnly DefaultAccountSharedKeyConfigurationString As String = "AccountSharedKey"

		''' <summary>
		''' The default configuration string in configuration files for setting the UsePathStyleUris option.
		''' </summary>
		Public Shared ReadOnly DefaultUsePathStyleUrisConfigurationString As String = "UsePathStyleUris"

		''' <summary>
		''' The default prefix string in application config and Web.config files to indicate that this setting should be looked up 
		''' in the fabric's configuration system.
		''' </summary>
		Public Shared ReadOnly CSConfigStringPrefix As String = "CSConfigName"

		Private _usePathStyleUris? As Boolean

		''' <summary>
		''' Constructor for creating account info objects.
		''' </summary>
		''' <param name="baseUri">The account's base URI.</param>
		''' <param name="usePathStyleUris">If true, path-style URIs (http://baseuri/accountname/containername/objectname) are used.
		''' If false host-style URIs (http://accountname.baseuri/containername/objectname) are used,
		''' where baseuri is the URI of the service..
		''' If null, the choice is made automatically: path-style URIs if host name part of base URI is an 
		''' IP addres, host-style otherwise.</param>
		''' <param name="accountName">The account name.</param>
		''' <param name="base64Key">The account's shared key.</param>
		Public Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String)
			Me.New(baseUri, usePathStyleUris, accountName, base64Key, False)
		End Sub

		''' <summary>
		''' Constructor for creating account info objects.
		''' </summary>
		''' <param name="baseUri">The account's base URI.</param>
		''' <param name="usePathStyleUris">If true, path-style URIs (http://baseuri/accountname/containername/objectname) are used.
		''' If false host-style URIs (http://accountname.baseuri/containername/objectname) are used,
		''' where baseuri is the URI of the service.
		''' If null, the choice is made automatically: path-style URIs if host name part of base URI is an 
		''' IP addres, host-style otherwise.</param>
		''' <param name="accountName">The account name.</param>
		''' <param name="base64Key">The account's shared key.</param>
		''' <param name="allowIncompleteSettings">true if it shall be allowed to only set parts of the StorageAccountInfo properties.</param>
		<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1308:NormalizeStringsToUppercase")> _
		Public Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String, ByVal allowIncompleteSettings As Boolean)
			If baseUri Is Nothing AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentNullException("baseUri")
			End If
			If String.IsNullOrEmpty(base64Key) AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentNullException("base64Key")
			End If
			If baseUri IsNot Nothing Then
				Dim newAccountName As String = Nothing
				Dim newBaseUri As Uri = Nothing
				If IsStandardStorageEndpoint(baseUri, newAccountName, newBaseUri) Then
					If (Not String.IsNullOrEmpty(newAccountName)) AndAlso (Not String.IsNullOrEmpty(accountName)) AndAlso String.Compare(accountName, newAccountName, StringComparison.Ordinal) <> 0 Then
						Throw New ArgumentException("The configured base URI " & baseUri.AbsoluteUri & " for the storage service is incorrect. " & "The configured account name " & accountName & " must match the account name " & newAccountName & " as specified in the storage service base URI." & GeneralAccountConfigurationExceptionString)
					End If
					Debug.Assert((newBaseUri Is Nothing AndAlso newAccountName Is Nothing) OrElse (newBaseUri IsNot Nothing AndAlso newAccountName IsNot Nothing))
					If newAccountName IsNot Nothing AndAlso newBaseUri IsNot Nothing Then
						accountName = newAccountName
						baseUri = newBaseUri
					End If
				End If
			End If
			If String.IsNullOrEmpty(accountName) AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentNullException("accountName")
			End If
			If (Not String.IsNullOrEmpty(accountName)) AndAlso accountName.ToLowerInvariant() <> accountName Then
				Throw New ArgumentException("The account name must not contain upper-case letters. " & "The account name is the first part of the URL for accessing the storage services as presented to you by the portal or " & "the predefined storage account name when using the development storage tool. " & GeneralAccountConfigurationExceptionString)
			End If

            Me.BaseUri = baseUri
            Me.AccountName = accountName
            Me.Base64Key = base64Key
			If Not usePathStyleUris.HasValue AndAlso baseUri Is Nothing AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentException("Cannot determine setting from empty URI.")
			ElseIf Not usePathStyleUris.HasValue Then
				If baseUri IsNot Nothing Then
					_usePathStyleUris = Utilities.StringIsIPAddress(baseUri.Host)
				Else
					_usePathStyleUris = Nothing
				End If
			Else
				Me.UsePathStyleUris = usePathStyleUris.Value
			End If
		End Sub

		''' <summary>
		''' The base URI of the account.
		''' </summary>
		Private privateBaseUri As Uri
		Public Property BaseUri() As Uri
			Get
				Return privateBaseUri
			End Get
			Set(ByVal value As Uri)
				privateBaseUri = value
			End Set
		End Property

		''' <summary>
		''' The account name.
		''' </summary>
		Private privateAccountName As String
		Public Property AccountName() As String
			Get
				Return privateAccountName
			End Get
			Set(ByVal value As String)
				privateAccountName = value
			End Set
		End Property

		''' <summary>
		''' The account's key.
		''' </summary>
		Private privateBase64Key As String
		Public Property Base64Key() As String
			Get
				Return privateBase64Key
			End Get
			Set(ByVal value As String)
				privateBase64Key = value
			End Set
		End Property

		''' <summary>
		''' If set, returns the UsePathStyleUris properties. If the property has not been explicitly set, 
		''' the implementation tries to derive the correct value from the base URI.
		''' </summary>
		Public Property UsePathStyleUris() As Boolean
			Get
				If Not _usePathStyleUris.HasValue Then
					If BaseUri Is Nothing Then
						Return False
					Else
						Return Utilities.StringIsIPAddress(BaseUri.Host)
					End If
				Else
					Return _usePathStyleUris.Value
				End If
			End Get
			Set(ByVal value As Boolean)
				_usePathStyleUris = value
			End Set
		End Property

		''' <summary>
		''' Retrieves account settings for the queue service from the default settings. 
		''' </summary>
		Public Shared Function GetDefaultQueueStorageAccountFromConfiguration(ByVal allowIncompleteSettings As Boolean) As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultQueueStorageEndpointConfigurationString, allowIncompleteSettings)
		End Function

		''' <summary>
		''' Retrieves account settings for the queue service from the default settings. 
		''' Throws an exception in case of incomplete settings.
		''' </summary>
		Public Shared Function GetDefaultQueueStorageAccountFromConfiguration() As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultQueueStorageEndpointConfigurationString, False)
		End Function

		''' <summary>
		''' Retrieves account settings for the table service from the default settings. 
		''' </summary>
		Public Shared Function GetDefaultTableStorageAccountFromConfiguration(ByVal allowIncompleteSettings As Boolean) As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultTableStorageEndpointConfigurationString, allowIncompleteSettings)
		End Function

		''' <summary>
		''' Retrieves account settings for the table service from the default settings. 
		''' Throws an exception in case of incomplete settings.
		''' </summary>
		Public Shared Function GetDefaultTableStorageAccountFromConfiguration() As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultTableStorageEndpointConfigurationString, False)
		End Function

		''' <summary>
		''' Retrieves account settings for the blob service from the default settings. 
		''' </summary>
		Public Shared Function GetDefaultBlobStorageAccountFromConfiguration(ByVal allowIncompleteSettings As Boolean) As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultBlobStorageEndpointConfigurationString, allowIncompleteSettings)
		End Function

		''' <summary>
		''' Retrieves account settings for the blob service from the default settings. 
		''' Throws an exception in case of incomplete settings.
		''' </summary>
		Public Shared Function GetDefaultBlobStorageAccountFromConfiguration() As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultBlobStorageEndpointConfigurationString, False)
		End Function

		''' <summary>
		''' Gets settings from default configuration names except for the endpoint configuration string.
		''' </summary>
		Public Shared Function GetAccountInfoFromConfiguration(ByVal endpointConfiguration As String, ByVal allowIncompleteSettings As Boolean) As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultAccountNameConfigurationString, DefaultAccountSharedKeyConfigurationString, endpointConfiguration, DefaultUsePathStyleUrisConfigurationString, allowIncompleteSettings)
		End Function

		''' <summary>
		''' Gets settings from default configuration names except for the endpoint configuration string. Throws an exception 
		''' in the case of incomplete settings.
		''' </summary>
		Public Shared Function GetAccountInfoFromConfiguration(ByVal endpointConfiguration As String) As StorageAccountInfo
			Return GetAccountInfoFromConfiguration(DefaultAccountNameConfigurationString, DefaultAccountSharedKeyConfigurationString, endpointConfiguration, DefaultUsePathStyleUrisConfigurationString, False)
		End Function

		''' <summary>
		''' Gets a configuration setting from application settings in the Web.config or App.config file. 
		''' When running in a hosted environment, configuration settings are read from .cscfg
		''' files.
		''' </summary>
		Public Shared Function GetConfigurationSetting(ByVal configurationSetting As String, ByVal defaultValue As String, ByVal throwIfNull As Boolean) As String
			If String.IsNullOrEmpty(configurationSetting) Then
				Throw New ArgumentException("configurationSetting cannot be empty or null", "configurationSetting")
			End If

			Dim ret As String = Nothing

			' first, try to read from appsettings
			ret = TryGetAppSetting(configurationSetting)

			' settings in the csc file overload settings in Web.config
			If RoleManager.IsRoleManagerRunning Then
				Dim cscRet As String = TryGetConfigurationSetting(configurationSetting)
				If (Not String.IsNullOrEmpty(cscRet)) Then
					ret = cscRet
				End If

				' if there is a csc config name in the app settings, this config name even overloads the 
				' setting we have right now
				Dim refWebRet As String = TryGetAppSetting(StorageAccountInfo.CSConfigStringPrefix & configurationSetting)
				If (Not String.IsNullOrEmpty(refWebRet)) Then
					cscRet = TryGetConfigurationSetting(refWebRet)
					If (Not String.IsNullOrEmpty(cscRet)) Then
						ret = cscRet
					End If
				End If
			End If

			' if we could not retrieve any configuration string set return value to the default value
			If String.IsNullOrEmpty(ret) AndAlso defaultValue IsNot Nothing Then
				ret = defaultValue
			End If

			If String.IsNullOrEmpty(ret) AndAlso throwIfNull Then
				Throw New ConfigurationErrorsException(String.Format(CultureInfo.InvariantCulture, "Cannot find configuration string {0}.", configurationSetting))
			End If
			Return ret
		End Function

		''' <summary>
		''' Retrieves account information settings from configuration settings. First, the implementation checks for 
		''' settings in an application config section of an app.config or Web.config file. These values are overwritten 
		''' if the same settings appear in a .csdef file.
		''' The implementation also supports indirect settings. In this case, indirect settings overwrite all other settings.
		''' </summary>        
		''' <param name="accountNameConfiguration">Configuration string for the account name.</param>
		''' <param name="accountSharedKeyConfiguration">Configuration string for the key.</param>
		''' <param name="endpointConfiguration">Configuration string for the endpoint.</param>
		''' <param name="usePathStyleUrisConfiguration">Configuration string for the path style.</param>
		''' <param name="allowIncompleteSettings">If false, an exception is thrown if not all settings are available.</param>
		''' <returns>StorageAccountInfo object containing the retrieved settings.</returns>        
		Public Shared Function GetAccountInfoFromConfiguration(ByVal accountNameConfiguration As String, ByVal accountSharedKeyConfiguration As String, ByVal endpointConfiguration As String, ByVal usePathStyleUrisConfiguration As String, ByVal allowIncompleteSettings As Boolean) As StorageAccountInfo
			If String.IsNullOrEmpty(endpointConfiguration) Then
				Throw New ArgumentException("Endpoint configuration is missing", "endpointConfiguration")
			End If
			Dim endpoint As String = Nothing
			Dim name As String = Nothing
			Dim key As String = Nothing
			Dim pathStyle As String = Nothing

			name = TryGetAppSetting(accountNameConfiguration)
			key = TryGetAppSetting(accountSharedKeyConfiguration)
			endpoint = TryGetAppSetting(endpointConfiguration)
			pathStyle = TryGetAppSetting(usePathStyleUrisConfiguration)


			' settings in the csc file overload settings in Web.config
			If RoleManager.IsRoleManagerRunning Then
				' get config settings from the csc file
				Dim cscName As String = TryGetConfigurationSetting(accountNameConfiguration)
				If (Not String.IsNullOrEmpty(cscName)) Then
					name = cscName
				End If
				Dim cscKey As String = TryGetConfigurationSetting(accountSharedKeyConfiguration)
				If (Not String.IsNullOrEmpty(cscKey)) Then
					key = cscKey
				End If
				Dim cscEndpoint As String = TryGetConfigurationSetting(endpointConfiguration)
				If (Not String.IsNullOrEmpty(cscEndpoint)) Then
					endpoint = cscEndpoint
				End If
				Dim cscPathStyle As String = TryGetConfigurationSetting(usePathStyleUrisConfiguration)
				If (Not String.IsNullOrEmpty(cscPathStyle)) Then
					pathStyle = cscPathStyle
				End If

				' the Web.config can have references to csc setting strings
				' these count event stronger than the direct settings in the csc file
				Dim refWebName As String = TryGetAppSetting(CSConfigStringPrefix & accountNameConfiguration)
				If (Not String.IsNullOrEmpty(refWebName)) Then
					cscName = TryGetConfigurationSetting(refWebName)
					If (Not String.IsNullOrEmpty(cscName)) Then
						name = cscName
					End If
				End If
				Dim refWebKey As String = TryGetAppSetting(CSConfigStringPrefix & accountSharedKeyConfiguration)
				If (Not String.IsNullOrEmpty(refWebKey)) Then
					cscKey = TryGetConfigurationSetting(refWebKey)
					If (Not String.IsNullOrEmpty(cscKey)) Then
						key = cscKey
					End If
				End If
				Dim refWebEndpoint As String = TryGetAppSetting(CSConfigStringPrefix & endpointConfiguration)
				If (Not String.IsNullOrEmpty(refWebEndpoint)) Then
					cscEndpoint = TryGetConfigurationSetting(refWebEndpoint)
					If (Not String.IsNullOrEmpty(cscEndpoint)) Then
						endpoint = cscEndpoint
					End If
				End If
				Dim refWebPathStyle As String = TryGetAppSetting(CSConfigStringPrefix & usePathStyleUrisConfiguration)
				If (Not String.IsNullOrEmpty(refWebPathStyle)) Then
					cscPathStyle = TryGetConfigurationSetting(refWebPathStyle)
					If (Not String.IsNullOrEmpty(cscPathStyle)) Then
						pathStyle = cscPathStyle
					End If
				End If
			End If

			If String.IsNullOrEmpty(key) AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentException("No account key specified!")
			End If
			If String.IsNullOrEmpty(endpoint) AndAlso (Not allowIncompleteSettings) Then
				Throw New ArgumentException("No endpoint specified!")
			End If
			If String.IsNullOrEmpty(name) Then
				' in this case let's try to derive the account name from the Uri
				Dim newAccountName As String = Nothing
				Dim newBaseUri As Uri = Nothing
				If IsStandardStorageEndpoint(New Uri(endpoint), newAccountName, newBaseUri) Then
					Debug.Assert((newAccountName IsNot Nothing AndAlso newBaseUri IsNot Nothing) OrElse (newAccountName Is Nothing AndAlso newBaseUri Is Nothing))
					If newAccountName IsNot Nothing AndAlso newBaseUri IsNot Nothing Then
						endpoint = newBaseUri.AbsoluteUri
						name = newAccountName
					End If
				End If
				If String.IsNullOrEmpty(name) AndAlso (Not allowIncompleteSettings) Then
					Throw New ArgumentException("No account name specified.")
				End If
			End If

			Dim usePathStyleUris? As Boolean = Nothing
			If (Not String.IsNullOrEmpty(pathStyle)) Then
				Dim b As Boolean
				If (Not Boolean.TryParse(pathStyle, b)) Then
					Throw New ConfigurationErrorsException("Cannot parse value of setting UsePathStyleUris as a boolean")
				End If
				usePathStyleUris = b
			End If
			Dim tmpBaseUri As Uri = Nothing
			If (Not String.IsNullOrEmpty(endpoint)) Then
				tmpBaseUri = New Uri(endpoint)
			End If
			Return New StorageAccountInfo(tmpBaseUri, usePathStyleUris, name, key, allowIncompleteSettings)
		End Function


		''' <summary>
		''' Checks whether all essential properties of this object are set. Only then, the account info object 
		''' should be used in ohter APIs of this library.
		''' </summary>
		''' <returns></returns>
		Public Function IsCompleteSetting() As Boolean
			Return BaseUri IsNot Nothing AndAlso Base64Key IsNot Nothing AndAlso AccountName IsNot Nothing
		End Function

		''' <summary>
		''' Checks whether this StorageAccountInfo object is complete in the sense that all properties are set.
		''' </summary>
		Public Sub CheckComplete()
			If (Not IsCompleteSetting()) Then
				Throw New ConfigurationErrorsException("Account information incomplete!")
			End If
		End Sub

		#Region "Private methods"

		Private Shared Function TryGetConfigurationSetting(ByVal configName As String) As String
			Dim ret As String = Nothing
			Try
				ret = RoleManager.GetConfigurationSetting(configName)
			Catch e1 As RoleException
				Return Nothing
			End Try
			Return ret
		End Function

		<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification := "Make sure that nothing prevents us to read from the fabric's configuration envrionment.")> _
		Private Shared Function TryGetAppSetting(ByVal configName As String) As String
			Dim ret As String = Nothing
			Try
				ret = ConfigurationSettings.AppSettings(configName)
			' some exception happened when accessing the app settings section
			' most likely this is because there is no app setting file
			' we assume that this is because there is no app settings file; this is not an error
			' and explicitly all exceptions are captured here
			Catch e1 As Exception
				Return Nothing
			End Try
			Return ret
		End Function

		Private Shared ReadOnly Property GeneralAccountConfigurationExceptionString() As String
			Get
				Return "If the portal defines http://test.blob.core.windows.net as your blob storage endpoint, the string ""test"" " & "is your account name, and you can specify http://blob.core.windows.net as the BlobStorageEndpoint in your " & "service's configuration file(s)."
			End Get
		End Property

		<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1308:NormalizeStringsToUppercase")> _
		Private Shared Function IsStandardStorageEndpoint(ByVal baseUri As Uri, <System.Runtime.InteropServices.Out()> ByRef newAccountName As String, <System.Runtime.InteropServices.Out()> ByRef newBaseUri As Uri) As Boolean
			If baseUri Is Nothing Then
				Throw New ArgumentNullException("baseUri")
			End If

			newAccountName = Nothing
			newBaseUri = Nothing

			Dim host As String = baseUri.Host
			If String.IsNullOrEmpty(host) Then
				Throw New ArgumentException("The host part of the Uri " & baseUri.AbsoluteUri & " must not be null or empty.")
			End If
			If host <> host.ToLowerInvariant() Then
				Throw New ArgumentException("The specified host string " & host & " must not contain upper-case letters.")
			End If

			Dim suffix As String = Nothing
			If host.EndsWith(StorageHttpConstants.StandardPortalEndpoints.TableStorageEndpoint, StringComparison.Ordinal) Then
				suffix = StorageHttpConstants.StandardPortalEndpoints.TableStorageEndpoint
			End If
			If host.EndsWith(StorageHttpConstants.StandardPortalEndpoints.BlobStorageEndpoint, StringComparison.Ordinal) Then
				suffix = StorageHttpConstants.StandardPortalEndpoints.BlobStorageEndpoint
			End If
			If host.EndsWith(StorageHttpConstants.StandardPortalEndpoints.QueueStorageEndpoint, StringComparison.Ordinal) Then
				suffix = StorageHttpConstants.StandardPortalEndpoints.QueueStorageEndpoint
			End If
			' a URL as presented on the portal was specified, lets find out whether it is in the correct format
			If suffix IsNot Nothing Then
				Dim index As Integer = host.IndexOf(suffix, StringComparison.Ordinal)
				Debug.Assert(index <> -1)
				If index > 0 Then
					Dim first As String = host.Substring(0, index)
					Debug.Assert((Not String.IsNullOrEmpty(first)))
					If first.Chars(first.Length-1) <> StorageHttpConstants.ConstChars.Dot(0) Then
						Return False
					End If
					first = first.Substring(0, first.Length - 1)
					If String.IsNullOrEmpty(first) Then
						Throw New ArgumentException("The configured base URI " & baseUri.AbsoluteUri & " for the storage service is incorrect. " & GeneralAccountConfigurationExceptionString)
					End If
					If first.Contains(StorageHttpConstants.ConstChars.Dot) Then
						Throw New ArgumentException("The configured base URI " & baseUri.AbsoluteUri & " for the storage service is incorrect. " & GeneralAccountConfigurationExceptionString)
					End If
					newAccountName = first
					newBaseUri = New Uri(baseUri.Scheme + Uri.SchemeDelimiter & suffix & baseUri.PathAndQuery)
				End If
				Return True
			End If
			Return False
		End Function


		#End Region

	End Class
End Namespace
