﻿Imports System
Imports System.Collections.Generic
Imports System.Data.Services.Client
Imports System.Data.Services.Common
Imports System.Diagnostics
Imports System.Diagnostics.CodeAnalysis
Imports System.Globalization
Imports System.Linq
Imports System.Net
Imports System.Reflection
Imports System.Runtime.Serialization
'' <copyright file="TableStorage.cs" company="Microsoft">
''     Copyright (c) Microsoft Corporation.  All rights reserved.
'' </copyright>
''
'﻿' This file contains helper classes for accessing the table storage service:
''      - base classes for tables and table entities (table rows) containing the necessary 
''        partition key and row key values
''      - methods for applying the table storage authentication scheme and for handling 
''        authentication in DataServiceContext objects
''      - helper methods for creating, listing, and managing tables
''      - a set of table storage constants
''      - convenience methods for dealing with continuation tokens and paging
''      - simple error handling helpers
''      - retry semantics for table storage requests
'' 
'' Examples of how to make use of the classes and methods in this file are available 
'' in the simple sample application contained in this solution.  


Namespace Microsoft.Samples.ServiceHosting.StorageClient

	''' <summary>
	''' Class representing some important table storage constants.
	''' </summary>
	Public NotInheritable Class TableStorageConstants
		''' <summary>
		''' The maximum size of strings per property/column is 64 kB (that is 32k characters.)
		''' Note: This constant is smaller for the development storage table service.
		''' </summary>
		Public Shared ReadOnly MaxStringPropertySizeInBytes As Integer = 64 * 1024


		''' <summary>
		''' One character in the standard UTF-16 character presentation is 2 bytes.
		''' Note: This constant is smaller for the development storage table service.
		''' </summary>
		Public Shared ReadOnly MaxStringPropertySizeInChars As Integer = MaxStringPropertySizeInBytes \ 2

		''' <summary>
		''' We want to prevent users from the pitfall of mixing up Utc and local time.
		''' Because of this we add some time to the minimum supported datetime.
		''' As a result, there will be no error condition from the server even 
		''' if a user converts the minimum supported date time to a local time and 
		''' stores this in a DateTime field.
		''' The local development storage support the SQL range of dates which is narrower than the
		''' one for the table storage service and so we use that value here. 
		''' </summary>
		Public Shared ReadOnly MinSupportedDateTime As DateTime = DateTime.FromFileTime(0).ToUniversalTime().AddYears(200)

		'You can use this if you are programming against the real table storage service only but then your
		'code will not work against the local development table storage.
		'public static readonly DateTime MinSupportedDateTime = DateTime.FromFileTime(0).ToUniversalTime().AddDays(7);

		''' <summary>
		''' Internal constant for querying tables.
		''' </summary>
		Friend Const TablesName As String = "Tables"

		''' <summary>
		''' Internal constant for querying tables.
		''' </summary>
		Friend Const TablesQuery As String = "/" & TablesName
	End Class


	''' <summary>
	''' API entry point for using structured storage. The underlying usage pattern is designed to be 
	''' similar to the one used in blob and queue services in this library. 
	''' Users create a TableStorage object by calling the static Create() method passing account credential 
	''' information to this method. The TableStorage object can then be used to create, delete and list tables. 
	''' There are two methods to get DataServiceContext objects that conform to the appropriate security scheme. 
	''' The first way is to call the GetDataServiceContext() method on TableStorage objects. The naming is again 
	''' chosen to conform to the convention in the other APIs for blob and queue services in this library. 
	''' This class can also be used as an adapter pattern. I.e., DataServiceContext objects can be created 
	''' independnt from a TableStorage object. Calling the Attach() method will make sure that the appropriate 
	''' security signing is used on these objects. This design was chosen to support various usage patterns that 
	''' might become necessary for autogenerated code.
	''' </summary>
	Public Class TableStorage

		''' <summary>
		''' The default retry policy
		''' </summary>
        Public Shared ReadOnly DefaultRetryPolicy As RetryPolicy = AddressOf RetryPolicies.NoRetry


		''' <summary>
		''' Creates a TableStorage service object. This object is the entry point into the table storage API.
		''' </summary>
		''' <param name="baseUri">The base URI of the table storage service.</param>
		''' <param name="usePathStyleUris">Type of URI scheme used.</param>
		''' <param name="accountName">The account name.</param>
		''' <param name="base64Key">Base64 encoded version of the key.</param>
		''' <returns></returns>
		Public Shared Function Create(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String) As TableStorage
			Return New TableStorage(baseUri, usePathStyleUris, accountName, base64Key)
		End Function

		''' <summary>
		''' Creates a TableStorage object.
		''' </summary>
		Public Shared Function Create(ByVal info As StorageAccountInfo) As TableStorage
			Return TableStorage.Create(info.BaseUri, info.UsePathStyleUris, info.AccountName, info.Base64Key)
		End Function

		''' <summary>
		''' Infers a list of tables from a DataServiceContext-derived type and makes sure
		''' those tables exist in the given service. The table endpoint information is retrieved from the 
		''' standard configuration settings.
		''' </summary>
		''' <remarks>
		''' Tables are inferred by finding all the public properties of type IQueryable&lt;T&gt; in 
		''' the provided type, where T is a type with an ID (in the case of table storage, this means it either
		''' has a [DataServiceKey("PartitionKey", "RowKey")] attribute in the class, or derives from
		''' the TableStorageEntity class included in this sample library (which in turn has that attribute).
		''' </remarks>
        Public Shared Sub CreateTablesFromModel(ByVal serviceContextType As Type)
            CreateTablesFromModel(serviceContextType, StorageAccountInfo.DefaultTableStorageEndpointConfigurationString)
        End Sub

		''' <summary>
		''' Infers a list of tables from a DataServiceContext-derived type and makes sure
		''' those tables exist in the given service.        
		''' </summary>
		''' <param name="serviceContextType">The DataServiceContext type from which the tables are inferred.</param>
		''' <param name="endpointConfiguration">A configuration string that is used to determine the table storage endpoint.</param>
		Public Shared Sub CreateTablesFromModel(ByVal serviceContextType As Type, ByVal endpointConfiguration As String)
			Dim account As StorageAccountInfo = StorageAccountInfo.GetAccountInfoFromConfiguration(endpointConfiguration)
			CreateTablesFromModel(serviceContextType, account)
		End Sub

		''' <summary>
		''' Infers a list of tables from a DataServiceContext-derived type and makes sure
		''' those tables exist in the given service.        
		''' </summary>
		''' <param name="serviceContextType">The type of the DataServiceContext.</param>
		''' <param name="account">An object containing information about the table storage endpoint to be used.</param>
		Public Shared Sub CreateTablesFromModel(ByVal serviceContextType As Type, ByVal account As StorageAccountInfo)
			Dim tableStorage As TableStorage = TableStorage.Create(account)
			For Each tableName As String In DataServiceUtilities.EnumerateEntitySetNames(serviceContextType)
				tableStorage.TryCreateTable(tableName)
			Next tableName
		End Sub

		''' <summary>
		''' Creates a DataServiceContext object that takes care of implementing the table storage signing process.
		''' </summary>
		Public Function GetDataServiceContext() As TableStorageDataServiceContext
			Dim uriComponents As New ResourceUriComponents(_accountName)
			Dim uri As Uri = HttpRequestAccessor.ConstructResourceUri(_baseUri, uriComponents, _usePathStyleUris)
			Dim svc As New TableStorageDataServiceContext(uri, _accountName, _base64Key)
			If svc IsNot Nothing Then
				svc.RetryPolicy = Me.RetryPolicy
			End If
			Return svc
		End Function


		''' <summary>
		''' If the adaptor pattern with Attach() shall be used, this function can be used to generate the 
		''' table service base Uri depending on the path style syntax.
		''' </summary>
		Public Shared Function GetServiceBaseUri(ByVal baseUri As Uri, ByVal usePathStyleUris As Boolean, ByVal accountName As String) As Uri
			Dim uriComponents As New ResourceUriComponents(accountName)
			Dim uri As Uri = HttpRequestAccessor.ConstructResourceUri(baseUri, uriComponents, usePathStyleUris)
			Return uri
		End Function

		''' <summary>
		''' If the adaptor pattern with Attach() shall be used, this function can be used to generate the 
		''' table service base Uri depending on the path style syntax.
		''' </summary>
		Public Shared Function GetServiceBaseUri(ByVal account As StorageAccountInfo) As Uri
			Return GetServiceBaseUri(account.BaseUri, account.UsePathStyleUris, account.AccountName)
		End Function

		''' <summary>
		''' If DataServiceContext objects are created at different places, this method can be called to configure the 
		''' DataServiceContext object to implement the required security scheme.
		''' </summary>
		Public Sub Attach(ByVal svc As DataServiceContext)
			' this is an explicit way of dealing with situations where Attach() is called on objects that already 
			' have the necessary events hooked up to deal with table storage
			' in the event Attach() is called multiple times on a normal DataServiceContext object, we make sure to 
			' not doing authentication twice in the sending event itself
			If TypeOf svc Is TableStorageDataServiceContext Then
				Throw New ArgumentException("Cannot attach to a TableStorageDataServiceContext object. " & "These objects already contain the functionality for accessing the table storage service.")
			End If
			Dim TempContextRef As ContextRef = New ContextRef(Me, svc)
		End Sub

		Friend Function ListTableImpl(ByVal query As DataServiceQuery(Of TableStorageTable)) As IEnumerable(Of TableStorageTable)
			Dim ret As IEnumerable(Of TableStorageTable) = Nothing

            RetryPolicy.Invoke(Function() AnonymousMethod1(query, ret))
			Return ret
		End Function
		
        Private Function AnonymousMethod1(ByVal query As DataServiceQuery(Of TableStorageTable), ByRef ret As IEnumerable(Of TableStorageTable)) As Boolean
            Try
                ret = query.Execute()
            Catch e As InvalidOperationException
                If TableStorageHelpers.CanBeRetried(e) Then
                    Throw New TableRetryWrapperException(e)
                End If
                Throw
            End Try
            Return True
        End Function

        ''' <summary>
        ''' Lists all the tables under this service's URL
        ''' </summary>
        Public Function ListTables() As IEnumerable(Of String)
            Dim svc As DataServiceContext = GetDataServiceContext()
            Dim nextKey As String = Nothing
            Dim localQuery As DataServiceQuery(Of TableStorageTable)
            Dim tmp As IEnumerable(Of TableStorageTable)
            svc.MergeOption = MergeOption.NoTracking
            Dim query As IQueryable(Of TableStorageTable) = _
             From t In svc.CreateQuery(Of TableStorageTable)(TableStorageConstants.TablesName) _
             Select t
            ' result chunking
            ' if we would not do this, the default value of 1000 is used before query result pagination
            ' occurs
            query = query.Take(StorageHttpConstants.ListingConstants.MaxTableListResults)


            Dim orig As DataServiceQuery(Of TableStorageTable) = TryCast(query, DataServiceQuery(Of TableStorageTable))
            Try
                tmp = ListTableImpl(orig)
            Catch e As InvalidOperationException
                Dim status As HttpStatusCode
                If TableStorageHelpers.EvaluateException(e, status) AndAlso status = HttpStatusCode.NotFound Then
                    Return New List(Of String)
                End If
                Throw
            End Try
            If tmp Is Nothing Then
                Return New List(Of String)
            End If
            Dim tables As New List(Of String)
            For Each table As TableStorageTable In tmp
                tables.Add(table.TableName)
            Next table

            Dim qor As QueryOperationResponse = TryCast(tmp, QueryOperationResponse)
            qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextTableName, nextKey)

            Do While nextKey IsNot Nothing
                localQuery = orig
                localQuery = localQuery.AddQueryOption(StorageHttpConstants.HeaderNames.NextTableName, nextKey)
                Try
                    tmp = ListTableImpl(localQuery)
                Catch e As InvalidOperationException
                    Dim status As HttpStatusCode
                    If TableStorageHelpers.EvaluateException(e, status) AndAlso status = HttpStatusCode.NotFound Then
                        Return New List(Of String)
                    End If
                    Throw
                End Try
                If tmp Is Nothing Then
                    Return New List(Of String)
                End If

                For Each table As TableStorageTable In tmp
                    tables.Add(table.TableName)
                Next table
                qor = TryCast(tmp, QueryOperationResponse)
                qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextTableName, nextKey)
            Loop
            Return tables
        End Function

		''' <summary>
		''' Creates a new table in the service
		''' </summary>
		''' <param name="tableName">The name of the table to be created</param>
		Public Sub CreateTable(ByVal tableName As String)
			ParameterValidator.CheckStringParameter(tableName, False, "tableName")
			If (Not Utilities.IsValidTableName(tableName)) Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "The specified table name ""{0}"" is not valid!" & "Please choose a name that conforms to the naming conventions for tables!", tableName))
			End If
				' exceptions are DataServiceClientException, DataServiceQueryException and DataServiceRequestException
				' all of the above exceptions are InvalidOperationExceptions
							' don't retry in this case as this is an expected exception if the table exists
							' just return the exception
            RetryPolicy.Invoke(Function() AnonymousMethod2(tableName))
		End Sub
		
        Private Function AnonymousMethod2(ByRef tableName As String) As Boolean
            Try
                Dim svc As DataServiceContext = GetDataServiceContext()
                svc.AddObject(TableStorageConstants.TablesName, New TableStorageTable() With {.TableName = tableName})
                svc.SaveChanges()
            Catch e As InvalidOperationException
                Dim status As HttpStatusCode
                If TableStorageHelpers.CanBeRetried(e, status) Then
                    If status = HttpStatusCode.Conflict Then
                        Throw
                    Else
                        Throw New TableRetryWrapperException(e)
                    End If
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Tries to create a table with the given name.
		''' The main difference to the CreateTable method is that this function first queries the 
		''' table storage service whether the table already exists, before it tries to actually create 
		''' the table. The reason is that this 
		''' is more lightweight for the table storage service than always trying to create a table that 
		''' does already exist. Furthermore, as we expect that applications don't really randomly create
		''' tables, the additional roundtrip that is required for creating the table is necessary only very
		''' rarely.
		''' </summary>
		''' <param name="tableName">The name of the table.</param>
		''' <returns>True if the operation was completed successfully. False if the table already exists.</returns>
		Public Function TryCreateTable(ByVal tableName As String) As Boolean
			ParameterValidator.CheckStringParameter(tableName, False, "tableName")
			If DoesTableExist(tableName) Then
				Return False
			End If
			Try
				CreateTable(tableName)
				Return True
			Catch e As InvalidOperationException
				Dim status As HttpStatusCode
				If TableStorageHelpers.EvaluateException(e, status) AndAlso status = HttpStatusCode.Conflict Then
					Return False
				End If
				Throw
			End Try
		End Function

		''' <summary>
		''' Checks whether a table with the same name already exists.
		''' </summary>
		''' <param name="tableName">The name of the table to check.</param>
		''' <returns>True iff the table already exists.</returns>
		Public Function DoesTableExist(ByVal tableName As String) As Boolean
			ParameterValidator.CheckStringParameter(tableName, False, "tableName")
			Dim tableExists As Boolean = False

						' the query contains the whole primary key
						' thus, if the query succeeds we can be sure that the table exists
						'This is a workaround for bug in DataServiceQuery<T>.Execute. It throws a
						'NullReferenceException instead of a DataServiceRequestException when it
						'cannot connect to the the server. This workaround will be removed when
						'the fix for this bug is released.
            RetryPolicy.Invoke(Function() AnonymousMethod3(tableName, tableExists))
			Return tableExists
		End Function
		
        Private Function AnonymousMethod3(ByVal tableName As String, ByRef tableExists As Boolean) As Boolean
            Try
                Dim svc As DataServiceContext = GetDataServiceContext()
                svc.MergeOption = MergeOption.NoTracking
                Dim query As IEnumerable(Of TableStorageTable) = _
                 From t In svc.CreateQuery(Of TableStorageTable)(TableStorageConstants.TablesName) _
                 Where t.TableName = tableName _
                 Select t
                tableExists = False
                Try
                    TryCast(query, DataServiceQuery(Of TableStorageTable)).Execute()
                    tableExists = True
                Catch e As DataServiceQueryException
                    Dim s As HttpStatusCode
                    If TableStorageHelpers.EvaluateException(e, s) AndAlso s = HttpStatusCode.NotFound Then
                        tableExists = False
                    Else
                        Throw
                    End If
                Catch ne As NullReferenceException
                    Throw New DataServiceRequestException("Unable to connect to server.", ne)
                End Try
            Catch e As InvalidOperationException
                Dim status As HttpStatusCode
                If TableStorageHelpers.CanBeRetried(e, status) Then
                    Throw New TableRetryWrapperException(e)
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Deletes a table from the service.
		''' </summary>
		''' <param name="tableName">The name of the table to be deleted</param>
		Public Sub DeleteTable(ByVal tableName As String)
			ParameterValidator.CheckStringParameter(tableName, False, "tableName")

						' we do this even thouh NoContent is currently not in the exceptions that 
						' are retried
							' don't retry in this case, just return the exception
            RetryPolicy.Invoke(Function() AnonymousMethod4(tableName))
		End Sub
		
        Private Function AnonymousMethod4(ByRef tableName As String) As Boolean
            Try
                Dim svc As DataServiceContext = GetDataServiceContext()
                Dim table As New TableStorageTable() With {.TableName = tableName}
                svc.AttachTo(TableStorageConstants.TablesName, table)
                svc.DeleteObject(table)
                svc.SaveChanges()
            Catch e As InvalidOperationException
                Dim status As HttpStatusCode
                If TableStorageHelpers.CanBeRetried(e, status) Then
                    If status = HttpStatusCode.NoContent OrElse status = HttpStatusCode.NotFound Then
                        Throw
                    Else
                        Throw New TableRetryWrapperException(e)
                    End If
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Tries to delete the table with the given name. 
		''' </summary>
		''' <param name="tableName">The name of the table to delete.</param>
		''' <returns>True if the table was successfully deleted. False if the table does not exists.</returns>
		Public Function TryDeleteTable(ByVal tableName As String) As Boolean
			ParameterValidator.CheckStringParameter(tableName, False, "tableName")
			Try
				DeleteTable(tableName)
				Return True
			Catch e As InvalidOperationException
				Dim status As HttpStatusCode
				If TableStorageHelpers.EvaluateException(e, status) AndAlso status = HttpStatusCode.NotFound Then
					Return False
				End If
				Throw
			End Try
		End Function

		''' <summary>
		''' The retry policy used for retrying requests
		''' </summary>
		Private privateRetryPolicy As RetryPolicy
		Public Property RetryPolicy() As RetryPolicy
			Get
				Return privateRetryPolicy
			End Get
			Set(ByVal value As RetryPolicy)
				privateRetryPolicy = value
			End Set
		End Property

		''' <summary>
		''' The base URI of the table storage service
		''' </summary>
		Public ReadOnly Property BaseUri() As Uri
			Get
				Return Me._baseUri
			End Get
		End Property

		''' <summary>
		''' The name of the storage account
		''' </summary>
		Public ReadOnly Property AccountName() As String
			Get
				Return Me._accountName
			End Get
		End Property

		''' <summary>
		''' Indicates whether to use/generate path-style or host-style URIs
		''' </summary>
		Public ReadOnly Property UsePathStyleUris() As Boolean
			Get
				Return Me._usePathStyleUris
			End Get
		End Property

		''' <summary>
		''' The base64 encoded version of the key.
		''' </summary>
		Friend ReadOnly Property Base64Key() As String
			Get
				Return _base64Key
			End Get
		End Property

		Protected Friend Sub New(ByVal baseUri As Uri, ByVal usePathStyleUris? As Boolean, ByVal accountName As String, ByVal base64Key As String)
			Me._baseUri = baseUri
			Me._accountName = accountName
			Me._base64Key = base64Key
			If Not usePathStyleUris.HasValue Then
				Me._usePathStyleUris = Utilities.StringIsIPAddress(baseUri.Host)
			Else
				Me._usePathStyleUris = usePathStyleUris.Value
			End If
			RetryPolicy = DefaultRetryPolicy
		End Sub

		Private _baseUri As Uri
		Private _usePathStyleUris As Boolean
		Private _accountName As String
		Private _base64Key As String
	End Class



	Public NotInheritable Class TableStorageHelpers

		#Region "Error handling helpers"

		''' <summary>
		''' Checks whether the exception is or contains a DataServiceClientException and extracts the 
		''' returned http status code and extended error information.
		''' </summary>
		''' <param name="exception">The exception from which to extract information</param>
		''' <param name="status">The Http status code for the exception</param>
		''' <param name="extendedErrorInfo">Extended error information including storage service specific
		''' error code and error message</param>
		''' <returns>True if the exception is or contains a DataServiceClientException.</returns>
		Private Sub New()
		End Sub
		Public Shared Function EvaluateException(ByVal exception As Exception, <System.Runtime.InteropServices.Out()> ByRef status As HttpStatusCode, <System.Runtime.InteropServices.Out()> ByRef extendedErrorInfo As StorageExtendedErrorInformation) As Boolean
			Return EvaluateExceptionImpl(exception, status, extendedErrorInfo, True)
		End Function

		''' <summary>
		''' Checks whether the exception is or contains a DataServiceClientException and extracts the 
		''' returned http status code.
		''' </summary>
		''' <param name="exception">The exception from which to extract information</param>
		''' <param name="status">The Http status code for the exception</param>
		''' <returns>True if the exception is or contains a DataServiceClientException.</returns>
		Public Shared Function EvaluateException(ByVal exception As Exception, <System.Runtime.InteropServices.Out()> ByRef status As HttpStatusCode) As Boolean
			Dim extendedErrorInfo As StorageExtendedErrorInformation
			Return EvaluateExceptionImpl(exception, status, extendedErrorInfo, False)
		End Function

		Private Shared Function EvaluateExceptionImpl(ByVal e As Exception, <System.Runtime.InteropServices.Out()> ByRef status As HttpStatusCode, <System.Runtime.InteropServices.Out()> ByRef extendedErrorInfo As StorageExtendedErrorInformation, ByVal getExtendedErrors As Boolean) As Boolean
			status = HttpStatusCode.Unused
			extendedErrorInfo = Nothing
			Do While e.InnerException IsNot Nothing
				e = e.InnerException

				Dim dsce As DataServiceClientException = TryCast(e, DataServiceClientException)
				If dsce IsNot Nothing Then
					status = CType(dsce.StatusCode, HttpStatusCode)
					If getExtendedErrors Then
						extendedErrorInfo = Utilities.GetExtendedErrorFromXmlMessage(dsce.Message)
					End If
					Return True
				End If
			Loop
			Return False
		End Function

		''' <summary>
		''' Checks whether the exception is either a DataServiceClientException, a DataServiceQueryException or a 
		''' DataServiceRequestException.
		''' </summary>
		Public Shared Function IsTableStorageException(ByVal exception As Exception) As Boolean
			Return ((TypeOf exception Is DataServiceClientException) OrElse (TypeOf exception Is DataServiceQueryException) OrElse (TypeOf exception Is DataServiceRequestException))
		End Function

		''' <summary>
		''' Only certain classes of errors should be retried. This method evaluates an exception 
		''' and returns whether this class of exception can be retried.
		''' </summary>
		''' <param name="e">The exception to analyze.</param>
		''' <param name="statusCode">The HttpStatusCode retrieved from the exception.</param>
		Friend Shared Function CanBeRetried(ByVal e As InvalidOperationException, <System.Runtime.InteropServices.Out()> ByRef statusCode As HttpStatusCode) As Boolean
			Dim status As HttpStatusCode

			statusCode = HttpStatusCode.Unused
			If EvaluateException(e, status) Then
				statusCode = status
					' server error codes above 500
				If status = HttpStatusCode.RequestTimeout OrElse status = HttpStatusCode.ServiceUnavailable OrElse status = HttpStatusCode.InternalServerError OrElse status = HttpStatusCode.BadGateway OrElse status = HttpStatusCode.GatewayTimeout Then
					Return True
				End If
			End If
			Return False
		End Function

		''' <summary>
		''' Overload that does not retrun the HttpStatusCode.
		''' </summary>
		Friend Shared Function CanBeRetried(ByVal e As InvalidOperationException) As Boolean
			Dim ignored As HttpStatusCode
			Return CanBeRetried(e, ignored)
		End Function

		#End Region

		#Region "Methods for checking properties to be inserted into a table"

		''' <summary>
		''' Checks whether the string can be inserted in a table storage table. Throws an exception if 
		''' this is not the case.
		''' </summary>
		''' <param name="propertyValue"></param>
		Public Shared Sub CheckStringProperty(ByVal propertyValue As String)
			If String.IsNullOrEmpty(propertyValue) Then
				Throw New ArgumentException("The string cannot be null or empty!")
			End If
			If propertyValue.Length > TableStorageConstants.MaxStringPropertySizeInChars Then
				Throw New ArgumentException("The string cannot be longer than the maximum string property size.")
			End If
		End Sub

		''' <summary>
		''' Checks whether the string can be inserted into a table storage table.
		''' </summary>
		Public Shared Function ValidateStringProperty(ByVal propertyValue As String) As Boolean
			If String.IsNullOrEmpty(propertyValue) Then
				Return False
			End If
			If propertyValue.Length > TableStorageConstants.MaxStringPropertySizeInChars Then
				Return False
			End If
			Return True
		End Function

		#End Region

	End Class

	<DataServiceKey("TableName")> _
	Public Class TableStorageTable
		Private _tableName As String


		''' <summary>
		''' The table name.
		''' </summary>
		Public Property TableName() As String
			Get
				Return Me._tableName
			End Get

			Set(ByVal value As String)
				ParameterValidator.CheckStringParameter(value, False, "TableName")
				Me._tableName = value
			End Set
		End Property

		Public Sub New()
		End Sub

		''' <summary>
		''' Creates a table with the specified name.
		''' </summary>
		''' <param name="name">The name of the table.</param>
		Public Sub New(ByVal name As String)
			ParameterValidator.CheckStringParameter(name, False, "name")
			Me.TableName = name
		End Sub

		Public Overrides Overloads Function Equals(ByVal obj As Object) As Boolean
			If obj Is Nothing Then
				Return False
			End If
			Dim rhs As TableStorageTable = TryCast(obj, TableStorageTable)

			If rhs Is Nothing Then
				Return False
			End If

			Return (Me.TableName = rhs.TableName)
		End Function

		Public Overrides Function GetHashCode() As Integer
			Return Me.TableName.GetHashCode()
		End Function
	End Class


	''' <summary>
	''' This class represents an entity (row) in a table in table storage.
	''' </summary>
	<CLSCompliant(False), DataServiceKey("PartitionKey", "RowKey")> _
	Public MustInherit Class TableStorageEntity

		Private privateTimestamp As DateTime
		Public Property Timestamp() As DateTime
			Get
				Return privateTimestamp
			End Get
			Set(ByVal value As DateTime)
				privateTimestamp = value
			End Set
		End Property

		''' <summary>
		''' The partition key of a table entity. The concatenation of the partition key 
		''' and row key must be unique per table.
		''' </summary>
		Private privatePartitionKey As String
		Public Overridable Property PartitionKey() As String
			Get
				Return privatePartitionKey
			End Get
			Set(ByVal value As String)
				privatePartitionKey = value
			End Set
		End Property

		''' <summary>
		''' The row key of a table entity.
		''' </summary>
		Private privateRowKey As String
		Public Overridable Property RowKey() As String
			Get
				Return privateRowKey
			End Get
			Set(ByVal value As String)
				privateRowKey = value
			End Set
		End Property

		''' <summary>
		''' Creates a TableStorageEntity object.
		''' </summary>
		Protected Sub New(ByVal partitionKey As String, ByVal rowKey As String)
			PartitionKey = partitionKey
			RowKey = rowKey
		End Sub

		''' <summary>
		''' Creates a TableStorageEntity object.
		''' </summary>
		Protected Sub New()
		End Sub


		''' <summary>
		''' Compares to entities.
		''' </summary>
		Public Overrides Overloads Function Equals(ByVal obj As Object) As Boolean
			If obj Is Nothing Then
				Return False
			End If
			Dim rhs As TableStorageEntity = TryCast(obj, TableStorageEntity)

			If rhs Is Nothing Then
				Return False
			End If

			Return (Me.PartitionKey = rhs.PartitionKey AndAlso Me.RowKey = rhs.RowKey)
		End Function


		''' <summary>
		''' Computes a HashCode for this object.
		''' </summary>
		Public Overrides Function GetHashCode() As Integer
			If PartitionKey Is Nothing Then
				Return MyBase.GetHashCode()
			End If
			If (Not String.IsNullOrEmpty(Me.RowKey)) Then
				Return Me.PartitionKey.GetHashCode() Xor Me.RowKey.GetHashCode()
			Else
				Return Me.PartitionKey.GetHashCode()
			End If
		End Function
	End Class

	''' <summary>
	''' This class can be used for handling continuation tokens in TableStorage.
	''' </summary>
	''' <typeparam name="TElement"></typeparam>
	Public Class TableStorageDataServiceQuery(Of TElement)

		Private _query As DataServiceQuery(Of TElement)

		''' <summary>
		''' Objects of this class can be created using this constructor directly or by 
		''' calling a factory method on the TableStorageDataServiceContext class
		''' </summary>
		Public Sub New(ByVal query As DataServiceQuery(Of TElement))
            Me.New(query, AddressOf RetryPolicies.NoRetry)
		End Sub

		''' <summary>
		''' Objects of this class can be created using this constructor directly or by 
		''' calling a factory method on the TableStorageDataServiceContext class
		''' </summary>
		Public Sub New(ByVal query As DataServiceQuery(Of TElement), ByVal policy As RetryPolicy)
			If query Is Nothing Then
				Throw New ArgumentNullException("query")
			End If
			_query = query
			RetryPolicy = policy
		End Sub

		''' <summary>
		''' Gets the underlying normal query object.
		''' </summary>
		Public Property Query() As DataServiceQuery(Of TElement)
			Get
				Return _query
			End Get
			Set(ByVal value As DataServiceQuery(Of TElement))
				_query = value
			End Set
		End Property


		''' <summary>
		''' The retry policy used for retrying requests
		''' </summary>
		Private privateRetryPolicy As RetryPolicy
		Public Property RetryPolicy() As RetryPolicy
			Get
				Return privateRetryPolicy
			End Get
			Set(ByVal value As RetryPolicy)
				privateRetryPolicy = value
			End Set
		End Property

		''' <summary>
		''' Normal Execute() on the query without retry. Just maps to _query.Execute().
		''' </summary>
		''' <returns></returns>
		Public Function Execute() As IEnumerable(Of TElement)
            Return ExecuteWithRetries(AddressOf RetryPolicies.NoRetry)
		End Function

		''' <summary>
		''' Calling Execute() on the query with the current retry policy.
		''' </summary>
		''' <returns>An IEnumerable respresenting the results of the query.</returns>
		Public Function ExecuteWithRetries() As IEnumerable(Of TElement)
			Return ExecuteWithRetries(RetryPolicy)
		End Function

		''' <summary>
		''' Calling Execute() on the query with the current retry policy.
		''' </summary>
		''' <param name="retry">The retry policy to be used for this request.</param>
		''' <returns>An IEnumerable respresenting the results of the query.</returns>
		Public Function ExecuteWithRetries(ByVal retry As RetryPolicy) As IEnumerable(Of TElement)
			Dim ret As IEnumerable(Of TElement) = Nothing
			If retry Is Nothing Then
				Throw New ArgumentNullException("retry")
			End If
			retry(Function() AnonymousMethod5(ret))
			Return ret
		End Function
		
        Private Function AnonymousMethod5(ByRef ret As IEnumerable(Of TElement)) As Boolean
            Try
                ret = _query.Execute()
            Catch e As InvalidOperationException
                If TableStorageHelpers.CanBeRetried(e) Then
                    Throw New TableRetryWrapperException(e)
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Returns all results of the query and hides the complexity of continuation if 
		''' this is desired by a user. Users should be aware that this operation can return 
		''' many objects. Uses no retries.
		''' Important: this function does not call Execute immediately. Instead, it calls Execute() on 
		''' the query only when the result is enumerated. This is a difference to the normal 
		''' Execute() and Execute() with retry method.         
		''' </summary>
		''' <returns>An IEnumerable representing the results of the query.</returns>
		Public Function ExecuteAll() As IEnumerable(Of TElement)
			Return ExecuteAll(False)
		End Function

		''' <summary>
		''' Returns all results of the query and hides the complexity of continuation if 
		''' this is desired by a user. Users should be aware that this operation can return 
		''' many objects. This operation also uses the current retry policy.
		''' Important: this function does not call Execute immediately. Instead, it calls Execute() on 
		''' the query only when the result is enumerated. This is a difference to the normal 
		''' Execute() and Execute() with retry method. 
		''' </summary>
		''' <returns>An IEnumerable representing the results of the query.</returns>
		Public Function ExecuteAllWithRetries() As IEnumerable(Of TElement)
			Return ExecuteAll(True)
		End Function

		''' <summary>
		''' Returns all results of the query and hides the complexity of continuation if 
		''' this is desired by a user. Users should be aware that this operation can return 
		''' many objects. 
		''' Important: this function does not call Execute immediately. Instead, it calls Execute() on 
		''' the query only when the result is enumerated. This is a difference to the normal 
		''' Execute() and Execute() with retry method. 
		''' </summary>
		''' <param name="withRetry">Determines whether to use retries or not.</param>
		''' <returns>An IEnumerable representing the results of the query.</returns>
		Public Function ExecuteAll(ByVal withRetry As Boolean) As IEnumerable(Of TElement)
			Dim res As IEnumerable(Of TElement)
			Dim tmp As IEnumerable(Of TElement)
			Dim nextPartitionKey As String = Nothing
			Dim nextRowKey As String = Nothing
			Dim localQuery As DataServiceQuery(Of TElement)

			If _query Is Nothing Then
				Throw New ArgumentException("The local DataServiceQuery element cannot be null!")
			End If

			If withRetry Then
				res = ExecuteWithRetries()
			Else
				res = _query.Execute()
			End If
			If res Is Nothing Then
                Return New List(Of TElement)
            End If
            Dim elements As New List(Of TElement)
			For Each item As TElement In res
                elements.Add(item)
			Next item
			Dim qor As QueryOperationResponse = TryCast(res, QueryOperationResponse)
			qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextPartitionKey, nextPartitionKey)
			qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextRowKey, nextRowKey)

			Do While nextPartitionKey IsNot Nothing OrElse nextRowKey IsNot Nothing
				localQuery = _query
				If nextPartitionKey IsNot Nothing Then
					localQuery = localQuery.AddQueryOption(StorageHttpConstants.HeaderNames.NextPartitionKey, nextPartitionKey)
				End If
				If nextRowKey IsNot Nothing Then
					localQuery = localQuery.AddQueryOption(StorageHttpConstants.HeaderNames.NextRowKey, nextRowKey)
				End If
				If withRetry Then
					Dim retryQuery As New TableStorageDataServiceQuery(Of TElement)(localQuery, RetryPolicy)
					tmp = retryQuery.ExecuteWithRetries()
				Else
					tmp = localQuery.Execute()
				End If
				If tmp Is Nothing Then
                    Return New List(Of TElement)
				End If
				For Each item As TElement In tmp
                    elements.Add(item)
				Next item
				qor = TryCast(tmp, QueryOperationResponse)
				qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextPartitionKey, nextPartitionKey)
				qor.Headers.TryGetValue(StorageHttpConstants.HeaderNames.PrefixForTableContinuation + StorageHttpConstants.HeaderNames.NextRowKey, nextRowKey)
            Loop
            Return elements
        End Function

	End Class

	''' <summary>
	''' The table storage-specific DataServiceContext class. It adds functionality for handling 
	''' the authentication process required by the table storage service.
	''' </summary>
	Public Class TableStorageDataServiceContext
		Inherits DataServiceContext
		Private _sharedKey As String
		Private _accountName As String

		''' <summary>
		''' Creates a DataServiceContext object and configures it so that it can be used with the table storage service.
		''' </summary>
		''' <param name="serviceRoot">The root URI of the service.</param>
		''' <param name="accountName">The account name.</param>
		''' <param name="sharedKey">The shared key associated with this service.</param>
		Friend Sub New(ByVal serviceRoot As Uri, ByVal accountName As String, ByVal sharedKey As String)
			MyBase.New(serviceRoot)
			If String.IsNullOrEmpty(accountName) Then
				Throw New ArgumentNullException("accountName")
			End If
			If String.IsNullOrEmpty(sharedKey) Then
				Throw New ArgumentNullException("sharedKey")
			End If
			_sharedKey = sharedKey
			_accountName = accountName
			AddHandler SendingRequest, AddressOf DataContextSendingRequest
			IgnoreMissingProperties = True

			' we assume that this is the expected client behavior
			MergeOption = MergeOption.PreserveChanges

			RetryPolicy = TableStorage.DefaultRetryPolicy
		End Sub

		''' <summary>
		''' Creates a DataServiceContext object and configures it so that it can be used with the table storage service.
		''' </summary>
		''' <param name="account">A StorageAccountInfo object containing information about how to access the table storage service.</param>
		Public Sub New(ByVal account As StorageAccountInfo)
			Me.New(TableStorage.GetServiceBaseUri(account), account.AccountName, account.Base64Key)
		End Sub


		''' <summary>
		''' Creates a DataServiceContext object and configures it so that it can be used with the table storage service.
		''' Information on the table storage endpoint is retrieved by accessing configuration settings in the app config section 
		''' of a Web.config or app config file, or by accessing settings in cscfg files.
		''' </summary>
		Public Sub New()
			Me.New(StorageAccountInfo.GetAccountInfoFromConfiguration(StorageAccountInfo.DefaultTableStorageEndpointConfigurationString))
		End Sub


		''' <summary>
		''' The retry policy used for retrying requests
		''' </summary>
		Private privateRetryPolicy As RetryPolicy
		Public Property RetryPolicy() As RetryPolicy
			Get
				Return privateRetryPolicy
			End Get
			Set(ByVal value As RetryPolicy)
				privateRetryPolicy = value
			End Set
		End Property

		''' <summary>
		''' Calls the SaveChanges() method and applies retry semantics.
		''' </summary>
		Public Function SaveChangesWithRetries() As DataServiceResponse
			Dim ret As DataServiceResponse = Nothing
            RetryPolicy.Invoke(Function() AnonymousMethod6(ret))
			Return ret
		End Function
		
        Private Function AnonymousMethod6(ByRef ret As DataServiceResponse) As Boolean
            Try
                ret = SaveChanges()
            Catch e As InvalidOperationException
                If TableStorageHelpers.CanBeRetried(e) Then
                    Throw New TableRetryWrapperException(e)
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Calls the SaveChanges() method and applies retry semantics.
		''' </summary>
		Public Function SaveChangesWithRetries(ByVal options As SaveChangesOptions) As DataServiceResponse
			Dim ret As DataServiceResponse = Nothing
            RetryPolicy.Invoke(Function() AnonymousMethod7(options, ret))
			Return ret
		End Function
		
        Private Function AnonymousMethod7(ByVal options As SaveChangesOptions, ByRef ret As DataServiceResponse) As Boolean
            Try
                ret = SaveChanges(options)
            Catch e As InvalidOperationException
                If TableStorageHelpers.CanBeRetried(e) Then
                    Throw New TableRetryWrapperException(e)
                End If
                Throw
            End Try
            Return True
        End Function

		''' <summary>
		''' Callback method called whenever a request is sent to the table service. This 
		''' is where the signing of the request takes place.
		''' </summary>
		Private Sub DataContextSendingRequest(ByVal sender As Object, ByVal e As SendingRequestEventArgs)
			Dim request As HttpWebRequest = TryCast(e.Request, HttpWebRequest)

			' this setting can potentially result in very rare error conditions from HttpWebRequest
			' using retry policies, these error conditions are dealt with
			request.KeepAlive = True

			' do the authentication
			Dim key() As Byte
			Dim credentials As SharedKeyCredentials

			Debug.Assert(_sharedKey IsNot Nothing)
			key = Convert.FromBase64String(_sharedKey)
			credentials = New SharedKeyCredentials(_accountName, key)
			credentials.SignRequestForSharedKeyLite(request, New ResourceUriComponents(_accountName, _accountName))
		End Sub
	End Class

	#Region "Internal and private helpers"

	''' <summary>
	''' Helper class to avoid long-lived references to context objects
	''' </summary>
	''' <remarks>
	''' Need to be careful not to maintain a reference to the context
	''' object from the auth adapter, since the adapter is probably
	''' long-lived and the context is not. This intermediate helper
	''' class is the one subscribing to context events, so when the
	''' context can be collected then this will be collectable as well.
	''' </remarks>
	Friend Class ContextRef
		Private _adapter As TableStorage

		Public Sub New(ByVal adapter As TableStorage, ByVal context As DataServiceContext)
			If adapter Is Nothing Then
				Throw New ArgumentNullException("adapter")
			End If
			If context Is Nothing Then
				Throw New ArgumentNullException("context")
			End If
			Me._adapter = adapter
			AddHandler context.SendingRequest, AddressOf HandleSendingRequest
		End Sub

		Private Sub HandleSendingRequest(ByVal sender As Object, ByVal e As SendingRequestEventArgs)
			Dim request As HttpWebRequest = TryCast(e.Request, HttpWebRequest)

			' first, we have to make sure that the request is not signed twice
			' this could happen if Attach() is called multiple times on the same DataServiceContext object
			Dim col As WebHeaderCollection = request.Headers
			If col IsNot Nothing Then
				For Each header As String In col
					If String.Compare(header, StorageHttpConstants.HeaderNames.StorageDateTime, StringComparison.Ordinal) = 0 Then
						Return
					End If
				Next header
			End If

			' this setting can potentially result in very rare error conditions from HttpWebRequest
			' using retry policies, these error conditions are dealt with
			request.KeepAlive = True

			' do the authentication
			Dim key() As Byte
			Dim credentials As SharedKeyCredentials

			Debug.Assert(_adapter.Base64Key IsNot Nothing)
			key = Convert.FromBase64String(_adapter.Base64Key)
			credentials = New SharedKeyCredentials(_adapter.AccountName, key)
			credentials.SignRequestForSharedKeyLite(request, New ResourceUriComponents(_adapter.AccountName, _adapter.AccountName))
		End Sub
	End Class

	''' <summary>
	''' The retry policies for blobs and queues deal with special StorageClient and StorageServer exceptions.
	''' In case of tables, we don't want to return these exceptions but instead the normal data service 
	''' exception. This class serves as a simple wrapper for these exceptions, and indicates that we 
	''' need retries.
	''' Data service exceptions are stored as inner exceptions.
	''' </summary>
	<Serializable> _
	Public Class TableRetryWrapperException
		Inherits Exception
		''' <summary>
		''' Creates a TableRetryWrapperException object.
		''' </summary>
		Public Sub New()
			MyBase.New()
		End Sub

		''' <summary>
		''' Creates a TableRetryWrapperException object.
		''' </summary>
		Public Sub New(ByVal innerException As Exception)
			MyBase.New(String.Empty, innerException)
		End Sub

		''' <summary>
		''' Creates a TableRetryWrapperException object.
		''' </summary>
		Public Sub New(ByVal message As String)
			MyBase.New(message)
		End Sub

		''' <summary>
		''' Creates a TableRetryWrapperException object.
		''' </summary>
		Public Sub New(ByVal message As String, ByVal innerException As Exception)
			MyBase.New(message, innerException)
		End Sub

		''' <summary>
		''' Creates a TableRetryWrapperException object.
		''' </summary>
		Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
			MyBase.New(info, context)
		End Sub

	End Class

	Friend NotInheritable Class ParameterValidator
		Private Sub New()
		End Sub
		Friend Shared Sub CheckStringParameter(ByVal s As String, ByVal canBeNullOrEmpty As Boolean, ByVal name As String)
			If String.IsNullOrEmpty(s) AndAlso (Not canBeNullOrEmpty) Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "The parameter {0} cannot be null or empty.", name))
			End If
			If s.Length > TableStorageConstants.MaxStringPropertySizeInChars Then
				Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "The parameter {0} cannot be longer than {1} characters.", name, TableStorageConstants.MaxStringPropertySizeInChars))
			End If
		End Sub
	End Class

	Friend NotInheritable Class DataServiceUtilities
		Private Sub New()
		End Sub
		Public Shared Function IsEntityType(ByVal t As Type, ByVal contextType As Type) As Boolean
			' ADO.NET data services convention: a type 't' is an entity if
			' 1) 't' has at least one key column
			' 2) there is a top level IQueryable<T> property in the context where T is 't' or a supertype of 't'
			' Non-primitive types that are not entity types become nested structures ("complex types" in EDM)

			If (Not t.GetProperties().Any(Function(p) IsKeyColumn(p))) Then
				Return False
			End If

			For Each pi As PropertyInfo In contextType.GetProperties()
				If GetType(IQueryable).IsAssignableFrom(pi.PropertyType) Then
					If pi.PropertyType.GetGenericArguments()(0).IsAssignableFrom(t) Then
						Return True
					End If
				End If
			Next pi

			Return False
		End Function

		Public Shared Function IsKeyColumn(ByVal pi As PropertyInfo) As Boolean
			' Astoria convention:
			' 1) try the DataServiceKey attribute
			' 2) if not attribute, try <typename>ID
			' 3) finally, try just ID

			Dim attribs() As Object = pi.DeclaringType.GetCustomAttributes(GetType(DataServiceKeyAttribute), True)
			If attribs IsNot Nothing AndAlso attribs.Length > 0 Then
				Debug.Assert(attribs.Length = 1)
				Return (CType(attribs(0), DataServiceKeyAttribute)).KeyNames.Contains(pi.Name)
			End If

			If pi.Name.Equals(pi.DeclaringType.Name & "ID", System.StringComparison.OrdinalIgnoreCase) Then
				Return True
			End If

			If pi.Name = "ID" Then
				Return True
			End If

			Return False
		End Function

        Public Shared Function EnumerateEntitySetProperties(ByVal contextType As Type) As IEnumerable(Of PropertyInfo)
            Dim props As New List(Of PropertyInfo)
            For Each prop As PropertyInfo In contextType.GetProperties()
                If GetType(IQueryable).IsAssignableFrom(prop.PropertyType) AndAlso prop.PropertyType.GetGenericArguments().Length > 0 AndAlso DataServiceUtilities.IsEntityType(prop.PropertyType.GetGenericArguments()(0), contextType) Then
                    props.Add(prop)
                End If
            Next prop
            Return props
        End Function

		Public Shared Function EnumerateEntitySetNames(ByVal contextType As Type) As IEnumerable(Of String)
			Return EnumerateEntitySetProperties(contextType).Select(Function(p) p.Name)
		End Function

	End Class

	#End Region
End Namespace
